/*
 *	alsa-midi-io.cc
 *
 */

#ifdef WITH_ALSA

#include <alsa/asoundlib.h>
#include <sys/errno.h>

#include "alsa-midi-io.h"

#define DBG_ERRORS	2
#define DBG_EVENTS	0

/* for snd_seq_event_input_pending(): do what it's called here */
#define FETCH_FROM_KERNEL 1

static void
alsa_snd_lib_error_handler (
	const char *file, 
	int line, 
	const char *function, 
	int err, 
	const char *fmt, ...
)
{
	_dbgprintf (1, "%s:%d:%s():%s\n", file, line, function, strerror (err));
}

alsa_port::alsa_port()
:
m_port ("alsa_port")
{
	handle = 0;
	i_can_read = 0;
	i_can_write = 0;

	snd_lib_error_set_handler (alsa_snd_lib_error_handler);
	
	int err = snd_seq_open (
		(snd_seq_t **) &handle, "hw",
		SND_SEQ_OPEN_OUTPUT | SND_SEQ_OPEN_INPUT, 
		SND_SEQ_NONBLOCK
	);
	
	if (!handle)
	{
		fprintf (stderr, "no snd_seq_handle obtained (%s)\n", snd_strerror (err));
		return;
	}

	q = snd_seq_alloc_queue ((snd_seq_t *) handle);
		
	port = snd_seq_create_simple_port (
		(snd_seq_t *) handle, 
		"thing:-:-",
		
		SND_SEQ_PORT_CAP_READ |
		SND_SEQ_PORT_CAP_SUBS_READ |
		SND_SEQ_PORT_CAP_WRITE |
		SND_SEQ_PORT_CAP_SUBS_WRITE,

		SND_SEQ_PORT_TYPE_MIDI_GENERIC
	);
	
	client = snd_seq_client_id ((snd_seq_t *) handle);

	errno = 0;
}

alsa_port::~alsa_port()
{
	if (handle)
	{
		snd_seq_delete_simple_port ((snd_seq_t *) handle, port);
		snd_seq_free_queue ((snd_seq_t *) handle, q);
		snd_seq_close ((snd_seq_t *) handle);
	}
}

int
alsa_port::get_read_fd()
{
	if (!i_can_read)
		return -1;
	
	#if SND_LIB_MINOR >= 6
	struct pollfd fds[2];

	int n = snd_seq_poll_descriptors ((snd_seq_t *) handle, fds, 2, POLLIN);

	if (n >= 1 && fds[0].events & POLLIN)
		return fds[0].fd;
	if (n >= 2 && fds[1].events & POLLIN)
		return fds[1].fd;
		
	return -1;
	#else
	return snd_seq_file_descriptor ((snd_seq_t *) handle);
	#endif
}

int
alsa_port::get_write_fd()
{
	if (!i_can_write)
		return -1;
	
	#if SND_LIB_MINOR >= 6
	struct pollfd fds[2];

	int n = snd_seq_poll_descriptors ((snd_seq_t *) handle, fds, 2, POLLOUT);
	
	if (n >= 1 && fds[0].events & POLLOUT)
		return fds[0].fd;
	if (n >= 2 && fds[1].events & POLLOUT)
		return fds[1].fd;
		
	return -1;
	#else
	return snd_seq_file_descriptor ((snd_seq_t *) handle);
	#endif
}

bool
alsa_port::connect (int major, int minor)
{
	snd_seq_port_info_t * pinfo;
	if (snd_seq_port_info_malloc (&pinfo))
		return false;

	snd_seq_get_any_port_info ((snd_seq_t *) handle, major, minor, pinfo);
	int caps = snd_seq_port_info_get_capability (pinfo);
	
	/* check for RW access */
	if (!((caps & SND_SEQ_PORT_CAP_SUBS_READ) || (caps & SND_SEQ_PORT_CAP_SUBS_WRITE)))
		return false;

	/* update my port info */
	snd_seq_get_port_info ((snd_seq_t *) handle, port, pinfo);

	char name [128];
	sprintf (name, "thing:%d:%d", major, minor);
	
	snd_seq_port_info_set_name (pinfo, name);
	snd_seq_set_port_info ((snd_seq_t *) handle, port, pinfo);
	
	snd_seq_port_info_free (pinfo);
	/* done with information setting */
	
	snd_seq_port_subscribe_t * subscription;
	if (snd_seq_port_subscribe_malloc (&subscription))
		return false;
	
	snd_seq_port_subscribe_set_queue (subscription, q);
	/* how to do this? subscription.convert_time = 0; */

	snd_seq_addr_t 
		me = {client, port},
		he = {major, minor};
		
	if (caps & SND_SEQ_PORT_CAP_SUBS_WRITE)
	{
		snd_seq_port_subscribe_set_sender (subscription, &me);
		snd_seq_port_subscribe_set_dest (subscription, &he);
	
		i_can_write = (snd_seq_subscribe_port ((snd_seq_t *) handle, subscription) >= 0);
	}

	if (caps & SND_SEQ_PORT_CAP_SUBS_READ)
	{
		snd_seq_port_subscribe_set_sender (subscription, &he);
		snd_seq_port_subscribe_set_dest (subscription, &me);
		
		i_can_read = (snd_seq_subscribe_port ((snd_seq_t *) handle, subscription) >= 0);
	}

	snd_seq_port_subscribe_free (subscription);

	return i_can_read || i_can_write;
}

bool
alsa_port::connect (char * name)
{
	snd_seq_client_info_t *cinfo;
	snd_seq_port_info_t *pinfo;
	
	if (snd_seq_client_info_malloc (&cinfo))
		return false;
	if (snd_seq_port_info_malloc (&pinfo))
	{
		snd_seq_client_info_free (cinfo);
		return false;
	}
	
	snd_seq_client_info_set_client (cinfo, -1);
	
	while (snd_seq_query_next_client ((snd_seq_t *) handle, cinfo) >= 0) 
	{
		snd_seq_port_info_set_client (pinfo, snd_seq_client_info_get_client (cinfo));
		snd_seq_port_info_set_port (pinfo, -1);

		while (snd_seq_query_next_port ((snd_seq_t *) handle, pinfo) >= 0) 
		{
			const char * pname = snd_seq_port_info_get_name (pinfo);
			if (!strncmp (pname, name, strlen (name)))
			{
				bool result = connect (
					snd_seq_port_info_get_client (pinfo),
					snd_seq_port_info_get_port (pinfo)
				);

				snd_seq_client_info_free (cinfo);
				snd_seq_port_info_free (pinfo);

				return result;
			}
		}
	}

	snd_seq_client_info_free (cinfo);
	snd_seq_port_info_free (pinfo);

	return false;
}

void
alsa_port::iterate (alsa_port_iterator * iterator)
{
	snd_seq_client_info_t *cinfo;
	snd_seq_port_info_t *pinfo;
	
	if (snd_seq_client_info_malloc (&cinfo))
		return;
	if (snd_seq_port_info_malloc (&pinfo))
	{
		snd_seq_client_info_free (cinfo);
		return;
	}
	
	while (snd_seq_query_next_client ((snd_seq_t *) handle, cinfo) >= 0) 
	{
		snd_seq_port_info_set_client (pinfo, snd_seq_client_info_get_client (cinfo));
		snd_seq_port_info_set_port (pinfo, -1);

		while (snd_seq_query_next_port ((snd_seq_t *) handle, pinfo) >= 0) 
			iterator->next (cinfo, pinfo);
	}
	
	snd_seq_client_info_free (cinfo);
	snd_seq_port_info_free (pinfo);
}

/*
 *	send an event
 *
 */

int
alsa_port::write_event (m_event * e, m_event_stamp * stamp)
{
	if (e->is_meta())
		return -1;
		
	if (!i_can_write)
		return -1;

	// e->debug ("alsa_port::write_event");

	int _channel = (stamp && stamp->channel >= 0) ? stamp->channel : e->get_channel();

	snd_seq_event_t packet;

	snd_seq_ev_clear (&packet);
	snd_seq_ev_set_source (&packet, port);

	packet.queue = q;

	packet.flags = SND_SEQ_TIME_STAMP_TICK | SND_SEQ_TIME_MODE_REL;
	packet.time.tick = 0;

	snd_seq_ev_set_subs (&packet);
	
	switch (e->get_type())
	{
		case M_NOTE_ON:
			packet.type = SND_SEQ_EVENT_NOTEON;
			packet.data.note.channel = _channel;
			packet.data.note.note = e->get_note();
			packet.data.note.velocity = e->get_velocity_on();
			break;

		case M_NOTE_OFF:
			packet.type = SND_SEQ_EVENT_NOTEOFF;
			packet.data.note.channel = _channel;
			packet.data.note.note = e->get_note();
			packet.data.note.velocity = e->get_velocity_on();
			packet.data.note.off_velocity = e->get_velocity_off();
			break;
		
		case M_NOTE_TOUCH:
			packet.type = SND_SEQ_EVENT_KEYPRESS;
			packet.data.note.channel = _channel;
			packet.data.note.note = e->get_note();
			packet.data.note.velocity = e->get_velocity_on();
			break;
		
		case M_PROGRAM_CHANGE:
			packet.type = SND_SEQ_EVENT_PGMCHANGE;
			packet.data.control.channel = _channel;
			packet.data.control.value = e->get_param_0();
			break;

		case M_CONTROL_CHANGE:
			packet.type = SND_SEQ_EVENT_CONTROLLER;
			packet.data.control.channel = _channel;
			packet.data.control.param = e->get_param_0();
			packet.data.control.value = e->get_param_1();
			break;

		case M_CHANNEL_TOUCH:
			packet.type = SND_SEQ_EVENT_CHANPRESS;
			packet.data.control.channel = _channel;
			packet.data.control.value = e->get_param_0();
			break;
		
		case M_PITCH_WHEEL:
			packet.type = SND_SEQ_EVENT_PITCHBEND;
			packet.data.control.channel = _channel;
			packet.data.control.value = (int) (M_PITCH_MAX * (1 + e->get_pitch()));
			break;

		case M_SYSEX:
			{
				c_char_buffer * buffer = e->get_data_buffer();
				snd_seq_ev_set_sysex (&packet, buffer->count, buffer->items);
			}
			break;

		default:
			switch (e->get_type())
			{
				case M_RT_QUARTER_FRAME:
				case M_RT_SONG_POS:
				case M_RT_SONG_SELECT:
				case M_RT_BUS_SELECT:
				case M_RT_TUNE_REQUEST:
					return 0;
				case M_EOX:
					return 0;
					
				case M_RT_CLOCK:
					packet.type = SND_SEQ_EVENT_CLOCK;
					break;
					
				case M_RT_START:
				case M_RT_CONTINUE:
				case M_RT_STOP:
				case M_RT_ACTIVE_SENSE:
					return 0;
			}
			return 0;
	}

	int r = snd_seq_event_output_direct ((snd_seq_t *) handle, &packet);
	if (r < 0)
		_dbgprintf (DBG_ERRORS, "event write failed %s\n", strerror (-r)); 
	
	return 0;
}

/*
 *	fetch events.
 *
 */

void
alsa_port::parse_input (m_tick tick, byte port, m_connector * out)
{
	int n = snd_seq_event_input_pending ((snd_seq_t *) handle, FETCH_FROM_KERNEL);
	if (!n)
		return;

	if (n < 0)
	{
		_dbgprintf (DBG_ERRORS, "snd_seq_event_input_pending = %s\n", strerror (-n));
		return;
	}

	while (n--)
	{
		snd_seq_event_t * packet;
		int error = snd_seq_event_input ((snd_seq_t *) handle, &packet);
		
		if (error)
		{
			_dbgprintf (2, "alsa_port::parse_input could not fetch packet: %s\n", snd_strerror (error));
			continue;
		}
		
		_dbgprintf (DBG_EVENTS, "[read_events] [%d]\n", packet->type);
		
		m_event * e = new m_event;
		C_INCREF (e);
		
		switch (packet->type)
		{
			case SND_SEQ_EVENT_NOTEON:
				e->init (
					M_NOTE_ON | packet->data.note.channel,
					packet->data.note.note,
					packet->data.note.velocity
				);
				break;
				
			case SND_SEQ_EVENT_KEYPRESS:
				e->init (
					M_NOTE_TOUCH | packet->data.note.channel,
					packet->data.note.note,
					packet->data.note.velocity
				);
				break;
	
			case SND_SEQ_EVENT_NOTEOFF:
				e->init (
					M_NOTE_OFF | packet->data.note.channel,
					packet->data.note.note,
					packet->data.note.velocity
				);
				break;
				
			case SND_SEQ_EVENT_CONTROLLER:
				e->init (
					M_CONTROL_CHANGE | packet->data.control.channel,
					packet->data.control.param,
					packet->data.control.value
				);
				break;

			case SND_SEQ_EVENT_PGMCHANGE:
				e->init (
					M_PROGRAM_CHANGE | packet->data.control.channel,
					packet->data.control.value
				);
				break;

			case SND_SEQ_EVENT_CHANPRESS:
				e->init (
					M_CHANNEL_TOUCH | packet->data.control.channel,
					packet->data.control.value
				);
				break;

			case SND_SEQ_EVENT_PITCHBEND:
				{
					int pitch = packet->data.control.value;
					e->init (
						M_PITCH_WHEEL | packet->data.control.channel,
						pitch >> 7,
						pitch & 0x7F
					);
				}
				break;

			case SND_SEQ_EVENT_SENSING:
				break;

			default: 
				_dbgprintf (1, "[read_events] unknown event [%d]\n", packet->type);
				break;
		}

		/* we have one. */
		if (e->midi_status() && out)	
		{
			e->set_tick (tick);
			e->set_port (port);
			out->event_in (e);
		}
		
		C_DECREF (e);

		/* clean up. */
		snd_seq_free_event (packet);
	}
}

#endif  /* WITH_ALSA */

