/*
 *	arch-threads.cc
 *
 *	this abstraction of POSIX threads has been introduced 
 *	mainly to ease debugging.
 *	
 */

#include <sys/time.h>
#include <unistd.h>
#include <stdio.h>
#include <pthread.h>
#include <errno.h>
#include <string.h>
#include <sched.h>

#include "arch-types.h"
#include "arch-threads.h"

#define DBG_MUTEX	0
#define DBG_LOCK	0

/*
 *	a_thread
 *
 */

a_thread::a_thread()
{
	_thread_running = 0;
	_quit_requested = 0;
}

a_thread::~a_thread()
{
}

bool
a_thread::thread_start()
{
	if (!
		(_thread_running = !pthread_create 
			(
				&thread_id, NULL, 
				(void * (*) (void *)) thread_init_func,
				this
			)
		)
	) _dbgprintf (1, "[a_thread] cannot start: %s\n", strerror (errno));

	return _thread_running;
}

bool
a_thread::rt_thread_start (int priority)
{
	if (seteuid (0))
		_dbgprintf (2, "cannot seteuid (0): %s\n", strerror (errno));

	struct sched_param sched;
	sched.sched_priority = priority;
	
	
	pthread_attr_t attr;
	
	if (pthread_attr_init (&attr))
		_dbgprintf (2, "pthread_attr_init: %s\n", strerror (errno));
	
	if (pthread_attr_setschedpolicy (&attr, SCHED_FIFO))
		_dbgprintf (2, "pthread_attr_setschedpolicy: %s\n", strerror (errno));
				
	if (pthread_attr_setschedparam (&attr, &sched))
		_dbgprintf (2, "pthread_attr_setschedparam: %s\n", strerror (errno));
	
	pthread_attr_setscope (&attr, PTHREAD_SCOPE_SYSTEM);
	
	int result = pthread_create 
	(
		&thread_id, &attr, 
		(void * (*) (void *)) thread_init_func,
		this
	);
	if (!result)
		_thread_running = true;
	else
		_dbgprintf (2, "rt_thread_start: %s\n", strerror (result));
	
	seteuid (getuid());

	return _thread_running;
}

void
a_thread::thread_exit (void * return_value)
{
	pthread_exit (return_value);
}

void *
a_thread::thread_init_func (a_thread * self)
{
	void * ret_value = self->thread();

	self->_thread_running = false;

	return ret_value;
}

/*
 *	a_mutex
 *
 */
 
a_mutex::a_mutex (int kind)
{
#ifdef DEBUG
	count = 0;
	_dbgprintf (DBG_MUTEX, "<mutex %x>\n", (int) this);
#endif

	pthread_mutexattr_t attr;
	
	pthread_mutexattr_init (&attr);
	
	switch (kind)
	{
		case Fast :
			pthread_mutexattr_settype (&attr, PTHREAD_MUTEX_DEFAULT);
			break;
		case Checking :
			pthread_mutexattr_settype (&attr, PTHREAD_MUTEX_ERRORCHECK);
			break;
		case Recursive :
		default :
			pthread_mutexattr_settype (&attr, PTHREAD_MUTEX_RECURSIVE);
			break;
	}

	pthread_mutex_init (&mutex, &attr);
}

a_mutex::~a_mutex()
{
#ifdef DBG_MUTEX
	_dbgprintf (DBG_MUTEX, "<~mutex %x>\n", (int) this);
#endif
	pthread_mutex_destroy (&mutex);
}

void
a_mutex::lock()
{
#ifdef DEBUG
	if (++count >= 0)
		_dbgprintf (DBG_LOCK, "<lock %dx %x>\n", count, (int) this);
#endif
	pthread_mutex_lock (&mutex);
}

void
a_mutex::unlock()
{
#ifdef DEBUG
	if (count-- >= 0)
		_dbgprintf (DBG_LOCK, "<unlock %dx %x>\n", count, (int) this);
#endif
	pthread_mutex_unlock (&mutex);
}

/*
 *	a_cond
 *	
 */

a_cond::a_cond()
: a_mutex (Fast)
{
	pthread_cond_init (&cond, NULL);
}

a_cond::~a_cond()
{
	pthread_cond_destroy (&cond);
}

void
a_cond::signal()
{
	pthread_cond_signal (&cond);
}

void
a_cond::wait()
{
	pthread_cond_wait (&cond, &mutex);
}

bool
a_cond::timed_wait (double seconds)
{
	struct timeval tv;
	gettimeofday (&tv, NULL);
	
	seconds += (double) tv.tv_sec + (double) tv.tv_usec / TEN_TO_THE_SIXTH;
		
	struct timespec ts;
	ts.tv_sec = (long) seconds;
	ts.tv_nsec = (long) ((seconds - (double) ts.tv_sec) * TEN_TO_THE_NINTH);

	int r;
	do {
		r = pthread_cond_timedwait (&cond, &mutex, &ts);
	} while (r == EINTR);

	return (r != ETIMEDOUT);
}

/*
 *	a_semaphore
 *
 */

a_semaphore::a_semaphore()
{
	#define NO_PROCESS_SHARING 0
	sem_init (&sem, NO_PROCESS_SHARING, 0);
	#undef NO_PROCESS_SHARING
}

a_semaphore::~a_semaphore()
{
	sem_destroy (&sem);
}

int			
a_semaphore::wait()
{
	return sem_wait (&sem);
}

int			
a_semaphore::try_wait()
{
	return !sem_trywait (&sem);
}

int			
a_semaphore::up()
{
	return sem_post (&sem);
}

int			
a_semaphore::value()
{
	int r;
	
	sem_getvalue (&sem, &r);
	
	return r;
}

/*
 * this code is adapted to c++ from the pthread-faq file:
 * 
 * $Id: rwlock.c,v 1.9 1999/02/27 14:19:35 lk Exp $
 *
 * Routines to implement a read-write lock. Multiple readers or one writer
 * can hold the lock at once. Writers are given priority over readers.
 * When compiled with RWLOCK_DEBUG defined verbose debugging output is
 * produced which can help track problems such as mismatches and
 * recursive locks.
 *
 * Copyright (C) 1998-99 Lee Kindness <leekindness@iname.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

a_rwlock::a_rwlock()
{
	pthread_mutex_init (&lock, NULL);
	pthread_cond_init (&wcond, NULL);
	pthread_cond_init (&rcond, NULL);
	
	lock_count = 0;
	waiting_writers = 0;
}

a_rwlock::~a_rwlock()
{
	pthread_mutex_destroy (&lock);
	pthread_cond_destroy (&wcond);
	pthread_cond_destroy (&rcond);
}

/*
 * rwlock_ReadLock()
 *
 * Obtain a read lock.
 */

void 
a_rwlock::lock_read()
{
    pthread_mutex_lock (&lock);
    pthread_cleanup_push (
			(void (*) (void *)) waiting_reader_cleanup, (void *) this);
    while ((lock_count < 0) && (waiting_writers))
			pthread_cond_wait (&rcond, &lock);
    lock_count++;
    /* Note that the pthread_cleanup_pop subroutine will
     * execute the rwlocki_WaitingReaderCleanup routine */
    pthread_cleanup_pop (1);
}

void 
a_rwlock::unlock_read()
{
	pthread_mutex_lock (&lock);
	lock_count--;
	if (!lock_count)
		pthread_cond_signal (&wcond);
	pthread_mutex_unlock(&lock);
}

void
a_rwlock::lock_write()
{
	pthread_mutex_lock (&lock);
	waiting_writers++;
	pthread_cleanup_push (
		(void (*) (void *)) waiting_writer_cleanup, (void *) this);
	while (lock_count)
		pthread_cond_wait (&wcond, &lock);
	lock_count = -1;
	/* Note that the pthread_cleanup_pop subroutine will
	 * execute the rwlocki_WaitingWriterCleanup routine */
	pthread_cleanup_pop(1);
}
	
void 
a_rwlock::unlock_write()
{
	pthread_mutex_lock (&lock);
	lock_count = 0;
	if (!waiting_writers)
		pthread_cond_broadcast (&rcond);
	else
		pthread_cond_signal (&wcond);
	pthread_mutex_unlock (&lock);
}

void 
a_rwlock::waiting_reader_cleanup (a_rwlock * self)
{
	pthread_mutex_unlock (&self->lock);
}

void 
a_rwlock::waiting_writer_cleanup (a_rwlock * self)
{
	self->waiting_writers--;

	if ((!self->waiting_writers) && (self->lock_count >= 0))
 		/*  This only happens if we have been cancelled */
		pthread_cond_broadcast (&self->wcond);
	
	pthread_mutex_unlock (&self->lock);
}


