/*
 *	common-list.h
 *	
 *	simple list of items
 *	
 */
#ifndef _COMMON_LIST_
#define _COMMON_LIST_

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "arch-types.h"
#include "arch-threads.h"
#include "common-refcount.h"

#define	DEFAULT_GRANULARITY	10

/*
 *	the mother of all lists.
 *
 *	thanks to gcc for compiling this :)
 *	
 */

template <class T> 
class c_list 
{
	public:
		T *			items;
		int			count;
		int			allocated;
		int			granularity;

		a_mutex	mutex;

									c_list (int _granularity = DEFAULT_GRANULARITY);
									c_list (c_list *);

		void					allocate (int _allocate);
		void					deallocate (int _allocate);
		void					set_length (int _length);

		T *						forget_allocation();

		T &						operator[] (int i)
										{return items[i];}
		
		virtual int		index_of (T & item)
										{int i = &item - items; 
										 return (i < 0 || i >= count) ? -1 : i;}

		void *				item (int i) {return (void *) &items[i];}

		virtual T &		insert_item (const T & t, int index = -1);
		virtual int		insert (int index = -1);
		
		virtual bool	remove_item (T & t);
		virtual bool	remove (int index);

		bool					remove_slice (int i0, int i1);

		void					move (T &, T &);
		void					move (int from, int to);

		virtual void	clear();
		virtual			 ~c_list();
};

/*
 *	keep a list of ref_obj'ects, incref'ing and decref'ing them
 *	on insertion and removal.
 *
 */

class c_ref_list
:
public c_list<c_ref_obj *>
{
	public:
		virtual				 ~c_ref_list();

		virtual void		clear();
		virtual int			find (c_ref_obj *);
		virtual int			insert (c_ref_obj *, int index = -1);
		virtual bool		remove (int);
		virtual bool		remove (c_ref_obj *);
		virtual bool		replace (int i, c_ref_obj *);
		virtual bool		swap (int i, int j);
};

/*
 *	c_char_buffer for string and data
 *
 *	use 
 *		buffer << (char) b
 *	for literals.
 */

class c_char_buffer
:
public c_list<char>
{
	public:
					c_char_buffer (int _granularity = sizeof (int));
					c_char_buffer (c_char_buffer * other);
					
		bool	read (FILE * fp, int length);
		void	bzero();
		void	set (void * data, int length);

		c_char_buffer & operator << (c_char_buffer * other);

		#ifdef DEBUG
		byte & operator[] (int i);
		#else
		byte & operator[] (int i) {return (byte) items[i];}
		#endif

		c_char_buffer & append (void * b, int len);
		c_char_buffer & operator << (char * s);
		c_char_buffer & operator << (int i);
		c_char_buffer & operator << (byte b);
		c_char_buffer & operator << (char c) {return append (&c, 1);};
		c_char_buffer & operator << (double d);
		
		int	operator >> (FILE * fp) {return fwrite (items, 1, count, fp);}
};

/*
 *	c_char_buffer with a write pointer and automatic cleanup
 *
 */

class c_out_buffer 
:
public c_char_buffer
{
	public:
		int	next_write;

							c_out_buffer() : c_char_buffer()
								{next_write = 0;}
		
							/* -1 = write all you have */
		int				write (int fd, int n_bytes = -1);
};

/*
 *	here goes.
 *
 */

template <class T>
c_list<T>::c_list<T> (int _granularity)
: mutex()
{
	allocated	= 0;
	granularity	= _granularity;

	count		= 0;
	items		= 0;
	
	allocate (granularity);
}

template <class T>
c_list<T>::~c_list<T>()
{
	if (items)
		free (items);
}

template <class T>
void
c_list<T>::allocate (int _allocated)
{
	allocated	+= _allocated;

	if (items)
		items	= (T *) realloc (items, sizeof (T) * allocated);
	else
		items	= (T *) malloc (sizeof (T) * allocated);
}

template <class T>
void
c_list<T>::deallocate (int _allocated)
{
	if (allocated - _allocated < count)
		return;

	allocated	-= _allocated;

	items = (T *) realloc (items, sizeof (T) * allocated);
}

template <class T>
void
c_list<T>::set_length (int _count)
{
	allocated = count = _count;
	if (!allocated)
		allocated = granularity;

	if (items)
		items	= (T *) realloc (items, sizeof (T) * allocated);
	else
		items	= (T *) malloc (sizeof (T) * allocated);
}

template <class T>
T *
c_list<T>::forget_allocation()
{
	allocated = 0;
	count = 0;

	T * was = items;
	items = 0;

	return was;
}

template <class T>
T &
c_list<T>::insert_item (const T & t, int i)
{
	return items [insert (i)] = t;
}

template <class T>
int
c_list<T>::insert (int index)
{
	if (!items) 
	{
		allocate	(granularity);
		count		= 1;
		index		= 0;
	} 
	else 
	{
		if (index < 0 || index > count)
			index	= count;

		if (count == allocated - 1)
			allocate	(granularity);

		if (index < count)
			memmove		(item (index + 1), item (index), (count - index) * sizeof (T));

		count++;
	}

	return index;
}

template <class T>
void
c_list<T>::move (int from, int to)
{
	if (from == to)
		return;

	assert (from >= 0 && from < count);
	assert (to >= 0 && to < count);

	void * tmp = malloc (sizeof (T));
	memcpy (tmp, item (from), sizeof (T));
	
	if (to > from)
		memmove (item (from), item (from + 1), (to - from) * sizeof (T));
	else
		memmove (item (to + 1), item (to), (from - to) * sizeof (T));

	memcpy (item (to), tmp, sizeof (T));
	
	free (tmp);
}

template <class T>
bool
c_list<T>::remove_item (T & t)
{
	int i = index_of (t);
	if (i < 0)
		return false;
	
	return remove (i);
}

template <class T>
bool
c_list<T>::remove (int index)
{
	if (index < 0 || index >= count)
		return 0;

	if (index < count - 1)
		memmove (item (index), item (index + 1), (count - index) * sizeof (T));	

	count--;
		
	if (count && allocated - count < granularity)
		deallocate (granularity);

	return true;
}

template <class T>
bool
c_list<T>::remove_slice (int i0, int i1)
{
	assert (i0 >= 0 && i0 < count);
	assert (i1 >= 0 && i1 <= count);
	assert (i1 > i0);
	
	if (count - i1)
		memmove (item (i0), item (i1), (count - i1) * sizeof (T));

	count -= (i1 - i0);

	return true;
}

template <class T>
void
c_list<T>::clear()
{
	count = 0;

	if (allocated > granularity) 
		deallocate (allocated - granularity);
}

#endif
