/*
 *	common-rgb-classes.cc
 *
 *	rgb image buffer handling routines
 *
 */

#include <malloc.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "common-rgb-classes.h"

#undef MIN
#define MIN(a, b)  (((a) < (b)) ? (a) : (b))
#define MAX(a, b)  (((a) > (b)) ? (a) : (b))

/*
 *	clipping rectangles
 *
 */

bool
c_rectangle::clip (c_rectangle & that)
{
	int x1 = x0 + width;
	int y1 = y0 + height;
	
	if (x0 < that.x0)
		x0 = that.x0;
	if (y0 < that.y0)
		y0 = that.y0;

	if (y1 > that.y0 + that.height)
		y1 = that.y0 + that.height;

	if (x1 > that.x0 + that.width)
		x1 = that.x0 + that.width;
	
	width = x1 - x0;
	height = y1 - y0;

	return (width > 0 && height > 0);
}

/*
 *	growing rectangles
 *
 */

void
c_rectangle::swallow (c_rectangle & that)
{
	int fx0 = MIN (x0, that.x0);
	int fx1 = MAX (x0 + width, that.x0 + that.width);

	int fy0 = MIN (y0, that.y0);
	int fy1 = MAX (y0 + height, that.y0 + that.height);

	x0 = fx0;
	y0 = fy0;
	width = fx1 - fx0;
	height = fy1 - fy0;
}

/*
 *	just a list of colors.
 *
 */
 
c_palette::c_palette()
:
c_list<c_color>()
{
}

int
c_palette::add (unsigned int packed)
{
	int index = c_list<c_color>::insert();
	if (index < 0)
		return -1;
		
	(*this)[index].set (packed);	

	return index;
}

/*
 * rgb buffer
 *
 */

c_rgb_buffer::c_rgb_buffer (int _width, int _height, int _pixelstride)
:fore (0,0,0), back (255,255,255)
{
	pixelstride = _pixelstride;

	buffer = NULL;

	resize (_width, _height);
}

c_rgb_buffer::~c_rgb_buffer()
{
	if (buffer)
		free (buffer);
}

/*
 * resize
 *
 */

bool
c_rgb_buffer::resize (int _width, int _height)
{
	width = _width;
	height = _height;

	return allocate();
}

/*
 * allocate the buffer
 *
 */

bool
c_rgb_buffer::allocate()
{
	rowstride = width * pixelstride;
	while (rowstride & pixelstride)
		rowstride++;
	
	int size = height * rowstride;

	if (buffer)
		buffer = (byte *) realloc (buffer, size);
	else
		buffer = (byte *) malloc (size);

	return (bool) buffer;
}


/*
 *	scroll the buffer
 *
 */

void
c_rgb_buffer::scroll (int dx, int dy)
{
	int bytes = pixelstride * (width - abs (dx));
	int lines = height - abs (dy);
	
	byte * src, * dest;
	
	int advance;

	if (dy > 0)
		advance = -rowstride,
		
		src = line_ptr (height - 1 - dy),
		dest = line_ptr (height - 1);
		
	else
		advance = rowstride,
		
		src = line_ptr (-dy),
		dest = buffer;

	if (dx < 0)
		src += pixelstride * -dx;
	else
		dest += pixelstride * dx;

	_dbgprintf (1, "-| %p %p %d\n", src, dest, bytes);

	while (lines--)
	{
		memcpy (dest, src, bytes);
		src += advance;
		dest += advance;
	}
}

/*
 *	fill a rectangle
 *
 */
 
void
c_rgb_buffer::rect (c_rectangle * r, c_color * c)
{
	rect (r->x0, r->y0, r->width, r->height, c);
}

void
c_rgb_buffer::rect (int x0, int y0, int width, int height, c_color * c)
{
	int x1 = x0 + width;
	int y1 = y0 + height;
	
	if (pixelstride == 3)
		for (int y = y0; y < y1; y++)
		{
			byte * p = pixel_ptr (x0, y);
			for (int x = x0; x < x1; x++)
			{
				*p++ = c->r;
				*p++ = c->g;
				*p++ = c->b;
			}
		}
	else
		for (int y = y0; y < y1; y++)
		{
			byte * p = pixel_ptr (x0, y);
			for (int x = x0; x < x1; x++)
			{
				*p++ = c->r;
				*p++ = c->g;
				*p++ = c->b;
				 p++;
			}
		}
}

/*
 *	color, alpha pixel
 *
 */

void
c_rgb_buffer::apixel (int x, int y, c_color * c, int alpha)
{
	byte * p = pixel_ptr (x, y);
	int inverse_alpha = 0xff ^ alpha;

	*p = (((((int) *p) * inverse_alpha) + (c->r * alpha)) >> 8);
	p++;
	*p = (((((int) *p) * inverse_alpha) + (c->g * alpha)) >> 8);
	p++;
	*p = (((((int) *p) * inverse_alpha) + (c->b * alpha)) >> 8);
}

/*
 *	color, alpha rect
 *
 */

void
c_rgb_buffer::arect (c_rectangle * r, c_color * c, int alpha)
{
	arect (r->x0, r->y0, r->width, r->height, c, alpha);
}

void
c_rgb_buffer::arect (int x0, int y0, int width, int height, c_color * c, int alpha)
{
	int x1 = x0 + width;
	int y1 = y0 + height;
	int inverse_alpha = 0xff - alpha;

	if (pixelstride == 3)
		for (int y = y0; y < y1; y++)
		{
			byte * p = pixel_ptr (x0, y);
			for (int x = x0; x < x1; x++)
			{
				*p = (((((int) *p) * inverse_alpha) + (c->r * alpha)) >> 8);
				p++;
				*p = (((((int) *p) * inverse_alpha) + (c->g * alpha)) >> 8);
				p++;
				*p = (((((int) *p) * inverse_alpha) + (c->b * alpha)) >> 8);
				p++;
			}
		}
	else
		for (int y = y0; y < y1; y++)
		{
			byte * p = pixel_ptr (x0, y);
			for (int x = x0; x < x1; x++)
			{
				*p = (((((int) *p) * inverse_alpha) + (c->r * alpha)) >> 8);
				 p++;
				*p = (((((int) *p) * inverse_alpha) + (c->g * alpha)) >> 8);
				 p++;
				*p = (((((int) *p) * inverse_alpha) + (c->b * alpha)) >> 8);
				 p++;
				 p++;
			}
		}
}

/*
 *	antialiased line
 *
 *	algorithm thanks to Wu (who?) and Bresenham
 *
 */

void
c_rgb_buffer::aaline (int x0, int y0, int x1, int y1, c_color * c)
{
	long x, y, x_inc, y_inc;
	long dx = x1 - x0, dy = y1 - y0;
	int h;
	
	if (abs (dx) > abs (dy))
	{
		if (dx < 0)
		{
			dx = -dx;
			dy = -dy;
			
			h = x1;
			x1 = x0;
			x0 = h;

			h = y1;
			y1 = y0;
			y0 = h;
		}
		x = x0 << 16 | 0x7fff;
		y = y0 << 16 | 0x7fff;
		y_inc = (dy << 16) / dx;

		while ((x >> 16) < x1)
		{
			apixel (x >> 16, y >> 16, c, (~y & 0xFFFF) >> 8);
			apixel (x >> 16, (y >> 16) + 1, c, (y & 0xFFFF) >> 8);
			x += (1 << 16);
			y += y_inc;
		}
	}
	else
	{
		if (dy < 0)
		{
			dx = -dx;
			dy = -dy;
			
			h = x1;
			x1 = x0;
			x0 = h;

			h = y1;
			y1 = y0;
			y0 = h;
		}
		x = x0 << 16 | 0x7fff;
		y = y0 << 16 | 0x7fff;
		x_inc = (dx << 16) / dy;

		while ((y >> 16) < y1)
		{
			apixel (x >> 16, y >> 16, c, (~x & 0xFFFF) >> 8);
			apixel ((x >> 16) + 1, y >> 16, c, (x & 0xFFFF) >> 8);
			x += x_inc;
			y += (1 << 16);
		}
	}
}

