/*
 *	gtk-event-view.h
 *
 */

#ifdef WITH_GTK

#include <stdlib.h>

#include "gtk-event-view.h"

#include <gdk/gdkx.h>
#include <gdk/gdkkeysyms.h>

#define GTK_CONNECT(a,b,c) gtk_signal_connect (GTK_OBJECT (a), b, (GtkSignalFunc) c, this)

/*
 *	gtk_event_drawer
 *
 */

gtk_event_drawer::gtk_event_drawer (gtk_event_view * _view, c_rectangle * _clip)
:
m_connector ("gtk_event_drawer")
{
	view = _view;
	clip = *_clip;

	bar_snap = 0;
	beat_snap = 0;
	fraction_snap = 1;
	
	/* easier clipping by remembering the displayed quantize in ticks */
	tick.start = view->x_mapper.x2tick (clip.x0 - 1);
	tick.end = view->x_mapper.x2tick (clip.x0 + clip.width + 1);
}

bool
gtk_event_drawer::event_in (m_event * e, m_event_stamp * stamp)
{
	//m_tick _tick = e->get_tick();
	int channel = (stamp && stamp->channel >= 0) ? stamp->channel : e->get_channel();
	int port = (stamp && stamp->port >= 0) ? stamp->port : e->get_port();

	switch (e->get_type())
	{
		case M_NOTE_ON:
		/*
			XXX ?!
			if (_tick >= tick.end)
				return false;
			if (_tick >= tick.start)
				view->draw_note (e, &clip);
			if (_tick + e->get_note_length() >= tick.start)
		*/
			
				view->draw_note (e, view->get_color (port, channel), &clip);
			return true;
	}

	return true;
}

void
gtk_event_drawer::click_in (m_tick tick, int bar, int beat, int fraction)
{
	if (bar_snap && (beat || fraction || (bar % bar_snap)))
		return;
	else if (beat_snap && (fraction || (beat % beat_snap)))
		return;
	else if (fraction_snap && (fraction % fraction_snap))
		return;

	int alpha;
	
	if (!fraction)
	{
		if (!beat)
			alpha = 0x40;
		else
			alpha = 0x24;
	}
	else
		alpha = 0x10;
		
	view->draw_click (tick, alpha, &clip);
}

/*
 *	gtk_event_view
 *
 */

gtk_event_view::gtk_event_view (GtkWidget * _widget)
:
c_ref_obj ("gtk_event_view"),
song(),
selection(),
x_mapper (&song),
y_mapper(),
exposures(),
velocity_to_alpha(),
fore (0x000000),
staff (0x000000),
back (0xfef7d3),
loop (0xff0000),
colors()
{
	song.disable_refcounting();
	selection.disable_refcounting();

	pixels.size.set (0, 0, 0, 0);
	
	pixels.map = NULL;
	pixels.buffer = NULL;

	scroll_pending.dx = 
	scroll_pending.dy = 0;

	clicked.dz = 0;
	clicked.button_mask = 0;

	magnification = 0;
	snap = 192;

	y_mapper.set_magnification (magnification, y_mapper.note2y (60));
	x_mapper.set_magnification (magnification, x_mapper.tick2x (4 * 96));

	set_magnification (1);

	/* velocity to alpha mapping ramp */
	for (int i = 0; i < 128; i++)
		velocity_to_alpha.insert_item (0x10 + (i * 240) / 127);

	init_colors();
	staff_alpha = 0x10;

	follow.song = NULL;
	follow.redraw = false;
	follow.tick = 0;

	/*	gtk initialization	*/
	gc = NULL;
	widget = _widget ? _widget : gtk_drawing_area_new();

	gtk_widget_ref (widget);

	gdk_rgb_init();

	GTK_CONNECT (widget, "configure-event", configure_event);
	GTK_CONNECT (widget, "expose-event", expose_event);
	GTK_CONNECT (widget, "motion-notify-event", motion_event);
	GTK_CONNECT (widget, "button-press-event", button_event);
	GTK_CONNECT (widget, "key-press-event", key_event);
	GTK_CONNECT (widget, "key-release-event", key_event);

	gtk_widget_set_events
	(
		widget,
		//GDK_EXPOSURE_MASK |
		GDK_POINTER_MOTION_MASK |
		GDK_BUTTON_PRESS_MASK |
		GDK_BUTTON_RELEASE_MASK |
		GDK_KEY_PRESS_MASK |
		GDK_KEY_RELEASE_MASK
	);
	
	timeout_id = gtk_timeout_add (40, (GtkFunction) timeout_event, this);
}

unsigned int
gtk_event_view::get_xwindow()
{
	return GDK_WINDOW_XWINDOW (widget->window);
}

gtk_event_view::~gtk_event_view()
{
	gtk_timeout_remove (timeout_id);

	if (widget)
	{
		if (widget->window)
			gdk_window_set_back_pixmap (widget->window, NULL, FALSE);
		gtk_widget_unref (widget);
	}

	if (pixels.map)
		gdk_pixmap_unref (pixels.map);

	if (pixels.buffer)
		delete pixels.buffer;
}



/* * * * * * * * * * * * * * 
 *
 *	set_*
 *
 */

void
gtk_event_view::set_follow (m_song * _song)
{
	if (follow.song != _song)
	{
		C_XINCREF (_song);
		C_XDECREF (follow.song);
		follow.song = _song;
	}
	follow.redraw = true;
}

void
gtk_event_view::set_magnification (int _magnification, int x, int y)
{
	if (magnification == _magnification)
		return;
	
	if (x < 0)
		x = pixels.size.width / 2;
	if (y < 0)
		y = pixels.size.height / 2;

	_dbgprintf (0, "(mag %d)\n", _magnification);

	magnification = _magnification;
	
	x_mapper.set_magnification (magnification, x);
	y_mapper.set_magnification (1 + magnification / 2, y);

	follow.redraw = true;
}

/* * * * * * * * * * * * * * *
 *
 *	periodic && scroll
 *
 */

void
gtk_event_view::periodic()
{
	periodcounter++;
	
	if (follow.redraw)
	{
		hide_position();
		follow.x = follow.tick = -1;
		follow.redraw = false;
	}
	update_position();
		
	/* prevent hogging the cpu when dragging the view */
	if ((clicked.button_mask & GDK_BUTTON2_MASK) && (periodcounter % 3))
		return;

	if (clicked.dz)
		delayed_scroll (clicked.dz);

	if (scroll_pending.dx || scroll_pending.dy)
	{
		scroll (scroll_pending.dx, scroll_pending.dy);
		scroll_pending.dx = 
		scroll_pending.dy = 0;
	}
	else 
	{
		process_exposures();
		show_position(); /* not always necessary but the sure way */
	}
}

void
gtk_event_view::delayed_scroll (int dx, int dy)
{
	scroll_pending.dx += dx;
	scroll_pending.dy += dy;
}

bool
gtk_event_view::scroll (int dx, int dy)
{
	if (!(dx | dy))
		return 0;

	x_mapper.scroll (dx);
	y_mapper.scroll (dy);
	
	int adx = abs (dx);
	int ady = abs (dy);
	if (adx > pixels.size.width || ady > pixels.size.height)
	{
		redraw (&pixels.size);
		return 0;
	}

	int x0 = (dx < 0) ? -dx : 0;
	int x1 = (dx < 0) ? 0 : dx;
	
	int y0 = (dy < 0) ? -dy : 0;
	int y1 = (dy < 0) ? 0 : dy;
	
	int width = pixels.size.width - adx;
	int height = pixels.size.height - ady;
	
	/* scroll areas that need an update, too */
	for (int i = 0; i < exposures.count; i++)
		exposures[i].move (dx, dy);

	/* scroll the pixmap first */
	GdkGC * gc = widget->style->fg_gc [GTK_WIDGET_STATE (widget)];

	gdk_window_copy_area
	(
			pixels.map, gc,
			x1, y1,
			pixels.map,
			x0, y0,
			width, height
	);

	if (dx && dy) /* better to redraw everything at once in this case */
		redraw (&pixels.size);
	else if (dx)
	{
		int x = (dx < 0) ? pixels.size.width + dx : 0;
		redraw (x, 0, adx, pixels.size.height);
	}
	else if (dy)
	{
		int y = (dy < 0) ? pixels.size.height + dy : 0;
		redraw (0, y, pixels.size.width, ady);
	}

	/* then scroll the window */
	gdk_window_copy_area
	(
			widget->window, gc,
			x1, y1,
			widget->window,
			x0, y0,
			width, height
	);

	process_exposures();

	/* scroll the position marker, too */
	
	follow.x += dx;
	
	if (dy)
		show_position();

	return true;
}

/* * * * * * * * * * * * * * * *
 *
 *	redraw
 *
 */

void
gtk_event_view::full_redraw()
{
	redraw (&pixels.size);
}

void
gtk_event_view::redraw (int x0, int y0, int width, int height)
{
	c_rectangle clip (x0, y0, width, height);
	redraw (&clip);
}

void
gtk_event_view::redraw (c_rectangle * clip)
{
	exposures.insert_item (*clip);

	pixels.buffer->rect (clip, &back);
	
	draw_stafflines (clip);


	m_tick _tick = x_mapper.x2tick (clip->x0 + clip->width);

	gtk_event_drawer drawer (this, clip);
	

	song.beat.tick.quantize = snap;
	
	if (magnification <= 0)
		drawer.bar_snap =	1 << (-magnification);
	else if (magnification < 2)
		drawer.beat_snap = 1;
	else
		drawer.fraction_snap = 0;
	_dbgprintf (0, "division=%d snap=%d, (%d,%d,%d)\n", song.division, snap, drawer.bar_snap, drawer.beat_snap, drawer.fraction_snap);

	song.seek (0.0, 0, &drawer);
	song.seek (TEN_TO_THE_NINTH, _tick, &drawer);
	song.stop_active_notes();


	draw_loop_markers (clip);

	gdk_draw_rgb_image 
	(
		pixels.map,
		widget->style->fg_gc [GTK_WIDGET_STATE (widget)],
		clip->x0, clip->y0,
		clip->width, clip->height,
		GDK_RGB_DITHER_NORMAL,
		pixels.buffer->pixel_ptr (clip->x0, clip->y0),
		pixels.buffer->rowstride
	);
}

void
gtk_event_view::show_position()
{
	gdk_draw_line 
	(
		widget->window,
		widget->style->black_gc,
		follow.x, 0, follow.x, pixels.size.height
	);
}

void
gtk_event_view::hide_position()
{
	blit (follow.x, 0, 1, pixels.size.height);
}

void
gtk_event_view::update_position()
{
	if (!follow.song)
		return;
		
	m_tick tick = follow.song->get_tick();
	if (follow.tick == tick)
		return;

	int x = x_mapper.tick2x (tick);
	if (follow.x == x)
		return;

	gdk_draw_line 
	(
		widget->window,
		widget->style->black_gc,
		x, 0, x, pixels.size.height
	);
	
	blit (follow.x, 0, 1, pixels.size.height);
	
	follow.tick = tick;
	follow.x = x;
}

int
gtk_event_view::process_exposures()
{
	if (!exposures.count)
		return 0;
	
	for (int i = 0; i < exposures.count; i++)
		blit (&exposures[i]);

	exposures.clear();

	return 1;
}

void
gtk_event_view::blit (int x, int y, int width, int height)
{
	gdk_draw_pixmap	(
			widget->window,
			gc,
			pixels.map,
			x,y,
			x,y, width, height
	);
}


/* * * * * * * * * * * * * * * * * * * *
 *
 *	event && click drawing
 *
 */

void
gtk_event_view::draw_click (m_tick tick, int alpha, c_rectangle * clip)
{
	int x = x_mapper.tick2x (tick);
	int y0, y1;
	
	if (!tick)
		y0 = 0, y1 = pixels.size.height;
	else
		y0 = y_mapper.note2y (127),
		y1 = y_mapper.note2y (0);
	
	c_rectangle r (x, y0, 1, y1 - y0);

	if (!r.clip (*clip))
		return;

	pixels.buffer->arect (&r, &fore, alpha);
}

void
gtk_event_view::draw_loop_markers (c_rectangle * clip)
{
	if (!follow.song)
		return;

	int x[2] = {
		x_mapper.tick2x (follow.song->loop.start),
		x_mapper.tick2x (follow.song->loop.end)
	};
	
	for (int i = 0; i < 2; i++)
	{
		c_rectangle r (x[i], 0, 1, pixels.size.height);

		if (!r.clip (*clip))
			continue;

		pixels.buffer->arect (&r, &loop, 128);
	}
}

void
gtk_event_view::draw_stafflines (c_rectangle * into)
{
	int x0 = x_mapper.tick2x (0);
	int y0, y1;

	if (staff_alpha > 0)
	{
		y0 = y_mapper.note2y (57);
		y1 = y_mapper.note2y (43) + 1;
		c_rectangle bottom_staff (0, y0, pixels.size.width, y1 - y0);

		if (bottom_staff.clip (*into))
			pixels.buffer->arect (&bottom_staff, &staff, staff_alpha);

		y0 = y_mapper.note2y (77);
		y1 = y_mapper.note2y (64) + 1;
		c_rectangle top_staff (0, y0, pixels.size.width, y1 - y0);

		if (top_staff.clip (*into))
			pixels.buffer->arect (&top_staff, &staff, staff_alpha);
	}
	
	for (int note = 0; note < 128; note++)
	{
		int alpha = y_mapper.alpha[note];
		if (!alpha)
			continue;
		
		c_rectangle r (x0, y_mapper.note2y (note), pixels.size.width - x0, 1);
		if (!r.clip (*into))
			continue;

		if (magnification < 2)
		{
			alpha >>= -(magnification - 2);
			if (alpha < 2)
				return;
		}
		pixels.buffer->arect (&r, &fore, alpha);
	}
}

c_color *
gtk_event_view::get_color (int port, int channel)
{
	return &colors[((port << 4) | channel) % colors.count];
}

/*
 *	draws a note
 *
 */

void
gtk_event_view::draw_note (m_event * e, c_color * color, c_rectangle * into)
{
	c_rectangle rectangle;

	if (!event_rectangle (e, &rectangle))
		return;

	int alpha = velocity_to_alpha [e->get_velocity_on()];

	c_rectangle r = rectangle;
	if (r.clip (*into))
		pixels.buffer->arect (&r, color, alpha);

	int note = e->get_note();
	
	int gravity = y_mapper.gravity[note];

	for (int i = note - gravity; i >= 0 && i < 128; i += gravity)
	{
		if (!y_mapper.gravity[i])
			break;

		if (!y_mapper.is_staff[i])
			continue;

		int y = y_mapper.note2y (i);
		
		c_rectangle helper;
		if (i == note)
			helper.set (rectangle.x0 - 2, y, rectangle.width + 4, 1);
		else
			helper.set (rectangle.x0 + 2, y, rectangle.width - 4, 1);

		if (!helper.clip (*into))
			continue;

		pixels.buffer->arect (&helper, &fore, 0x50);
	}
}

bool
gtk_event_view::event_rectangle (m_event * e, c_rectangle * r)
{
	if (e->get_type() != M_NOTE_ON)
		return false;

	m_tick t0 = e->get_tick();
	m_tick t1 = t0 + e->get_note_length();
	
	int x0 = x_mapper.tick2x (t0);
	int x1 = x_mapper.tick2x (t1);


	int note = e->get_note();
	
	int dy = y_mapper.note_height();
	int y = y_mapper.note2y (note) - (dy >> 1);
	
	r->set (x0, y, (x1 - x0), dy);

	return true;
}

/* * * * * * * * * * * * * * * * *
 *
 *	X events
 *
 */

gint
gtk_event_view::timeout_event (gtk_event_view * self)
{
	self->periodic();

	return TRUE;
}


/*
 *	button && motion
 *
 */

gint
gtk_event_view::button_event (GtkWidget * widget, GdkEventButton * event, gtk_event_view * self)
{
	self->clicked.x = (int) event->x;
	self->clicked.y = (int) event->y;
	self->clicked.xtick = (int) event->time;

	if (event->type != GDK_BUTTON_PRESS)
	{
		self->clicked.button_mask &= ~(1 << (7 + event->button));
		return false;
	}
	
	/* GDK_BUTTON1_MASK = 1 << 8 */
	self->clicked.button_mask = 1 << (7 + event->button);

	if (event->button == 1)
		/* add select code */;

	else if (event->button == 2)
		;
		
	else if (event->button == 3)
	{
		if (event->state & GDK_CONTROL_MASK)
			self->set_magnification (self->magnification - 1, (int) event->x, (int) event->y);
		else
			self->set_magnification (self->magnification + 1, (int) event->x, (int) event->y);
		self->full_redraw();
		_dbgprintf (0, "mag!\n");
	}

	else if (event->button == 4)
		self->clicked.dz++;
		
	else if (event->button == 5)
		self->clicked.dz--;
		
	return false;
}

gint
gtk_event_view::motion_event (GtkWidget * widget, GdkEventMotion * event, gtk_event_view * self)
{
	_dbgprintf (0, "(%d/%d) %f %f\n", event->state, self->clicked.button_mask, event->x, event->y);
	
	/* is it the same button */
	
	if (!(event->state & self->clicked.button_mask))
	{
		self->clicked.button_mask = 0;
		return false;
	}
	
	/* drag scroll */

	if (self->clicked.button_mask & GDK_BUTTON2_MASK)
	{
		int dx = (int) event->x - self->clicked.x;
		int dy = (int) event->y - self->clicked.y;
		
		self->delayed_scroll (dx, dy);

		self->clicked.x = (int) event->x;
		self->clicked.y = (int) event->y;
	}
	
	return false;
}

/*
 *	keypress
 *
 */

gint
gtk_event_view::key_event (GtkWidget * widget, GdkEventKey * event, gtk_event_view * self)
{
	_dbgprintf (2, "%04x : %s\n", event->keyval, event->string);
	switch (event->keyval)
	{
		case GDK_Left:
			self->scroll (-10);
			return true;
		case GDK_Right:
			self->scroll (+10);
			return true;
		case GDK_Escape:
			gtk_main_quit();
			return true;
	}
	
	return false;
}

/*
 *	configure
 *
 */

gint
gtk_event_view::configure_event (GtkWidget * widget, GdkEventConfigure * event, gtk_event_view * self)
{
	int	width = event->width, height = event->height;

	bool new_size = 
		width != self->pixels.size.width ||
		height != self->pixels.size.height;
		
	if (!new_size)
		return true;
		
	_dbgprintf (0, "CONFIGURE %dx%d\n", width, height);

	if (!self->gc)
	{
		self->gc = gdk_gc_new (self->widget->window);
		gdk_gc_copy (self->gc, self->widget->style->fg_gc [GTK_WIDGET_STATE (self->widget)]);
	}

	/* recenter y, don't know a better solution */
	self->y_mapper.set_position (60, height / 2);

	self->pixels.size.width = width;
	self->pixels.size.height = height;

	if (self->pixels.map)
		gdk_pixmap_unref (self->pixels.map);

	gtk_widget_push_visual (gdk_rgb_get_visual());
	gtk_widget_push_colormap (gdk_rgb_get_cmap());
	
	self->pixels.map = gdk_pixmap_new (self->widget->window, width, height, -1);

	gtk_widget_pop_visual();
	gtk_widget_pop_colormap();

	if (!self->pixels.buffer)
		self->pixels.buffer = new c_rgb_buffer (width, height);
	else
		self->pixels.buffer->resize (width, height);
	
	self->redraw (0, 0, width, height);

	gdk_window_set_back_pixmap (self->widget->window, self->pixels.map, FALSE);

	return true;
}

/*
 *	expose event
 *
 */

gint
gtk_event_view::expose_event (GtkWidget * widget, GdkEventExpose * event, gtk_event_view * self)
{
	_dbgprintf (0, "eXPOSE %d\n", event->count);
	_dbgprintf (0, "EXPOSE %d %d [%d x %d]\n", 
			event->area.x,
			event->area.y,
			event->area.width,
			event->area.height
	);

	/*
	self->exposures.insert_item (
		c_rectangle (
			event->area.x,
			event->area.y,
			event->area.width,
			event->area.height
		)
	);
	*/
	return true;
}

/*
 *	initialize the colors
 *
 */

void
gtk_event_view::init_colors()
{
	unsigned int defaults[] = 
	{
		0x297994, // output blue
		0xefaa5a, // nude44 flesh
		0x940c18, // toyota gpone red
		0x005184, // chamberlain viper blue
		0xdecb7b, // headlight yellow
		0x7b8242, // le mans dirt green
		0x94967b, // single gray
		0x84a68c, // curbstone green
		0x426939, // helmet green
		0x106db5, // riley blue
		0x6b92ad, // light curb blue
		0x1861de, // electric blue
		0xdecf94, // sand yellow
		0x844539, // gravel sunset brown
		0xf78a52, // neon orange
		0x84a631, // tree green
		0x5aa6ff, // light curb blue II
		0x4275a5, // rubber blue
		0xb55d39, // spanish dark orange
		0x94d7ff, // sky blue
		0xefcb00, // viper yellow
		0x318610, // belt green
		0xefd763, // sunbeam orange
		0x8c2042, // tag magenta
		0xde615a, // closed eye red
		0xbdcbc6, // face grey
		0x6b618c, // night magenta
		0xa5b6c6, // electric light
		0x849294, // street gray
		0xf71818, // cavallino red
		0xfffb5a, // mustang yellow
		0xd67900, // blink orange
		0xff756b, // passion orange
		0xc6dfad, // reflection green
		0x5a557b, // passion magenta
		0x94b652, // pipeline green
		0x2961b5, // blue view blue
		0xfffbb5, // cloud yellow
		0xce1829, // fine red
		0xc6f318, // neon view green
		0xf79e5a, // disc orange
		0x427563, // renown green
		0x18614a, // christmas green
		0x6ba2ce, // south blue
		0x5279de, // starlight blue
		0x7bcfa5, // flow green
		0xbda2a5, // church stone
		0xbd7163, // wide red
		0x9cbec6, // cat's eye green
		0x73aec6, // east blue
		0xf7f7ef, // dream white
		0xb5414a, // dream red
		0xffe77b, // dream yellow
		0xf7c739, // 24 yellow
		0xffffff  // cloud white
	};

	for (unsigned int i = 0; i < sizeof (defaults) / sizeof (defaults[0]); i++)
		colors.add (defaults[i]);
}


/*
 *	gtk_event_selector
 *
 */

gtk_event_selector::gtk_event_selector (gtk_event_view * _view, c_rectangle * _clip)
:
m_connector ("gtk_event_selector")
{
	view = _view;
	clip = *_clip;
}

bool
gtk_event_selector::event_in (m_event * e, m_event_stamp * stamp)
{
	c_rectangle r;
	if (!view->event_rectangle (e, &r))
		return true;
	if (r.clip (clip))
		view->selection.add_event (e);
	return true;
}

void
gtk_event_selector::click_in (m_tick tick, int bar, int beat, int fraction)
{
}

#endif WITH_GTK
