"""Miscellaneous widgets for the Flying Circus"""
from midithing import *
from gtk import *
import GDK, GTK
import pixmaps

class Gtk_XFrame (GtkVBox):
	def __init__ (self, widget, lighting):
		GtkVBox.__init__ (self)
		self.add (widget)
	
class Gtk_Frame (GtkFrame):
	def __init__ (self, widget, lighting):
		GtkFrame.__init__ (self)
		self.add (widget)
		self.set_shadow_type (lighting)

class Input:
	"""A base class for an input widget holding a value and methods to 
		 negotiate the value between the Main Thing and the Input itself."""
	def __init__ (self, pygmy, name):
		self.pygmy = pygmy
		self.name = name
		self.value = None
		self.set_name (name) # so we can single it out in a .gtkrc
		self.set_data ('self', self)

	def emit_signal (self, value = None):
		"""This method triggers a state change method in the Main Thing, which
			 is responsible for setting this method to something more useful."""
		pass
		
	def change (self, value):
		"""Called when a Gtk event led to a state change, this triggers the 
			 appropriate signal method."""
		self.value = value
		self.update_label()
		self.emit_signal (value, self)
		
	def set_value (self, value):
		"""Called from Main to update the Input."""
		if self.value == value: return
		self.value = value
		self.update_label()
		print self.name + ".set_value (%s)" % repr (self)

	def __repr__ (self):
		return repr (self.value)

	def sig_focussed (self, *args):
		self.pygmy.change_focus (self)

class SongPositionInput (GtkHBox, Input):
	"""A widget showing a song position. The position is stored as MIDI
		 ticks and represented as 'bar.beat.fraction'.
		 
		 This has grown quite a beast, my friend."""
	entry_widths = (42, 30, 42)
	entry_height = 22
	def __init__ (self, pygmy, name, song):
		GtkHBox.__init__ (self)
		Input.__init__ (self, pygmy, name)
		self.song = song
		self.insert_handler_ids = []
		self.bar			= self.add_entry (self.entry_widths[0])
		self.beat			= self.add_entry (self.entry_widths[1])
		self.fraction	= self.add_entry (self.entry_widths[2])
		self.next = {
			self.bar: self.beat, 
			self.beat: self.fraction, 
			self.fraction: self.bar
		}

	def add_entry (self, width = 42):
		entry = GtkEntry()
		entry.connect ('enter-notify-event', self.sig_focussed)
		# remember insert-text handler ids so they can be blocked when 
		# updating the entries
		id = entry.connect ('insert-text', self.sig_insert_text)
		self.insert_handler_ids += [(entry, id),]
		entry.connect ('key-press-event', self.sig_key_press)
		entry.connect_after ('button-press-event', self.sig_button_press)
		entry.connect_after ('button-release-event', self.sig_button_release)
		entry['width'] = width
		entry['height'] = self.entry_height
		self.add (entry)
		return entry
		
	def update_label (self, widget = None):
		(bar, beat, fraction) = self.song.beat
		
		for (entry, id) in self.insert_handler_ids:
			entry.signal_handler_block (id)

		if widget in (None, self.bar):
			self.bar.set_text (repr (bar))
		if widget in (None, self.beat):
			self.beat.set_text (repr (beat + 1))
		if widget in (None, self.fraction):
			self.fraction.set_text (repr (fraction))
		
		for (entry, id) in self.insert_handler_ids:
			entry.signal_handler_unblock (id)

	def set_value (self, value):
		if self.song.tick == value: return
		self.song.seek (value)
		self.update_label()

	def __repr__ (self):
		"""Convert tick to 'bar.beat.fraction'."""
		if not len (self.song): return '0.1.0'
		(bar, beat, fraction) = self.song.beat
		return "%d.%d.%d" % (bar, beat + 1, fraction)

	def move (self, delta):
		# normalize delta first.
		delta = delta > 0 or -1
		if self.focussed == self.bar: # measure/bar-wise
			delta *= self.song.numerator * self.song.ticks_per_beat
		elif self.focussed == self.beat: # beat-wise
			delta *= self.song.ticks_per_beat
		
		try:		self.song.tick += delta # will throw if tick goes < 0
		except:	self.song.tick = 0

		self.update_label()

	def update_from_label (self):
		"""Reads the entries' input as strings and tries to move the songpointer
			 to that position."""
		try:		bar = int (self.bar.get_text())
		except:	bar = 0
		try:		beat = int (self.beat.get_text()) - 1
		except:	beat = 0
		try:		fraction = int (self.fraction.get_text())
		except: fraction = 0
		self.song.seek (bar, beat, fraction)

	def sig_focussed (self, widget, event):
		self.focussed = widget
		Input.sig_focussed (self)

	def sig_insert_text (self, entry, text, length, position):
		"""Stop garbage input."""
		for c in text[:length]:
			if c not in "0123456789":
				entry.emit_stop_by_name ('insert-text')

	def sig_key_press (self, entry, event):
		move_focus = 0
		if event.keyval == GDK.Return or event.keyval == GDK.KP_Enter:
			self.update_from_label()
			return 1
		if event.keyval == GDK.Right:
			text = entry.get_text()
			if entry.get_position() == len (text):
				move_focus = +1
			else:
				return 0
		elif event.keyval == GDK.Left or event.keyval == GDK.BackSpace:
			if entry.get_position() == 0:
				move_focus = -1
			else:
				return 0
		elif event.string == '.':
			move_focus = +1
		else:
			return 0
				
		entry.emit_stop_by_name ('key-press-event')

		if not move_focus: 
			return 1
		
		entries = [self.bar, self.beat, self.fraction]
		next = (entries.index (entry) + move_focus) % len (entries)
		entries [next].grab_focus()

		if move_focus > 0:
			entries [next].set_position (0)
			
		return 1

	def sig_button_press (self, entry, event):
		if event.button == 4:
			self.move (+1)
		elif event.button == 5:
			self.move (-1)
		else:
			return 0
		grab_remove (entry)
		return 1

	def sig_button_release (self, entry, event):
		"""For some strange reason, it is necessary to remove the grab from
			 the label widget after buttons 4 and 5 (the wheel) have been used.
			 (GTK+ 1.2.8)"""
		if event.button not in (4, 5):
			return 0
		grab_remove (entry)
		return 1

class OldSongPositionInput (GtkEventBox, Input):
	"""A widget showing a song position. The position is stored as MIDI
		 ticks and represented as 'bar.beat.fraction'."""
	def __init__ (self, pygmy, name, song):
		GtkEventBox.__init__ (self)
		Input.__init__ (self, pygmy, name)
		self.song = song
		self.label = GtkEntry()
		self.add (self.label)
		self.label.connect ('enter-notify-event', self.sig_focussed)
		self.label.connect_after ('button-press-event', self.sig_button_press)
		self.label.connect_after ('button-release-event', self.sig_button_release)
		self.caret = 0 # where the caret stands: 0|1|2 for bar.beat.fraction

	def update_caret_position (self):
		"""Checks where (on bar, beat or fraction) the caret stands."""
		text = self.label.get_text()
		self.caret = 0
		i = self.label.get_position() - 1
		while i > 0:
			if text[i] == '.':
				self.caret += 1
			i -= 1

	def update_label (self):
		text = repr (self)
		self.label.set_text (text)
		c = self.caret
		i = 0
		while c:
			while text[i] != '.':
				i += 1
			i += 1
			c -= 1
		self.label.set_position (i)
		
	def set_value (self, value):
		if self.song.tick == value: return
		self.song.seek (value)
		self.update_label()

	def __repr__ (self):
		"""Convert tick to 'bar.beat.fraction'."""
		if not len (self.song): return '0.1.0'
		(bar, beat, fraction) = self.song.beat
		return "%d.%d.%d" % (bar, beat + 1, fraction)

	def move (self, delta):
		self.update_caret_position()
		# normalize delta first.
		delta = delta > 0 or -1
		if self.caret == 0: # measure (bar)-wise
			delta *= self.song.numerator * self.song.ticks_per_beat
		elif self.caret == 1: # beat-wise
			delta *= self.song.ticks_per_beat
		
		try:
			self.song.tick += delta
		except:
			self.song.tick = 0

		self.update_label()

	def sig_button_press (self, label, event):
		if event.button == 4:
			self.move (+1.0 / 12)
		elif event.button == 5:
			self.move (-1.0 / 12)
		else:
			return 0
		grab_remove (self.label)
		return 1

	def sig_button_release (self, label, event):
		"""For some strange reason, it is necessary to remove the grab from
			 the label widget after buttons 4 and 5 (the wheel) have been used.
			 This pertains to GTK+ 1.2.8."""
		if event.button not in (4, 5):
			return 0
		grab_remove (self.label)
		return 1

class ColorInput (GtkEventBox, Input):
	"""An Input for a color value. The looks: a sunken GtkFrame filled with
		 the color."""
	def __init__ (self, pygmy, name):
		GtkEventBox.__init__ (self)
		frame = GtkFrame()
		self.add (frame)
		frame.set_shadow_type (GTK.SHADOW_IN)
		Input.__init__ (self, pygmy, name)
		try:		self.label = pixmaps.XPM (pygmy.window, name)
		except:	self.label = pixmaps.SmallDot (pygmy.window, 'None')
		frame.add (self.label)
		
	def set_color (self, value):
		return self.set_value (value)

	def get_color_code (self):
		"""Returns the current color value as parseable by X."""
		return '#%02x%02x%02x' % (self.value[0], self.value[1], self.value[2])

	def update_label (self):
		style = self.get_style().copy()
		self.color = self.pygmy.window.get_colormap().alloc (self.get_color_code())
		# and how do you free a color with pygtk?
		for i in range (5):
			style.bg[i] = self.color
		self.set_style (style)
		self.label.set_style (style)
		self.queue_draw()
	

class ToggleInput (GtkFrame, Input):
	"""A two-state toggle input."""
	def __init__ (self, pygmy, name, color = None):
		self.button = GtkToggleButton()
		
		if hasattr (self, 'widget'):
			self.button.add (self.widget)
		elif color: 
			self.button.add (pixmaps.SmallDot (pygmy.window, color))
		else:			
			self.button.add (GtkLabel (name))
		
		self.sig_handler_id = self.button.connect ('toggled', self.sig_toggled)
		#self.button['can_focus'] = 0
		
		#Gtk_Frame.__init__ (self, self.button, GTK.SHADOW_IN)
		GtkFrame.__init__ (self)
		self.set_shadow_type (GTK.SHADOW_NONE)
		self.add (self.button)
		Input.__init__ (self, pygmy, name)

	def update_label (self):
		pass

	def sig_toggled (self, *args):
		self.change (self.button ['active'])

	def set_value (self, value):
		self.button.signal_handler_block (self.sig_handler_id)
		self.button['active'] = value
		self.button.signal_handler_unblock (self.sig_handler_id)


class ButtonInput (ToggleInput):
	"""A trigger rather than an input."""
	def __init__ (self, pygmy, name, color = None):
		ToggleInput.__init__ (self, pygmy, name, color)

	def sig_toggled (self, *args):
		self.set_value (0)
		self.emit_signal()

class XPMToggleInput (ToggleInput):
	"""A toggle with an XPM as its label."""
	def __init__ (self, pygmy, name):
		try:		self.widget = pixmaps.XPM (pygmy.window, name)
		except:	pass
		ToggleInput.__init__ (self, pygmy, name)
	
class XPMButtonInput (ButtonInput):
	"""A button with an XPM as its label."""
	def __init__ (self, pygmy, name):
		try:		self.widget = pixmaps.XPM (pygmy.window, name)
		except:	pass
		ButtonInput.__init__ (self, pygmy, name)
	
class GrayDotButton (ButtonInput, GtkButton):
	def __init__ (self, pygmy, name):
		GtkButton.__init__ (self)
		Input.__init__ (self, pygmy, name)
		self.connect ('clicked', getattr (pygmy, 'sig_' + self.name))
		self.add (pixmaps.SmallDot (pygmy.window, '#b8b8a0'))
	
		


class ParameterInput (GtkVBox, Input):
	"""The generic slider input with a label and a value entry."""
	def __init__ (self, pygmy, name, lower, upper):
		GtkVBox.__init__ (self)
		Input.__init__ (self, pygmy, name)

		self.lower = lower
		self.upper = upper

		box = GtkHBox() # contains the slider and the label
		box.set_border_width (0)

		self.value_label = GtkEntry()
		height = self.value_label['height']
		self.value_label.set_usize (40, height)

		self.pack_start (self.value_label, expand=1, fill=1)
		self.pack_start (box)

		# the slider
		self.scale = GtkVScale() 
		self.scale.set_digits (0)
		self.scale.set_draw_value (0)

		self.adjustment = GtkAdjustment (lower = lower, upper = upper - 1, step_incr = 1.0, page_incr = 2.0, page_size = 0.0)
		self.scale.set_adjustment (self.adjustment)
		
		try:		self.label = pixmaps.XPM (pygmy.window, name)
		except:	self.label = GtkLabel (name)

		box.pack_start (self.scale, expand = 0, fill = 1)
		box.pack_start (self.label, expand = 1, fill = 0)

		self.scale.connect ('enter-notify-event', self.sig_focussed)
		self.handler_id = self.adjustment.connect ('value-changed', self.sig_changed)
	
	def move (self, delta):
		"""Changes the value by delta, which should be between -1 and +1"""
		sign = delta >= 0 or -1
		d = int (sign * (delta * 2) ** 2)
		if not d:
			d = sign

		value = self.value + d

		if value < self.lower: value = self.lower
		if value > self.upper: value = self.upper
		if value == self.value:
			return 0

		self.adjustment.set_value (self.value2adjustment (value))
		return 1
		# and let it trigger.
		
	# adjustment signal
	def sig_changed (self, *args):
		value = self.adjustment2value (self.adjustment.value)
		self.change (value)
	
	def value2adjustment (self, value):
		"""Converts from value to adjustment coordinate space."""
		return self.adjustment.upper - value
	
	def adjustment2value (self, value):
		"""Converts from adjustment to value coordinate space."""
		return int (self.adjustment.upper - value)
		
		
	def set_value (self, value):
		#print self.name + ".set_value (%s)" % repr (self)
		self.adjustment.signal_handler_block (self.handler_id)
		self.adjustment.set_value (self.value2adjustment (value))
		self.adjustment.signal_handler_unblock (self.handler_id)

		self.value = value
		self.update_label()
		
	def update_label (self):
		self.value_label.set_text (repr (self))
		
	
class BankInput (ParameterInput):
	"""Alphabetic bank input"""
	def __repr__ (self):
		return chr (self.value + ord ('A'))


class PitchInput (ParameterInput):
	"""itchin' value input"""
	def __init__ (self, pygmy, name):
		ParameterInput.__init__ (self, pygmy, name, 0.0, 8192)
		self.adjustment.step_increment = 8
		self.adjustment.page_increment = 128

	def value2adjustment (self, value):
		return self.adjustment.upper - 4096 * value - 4096
	
	def adjustment2value (self, value):
		return (self.adjustment.upper - self.adjustment.value - 4096) / 4096.0
		
	def __repr__ (self):
		return '%0.3f' % (self.value)

class NoteInput (ParameterInput):
	"""Edit a note height"""
	def __init__ (self, pygmy, name):
		ParameterInput.__init__ (self, pygmy, name, 0, 128)
		
	def __repr__ (self):
		return note_name (self.value)

class VelocityInput (ParameterInput):
	"""Edit a velocity value"""
	def __init__ (self, pygmy, name):
		ParameterInput.__init__ (self, pygmy, name, 0, 127)
		
	def value2adjustment (self, value):
		return 127 - value
	
	def adjustment2value (self, value):
		return 127 - int (value)



