/*
midi-clock.cc

a m_clock is responsible for periodically calling pulse() in connected
 *
 */

#include <pthread.h>
#include <unistd.h>
#include <malloc.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/poll.h>
#include <asm/msr.h>
#include <fcntl.h>
#include <errno.h>
#include <assert.h>

#include <stdexcept>

#ifdef WITH_RTC
#include <sys/ioctl.h>
#if defined (HAVE_RTC_H)
#include <linux/rtc.h>
#elif defined (HAVE_MC146818RTC_H)
#include <linux/mc146818rtc.h>
#else
#error WITH_RTC defined, but no header file specified.
#endif
#endif

#include "arch-threads.h"
#include "midi-clock.h"

#define DBG_ERRORS		2
#define	DBG_CMDS			0
#define	DBG_TIME			0
#define DBG_LATENCY		1

#define TIMER_RESOLUTION	1024

/*
 *	there should be only one clock running
 *
 */
 
m_clock * m_clock::the_clock = NULL;

m_clock *
m_clock::get()
{
	if (!the_clock)
	{
		#ifdef WITH_RTC
		try 
		{
			the_clock = new m_rtc_clock();
			_dbgprintf (1, "m_clock using rtc.\n");
		}
		catch (std::exception &e) 
		#endif /* let op! */
		{
			the_clock = new m_clock();
			_dbgprintf (1, "m_clock using normal clock.\n");
		}
	}

	C_XINCREF (the_clock);
	return the_clock;
}

void
m_clock::release()
{
	C_XDECREF (the_clock);
}

/*
 *	the clock.
 *
 */

m_clock::m_clock()
:
c_ref_obj ("m_clock"),
clients(),
mutex()
{
	time.resolution = 
	time.hz = TIMER_RESOLUTION;

	cmd.fd[0] = 
	cmd.fd[1] = -1;

	int r = socketpair (AF_LOCAL, SOCK_STREAM, 0, cmd.fd);
	if (r < 0) throw std::runtime_error ("cannot open a socketpair");
	
	running = false;
}

m_clock::~m_clock()
{
	if (a_thread::thread_running())
		do_command (Quit);

	if (cmd.fd[0] != -1)
	{
		close (cmd.fd[0]);
		close (cmd.fd[1]);
	}

	the_clock = NULL;
}

/*
 *	list of connectors maintenance
 *
 */

int
m_clock::add_client (m_connector * _client)
{
	clients.mutex.lock();
	int i = clients.insert (-1);
	clients[i] = _client;
	clients.mutex.unlock();

	return i;
}

bool
m_clock::remove_client (m_connector * _client)
{
	bool done = false;
	
	clients.mutex.lock();
	for (int i = 0; i < clients.count; i++)
	{
		if (clients[i] == _client)
		{
			clients.remove (i);
			done = true;
			break;
		}
	}
	clients.mutex.unlock();

	return done;
}

/*
 *	pulse
 *
 *	advance all io's
 *
 */

void
m_clock::pulse()
{
	/* we have only one m_clock in the whole system. */
	#if DBG_LATENCY
	static double sum = 0, min = TEN_TO_THE_NINTH, max = 0;
	static int samples;
	#endif
	
	double nu = now();
	double dt = nu - time.last;
	_dbgprintf (DBG_TIME, "%.4f (nu) - %.4f = %.4f (time.last)\n", nu, dt, time.last);
	time.last = nu;
	
	#if DBG_LATENCY
	sum += dt;
	if (dt < min)
		min = dt;
	if (dt > max)
		max = dt;
	if (++samples == time.hz)
	{
		_dbgprintf (1, "%.4f < %.4f < %.4f [%1.4f / %d]\n", min, sum / (double) samples, max, sum, samples);
		sum = 0;
		min = TEN_TO_THE_NINTH;
		max = 0;
		samples = 0;
	}
	#endif

	clients.mutex.lock();
	
	for (int i = 0; i < clients.count; i++)
		clients[i]->advance (dt, NULL);
	
	clients.mutex.unlock();
}
				
double
m_clock::now()
{
	struct timeval tv;
	gettimeofday (&tv, NULL);

	return (double) tv.tv_sec + ((double) tv.tv_usec) / TEN_TO_THE_SIXTH;
}

/*
 *	start / stop
 *
 */

bool
m_clock::start()
{
	if (!a_thread::thread_running())
	{
		if (!rt_thread_start (20) && !thread_start())
		{
			return false;
		}
	}

	return !do_command (Start);
}

bool
m_clock::stop()
{
	return !do_command (Stop);
}

/*
 *	set resolution
 *
 */

bool
m_clock::set_resolution (int hz)
{
	time.hz = hz;
	return !do_command (Divide);
}

/*		 
 *	transmit a command to the thread and wait for completion.
 *
 */

int
m_clock::do_command (int code)
{
	if (cmd.fd[0] < 0)
		return EBADF;

	if (!a_thread::thread_running())
		return ENOENT;

	mutex.lock();

	write (cmd.fd[0], &code, sizeof (int));
	read (cmd.fd[0], &code, sizeof (int));
	
	mutex.unlock();

	return code;
}


/*		 
 *	standard m_clock thread
 *
 */

void *
m_clock::thread()
{
	static
	struct pollfd fds = {cmd.fd[1], POLLIN, 0};

	int msec_to_wait = 1000;
	
	while (1)
	{
		_dbgprintf (DBG_TIME, "%d msec\n", msec_to_wait);
		
		fds.revents = 0;
		poll (&fds, 1, msec_to_wait);
		
		if (fds.revents)
		{
			int command;
			read (cmd.fd[1], &command, sizeof (int));
			
			_dbgprintf (2, "cmd %d read\n", msec_to_wait);
			int result = EBADRQC;
			switch (command)
			{
				case Start:
					running = true;
					time.last = now();
					/* fallthrough */

				case Divide:
					msec_to_wait = (int) (1000.0 / (double) time.hz);
					if (!msec_to_wait)
						msec_to_wait = 1;
					result = 0;
					break;
					
				case Stop:
					running = false;
					msec_to_wait = 1000;
					result = 0;
					break;

				case Quit:
					result = 0;
					break;
			}
			write (cmd.fd[1], &result, sizeof (int));
			
			if (command == Quit)
				break;
		}
			
		/* even if poll blew it we get a tad more precision this way. */
		if (running)
			pulse();
	}

	return NULL;
}








/* 
 *	another m_clock, driven by /dev/rtc available on linux
 *
 */

#ifdef WITH_RTC

m_rtc_clock::m_rtc_clock()
:
m_clock()
{
	rtc.fd = -1;

	rtc.hz = time.hz;
	
	try {
		if (seteuid (0))
			throw std::runtime_error ("cannot obtain root privileges.");

		rtc.fd = open ("/dev/rtc", O_RDONLY);
		if (rtc.fd < 0)
			throw std::runtime_error ("cannot open /dev/rtc");
		
		if (ioctl (rtc.fd, RTC_IRQP_SET, rtc.hz) == -1)
			throw std::runtime_error ("cannot set irq resolution on /dev/rtc");
	}
	catch (std::exception &e)
	{
		_dbgprintf (2, "%s\n", e.what());
		seteuid (getuid());
		throw std::runtime_error (e.what());
	}

	seteuid (getuid());
}

m_rtc_clock::~m_rtc_clock()
{
	if (rtc.fd >= 0)
		close (rtc.fd);
}

/*
 *	rt clock thread.
 *
 */

void *
m_rtc_clock::thread()
{
	static 
	struct pollfd fds[] = 
	{
		{rtc.fd, POLLIN, 0},
		{cmd.fd[1], POLLIN, 0}
	};

	_dbgprintf (1, "Clock thread is >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> pid %d\n", getpid());

	while (1)
	{
		bool have_clock = false;
		bool have_cmd = false;
		
		int result = poll (fds, 2, 1000);
		if (result > 0)
		{
			have_clock = fds[0].revents & POLLIN;
			have_cmd = fds[1].revents & POLLIN;
		}
		
		if (have_clock)
		{
			int irqs;
			read (rtc.fd, &irqs, sizeof (unsigned long));

			pulse();
		}
		
		if (have_cmd)
		{
			int command;
			read (cmd.fd[1], &command, sizeof (int));
			
			result = thread_do_command (command);
			write (cmd.fd[1], &result, sizeof (int));
			
			if (command == Quit)
			{
				close (rtc.fd);
				return NULL;
			}
		}
	}

	return NULL;
}

/*
 *	thread-side command switch.
 *
 */

int
m_rtc_clock::thread_do_command (int code)
{
	_dbgprintf (DBG_CMDS, "[thread_do_command] code = %d\n", code);

	switch (code)
	{
		case Divide:
			if (ioctl (rtc.fd, RTC_IRQP_SET, time.hz) == -1)
			{
				_dbgprintf (2, "[thread] setting irq resolution (%d): %s\n", 
					rtc.hz, strerror (errno));
				time.resolution = time.hz = (unsigned) rtc.hz;
				return errno;
			}
			time.resolution = rtc.hz = time.hz;
			
			return 0;
		
		case Start:
			if (running)
				return EBUSY;
				
			if (ioctl (rtc.fd, RTC_PIE_ON, 0) == -1)
			{
				_dbgprintf (2, "[thread] cannot RTC_PIE_ON: %d / %s\n", errno, strerror (errno));
				return errno;
			}
			time.last = now();
			running = true;
			return 0;

		case Stop:
			if (!running)
				return ESRCH;
				
			if (ioctl (rtc.fd, RTC_PIE_OFF, 0) == -1)
			{
				_dbgprintf (2, "[thread] cannot RTC_PIE_OFF: %d / %s\n", errno, strerror (errno));
				return errno;
			}

			running = false;
			return 0;
	}			

	return EBADRQC;
}

#endif  /* WITH_RTC */
