/*

 m_connector.h
 
 event and timing input socket (for ports, filters, songs, parts etc).
 
 one common interface with many features that derived classes only 
 partially use.
 
 */

#ifndef _MIDI_CONNECTOR_
#define _MIDI_CONNECTOR_

#include <sys/fcntl.h>
#include <unistd.h>

#include "midi-event.h"

class m_state;

class m_connector
:
public c_ref_obj
{
	private:
		char * _m_connector_name;
	
	public:
											/* if > 0, return immediately on write_event() -
											 * only used in m_port. this is a counter since
											 * in theory an m_port can be in more than one
											 * m_io object.
											 */
		int								_async;
		
											m_connector (char * name); 
		virtual					 ~m_connector();

		/* just a description for debugging & overview */
		void							set_connector_name (char * fmt, ...);
		char *						get_connector_name();

		/* the current tick of whatever drives this connector */
		virtual void			set_tick (m_tick) {}

		/* feed an event to this connector */
		virtual bool			event_in (m_event *, m_event_stamp * = 0) = 0;
		virtual void			click_in (m_tick tick, int bar, int beat, int fraction) {}

		/* states for poll(2)ing */
		virtual bool			can_read() {return false;}
		virtual bool			can_write() {return false;}

		virtual int				get_read_fd() {return -1;}
		virtual int				get_write_fd() {return -1;}

		/* for actual midi ports, mostly */
		virtual void			reset() {}
		virtual void			parse_input (m_tick, byte port, m_connector *) {}
		virtual m_state *	get_state() {return NULL;}
		
		/* realtime interface */
		virtual void			advance (double dt, m_connector * out) {}
};

class m_event_debugger
:
public m_connector
{
	public:
											m_event_debugger (char * name)
											:
											m_connector (name) {}

		#ifdef DEBUG
		virtual bool			event_in (m_event * e, m_event_stamp * = 0) 
											{
												C_INCREF (e); 
												e->debug (get_connector_name()); 
												C_DECREF (e);
												return false;
											}
		#else
		virtual bool			event_in (m_event * e, m_event_stamp * = 0) {return false;}
		#endif
};

/*	m_in_parser 
 *
 *	thanks:
 *		Paul B-Davis (libmidi++),
 *		Tim Thompson (KeyKit).
 */

class m_in_parser
{
	private:
		m_event *		event;

	public:
								m_in_parser();
		virtual		 ~m_in_parser();
							 
		void				parse (byte data, m_tick, byte port, m_connector *);
	
	private:
		void		start_event (byte data, m_tick, byte port);
		void		dispatch (m_connector *);
		
		int			expect;
		enum {
			Status		= 0,
			ByteOne		= 1,
			ByteTwo		= 2,
			Data			= 3
		};
		
		int			running;
};


/*
 *	records the current midi controller && note state
 *	
 */

class m_state 
: 
public m_connector 
{
	public:
		double			pitch[16];
		signed char	note[16][127];
		byte				ctrl[16][127];
		byte				pres[16][127];
		
											m_state();
									
		void							reset();

		virtual bool			can_write() {return true;}
		virtual bool			event_in (m_event * e, m_event_stamp * = 0);

		virtual m_state *	get_state() {return this;}
};

/* 
 *	m_port
 *
 *	is nothing more than a connector with a state.
 * 
 */

class m_port 
:
public m_connector
{
	public:
		m_state						state;
		m_in_parser				parser;

											m_port (char * name); 
		virtual					 ~m_port();

		virtual bool			event_in (m_event *, m_event_stamp * = 0); 

		virtual int				get_read_fd() {return -1;}
		virtual int				get_write_fd() {return -1;}
		
		virtual void			parse_input (m_tick, byte port, m_connector *);

		virtual void			reset();

		/* additions */
		virtual bool			read (byte *) {return false;}
		virtual int				write_event (m_event * e, m_event_stamp * = 0) {return -1;}
		
		virtual m_state *	get_state() {return &state;}
};

/* m_port for a fd */

class m_fd 
: 
public m_port
{
	public:
		int		fd;
		int		mode;
		double	bytes_per_second;

										/* mode is |= O_NONBLOCK in constructor. */
										m_fd (char * name, int _mode = O_RDWR);

										/* takes over ownership of the fd */
										m_fd (int _fd, int _mode = O_RDWR);
									
		void						setup();
		
									 ~m_fd(); 
		
		virtual bool		can_read() {return get_read_fd() >= 0;}
		virtual bool		can_write() {return get_write_fd() >= 0;}

		virtual int			get_read_fd();
		virtual int			get_write_fd();

		virtual int			write_event (m_event * e, m_event_stamp * = 0);
		virtual bool		read (byte *);

		virtual void		advance (double dt, m_connector * = NULL);

	private:
		c_out_buffer buffer;

		struct {
			byte		running;	/* running status */
			int			pending;	/* bytes pending to be written */
			double	written;	/* approx. bytes in the os queue */
			double	dt;				/* delta time between advances */
			double	writable;	/* how many bytes can be written in this cycle */
		} out;
};

/* a port to pass ascii messages to connected programs like csound ea */
class m_ascii_port
:
public m_port
{
	public:
		int fd;
		int _own_fd;
		
										m_ascii_port (int _fd, bool __own_fd);
									 ~m_ascii_port();
									 
		virtual bool		can_read() {return 0;}
		virtual bool		can_write() {return true;}

		virtual int			get_read_fd() {return -1;}
		virtual int			get_write_fd() {return fd;}

		virtual bool		event_in (m_event * e, m_event_stamp * = 0);
};

#endif  /* _MIDI_CONNECTOR_ */
