/*
 *	midi-event-print.cc
 *
 *	very helpful for python, and debugging ')
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#include "midi-event.h"
#include "midi-part.h"

#ifdef WITH_FLATS
bool m_note_event::flat = true;
#else
bool m_note_event::flat = false;
#endif

char * 
m_note_event::note2string (int n, char * buffer)
{
	static char letters_flat[] = "CDDEEFGGAABB";
	static char letters_sharp[] = "CCDDEFFGGAAB";

	if (n & ~0x7f)
		sprintf (buffer, "%d", n);

	int octave = n / 12;
	int note = n % 12;
	
	char * b = buffer;
	char sign;
	
	if (flat)
	{
		*b++ = letters_flat[note];
		sign = 'b';
	} else {
		*b++ = letters_sharp[note];
		sign = '#';
	}
	
	switch (note)
	{
		case 1:
		case 3:
		case 6:
		case 8:
		case 10:
			*b++ = sign; break;
	}

	sprintf (b, "%d", octave);

	return buffer;
}

/*
 * return the note described by buffer.
 *
 */
 
int
m_note_event::string2note (char * buffer)
{
	/***************************	A		B		C		D		E		F		G */
	static int letter_notes[] = {	9,	11,	0,	2,	4,	5,	7};
	
	char c = toupper (buffer[0]);
	if (!c || c < 'A' || c > 'G')
		return -1;
		
	int note = letter_notes [c - 'A'];
	
	int sign = 0;
	
	char * b = buffer + 1;
	while (*b && !isdigit (*b))
	{
		switch (*b++)
		{
			case '#':
				sign++;
				break;

			case 'b':
				sign--;
				break;

			default:
				return -1;
		}
	}
	
	if (!*b)
		return -1;
		
	char * end = b;
	int octave = strtol (b, &end, 10);
	if (*end) 
		return -1;
	
	return octave * 12 + note + sign;
}


/*
 *	print
 *
 *	python and debugging helper.
 */

void
m_event::print (FILE *fp)
{
	c_char_buffer buffer;
	
	repr (&buffer);
	buffer << (char) 0;
	
	fputs (buffer.items, fp);
}

#ifdef DEBUG
void
m_event::debug (char * prefix, bool cond)
{
	if (!cond) return;
	
	c_char_buffer buffer;
	buffer << '[' << prefix << "] ";
	repr (&buffer);
	buffer << '\n' << (char) 0;

	buffer >> __debug_fp;

	fflush (__debug_fp);
}
#endif

/*
 *	repr()
 *
 *	python helper that prints what eval() would make the same event.
 */

void
m_event::repr (c_char_buffer * buffer)
{
	char note_name[8];
	
	if (get_type() >= M_NOTE_OFF && get_type() < M_SYSEX)
	{
		switch (get_type())
		{
		case M_NOTE_ON:
			*buffer 
				<< M_NOTE_ON_NAME" ('" 
				<< event.note.note2string (get_note(), note_name) 
				<< "' or " << (int) get_note()
				<< ", velocity=" << (int) get_velocity_on()
				<< ", channel=" << (int) get_channel()
				<< ", tick=" << get_tick()
				<< ", length=" << get_note_length()
				<< ", velocity_off=" << (int) get_velocity_off();
			break;
				
		case M_NOTE_OFF:
			*buffer 
				<< M_NOTE_OFF_NAME" ('" 
				<< event.note.note2string (get_note(), note_name) 
				<< "' or " << (int) get_note()
				<< ", velocity=" << (int) get_velocity_on()
				<< ", channel=" << (int) get_channel()
				<< ", tick=" << get_tick();
			break;
		
		case M_NOTE_TOUCH:
			*buffer 
				<< M_NOTE_TOUCH_NAME" ('" 
				<< event.note.note2string (get_note(), note_name) 
				<< "' or " << (int) get_note()
				<< ", velocity=" << (int) get_velocity_on()
				<< ", channel=" << (int) get_channel()
				<< ", tick=" << get_tick();
			break;

		case M_CHANNEL_TOUCH:
			*buffer
				<< M_CHANNEL_TOUCH_NAME" (touch=" << (int) get_param_0()
				<< ", channel=" << (int) get_channel()
				<< ", tick=" << get_tick();
			break;
		
		case M_CONTROL_CHANGE:
			*buffer
				<< M_CONTROL_CHANGE_NAME" (controller=" << (int) get_param_0()
				<< ", value=" << (int) get_param_1()
				<< ", channel=" << (int) get_channel()
				<< ", tick=" << get_tick();
			break;
		
		case M_PROGRAM_CHANGE:
			*buffer
				<< M_PROGRAM_CHANGE_NAME" (program=" << (int) get_param_0()
				<< ", channel=" << (int) get_channel()
				<< ", tick=" << get_tick();
			break;
		
		case M_PITCH_WHEEL:
			*buffer
				<< M_PITCH_WHEEL_NAME" (pitch=" << get_pitch()
				<< ", channel=" << (int) get_channel()
				<< ", tick=" << get_tick();
			break;
		}
		
		*buffer 
			<< ", port=" << (int) get_port()
			<< ")";
			
		return;
	}

	c_char_buffer & data = *get_data_buffer();

	switch (get_type())
	{
		case M_META:
			*buffer << M_META_NAME" (";
			{
				char * s = m_meta_event::types [get_meta_type()].name;
				if (s)
					*buffer << s;
				else
					*buffer << "0x" << get_meta_type();
			}
			*buffer << ", ";
			switch (get_meta_type())
			{
				case M_META_SEQ_NO:
					*buffer << (((int) data[0] << 8) | (int) data[1]) << ", ";
					goto TICK;

				case M_META_PORT:
				case M_META_CHANNEL:
					*buffer << (int) data[0] << ", ";
					goto TICK;
					
				case M_META_EOT:
					goto TICK;

				case M_META_TEMPO:
					*buffer << is_tempo()->bpm() << ", ";
					goto TICK;

				case M_META_SMPTE:
					{
						for (int i = 0; i < 5; i++)
							*buffer << (int) data[i] << ", ";
					}
					goto TICK;

				case M_META_TIME:
					*buffer 
						<< "(" << (int) is_time()->get_numerator() << ", "
						<< (int) is_time()->get_denominator() << "), "
						<< "ticks_per_click=" << (int) is_time()->ticks_per_click() << ", "
						<< "thirtyseconds_per_click=" << (int) is_time()->_32s_per_click() << ", ";
					goto TICK;

				case M_META_KEY:
					*buffer
						<< "key=" << (int) data[0] << ", "
						<< "flavour=" << (int) data[1] << ", ";
					goto TICK;

				default:
					goto DATA;
			} /* end of meta subtype switch */
			break;
		
		case M_SYSEX:
			*buffer 
				<< M_SYSEX_NAME" (";
			goto DATA;

		default:
			*buffer 
				<< M_EVENT_NAME" (0x" << midi_status() << ", ";
			goto DATA;
	}
	
	DATA:
	if (has_data())
	{
		*buffer << "data='";

		if (get_type() == M_SYSEX)
		{
			for (int i = 0; i < data.count; i++)
				*buffer << "\\x" << data[i];
		}
		else
		{
			for (int i = 0; i < data.count; i++)
			{
				if (isprint (data[i]))
					*buffer << (char) data[i];
				else
					*buffer << "\\x" << data[i];
			}
		}
		*buffer << "', ";
	}

	TICK:
	*buffer << "tick=" << get_tick() << ")";
}


