/*
 *	midi-event.cc
 *	
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "midi-event.h"
#include "midi-part.h"

/*
 *	initialize this event to some default values.
 *
 */

void
m_event::init (byte status, byte p0, byte p1)
{
	if (has_data())
		event.sysex.free_data();

	event.generic.tick = 0;

	if (status < M_SYSEX)
	{
		event.generic.type = status & M_TYPE_MASK;
		event.generic.channel = status & M_CHANNEL_MASK;

		if (p0 & 0x80)
		{
			switch (event.generic.type)
			{
			case M_NOTE_OFF:
			case M_NOTE_ON:
			case M_NOTE_TOUCH:
				p0 = DEFAULT_NOTE;
				break;
				
			case M_PITCH_WHEEL:
				/* pitch.set_pitch (0.0); */
				event.pitch.msb = 0x40;
				event.pitch.lsb = 0x00;
				return;
				
			default:
				p0 = 0;
			}
		}
		
		if (event.generic.type == M_NOTE_ON && !p1)
			p1 = DEFAULT_VELOCITY;
			
		if (p1 & 0x80)
		{
			switch (event.generic.type)
			{
			case M_NOTE_ON:
				p1 = DEFAULT_VELOCITY;
				break;
				
			case M_PITCH_WHEEL:
				event.pitch.set_pitch (0.0);
				return;
				
			default:
				p1 = 0;
			}
		}
			
		event.generic.p[0] = p0;
		event.generic.p[1] = p1;

		switch (event.generic.type)
		{
			case M_NOTE_ON:
				event.note.length = DEFAULT_LENGTH;
				break;
		}
	}
	else
	{
		event.generic.type = status;
		event.meta.subtype = p0;

		event.sysex.init();
	}
}

m_event::~m_event()
{
	if (has_data())
		event.sysex.free_data();
}

void
m_event_stamp::apply (m_event * e)
{
	if (this == NULL)
		return;
	if (port >= 0)
		e->set_port (port);
	if (channel >= 0)
		e->set_channel (channel);
}

/*
 *	the precedence of the event type in the row over that in the
 *	column is a 1, its equality a 0 and its 'postcedence' a -1.
 *
 *	it sounds better if program change and note, if they must coexist
 *	at the same point in time, are sent in that order, in case you wondered.
 *
 */
 
int 
m_event::precedence [8][8] =
{
	//NOTE_OFF	NOTE_ON	NOTE_TOUCH	CONTROL	PROGRAM	CHAN_TOUCH	PITCH	META
	{	0,				1,			1,					1,			1,			1,					1,		-1}, //OFF
	{	-1,				0,			1,					-1,			-1,			-1,					-1,		-1}, //ON
	{	-1,				-1,			0,					-1,			-1,			-1,					-1,		-1}, //TCH
	{	-1,				1,			1,					0,			1,			1,					1,		-1}, //CC
	{	-1,				1,			1,					-1,			0,			1,					1,		-1}, //PRG
	{	-1,				1,			1,					-1,			-1,			0,					-1,		-1}, //CTCH
	{	-1,				1,			1,					-1,			-1,			1,					0,		-1}, //PTCH
	{	1,				1,			1,					1,			1,			1,					1,		 0}	 //META
};

int				
m_event::compare (m_event * other)
{
	int dt = event.generic.tick - other->event.generic.tick;
	if (dt) 
		return dt;
		
	dt = precedence [event.generic.type >> 5][other->event.generic.type >> 5]; 
	_dbgprintf (0, "%2x < %2x: %d\n", event.generic.type >> 5, other->event.generic.type >> 5, dt); 
	if (dt)	
		return dt;
	
	dt = event.generic.p[0] - other->event.generic.p[0];
	return dt;
}
		
/*
 *	get the number of bytes this event needs.
 *
 */

int
m_event::get_bytecount (byte status)
{
	static int sizes[] = 
	{
		0,0,0,0, 0,0,0,0,
		2,2,2,2, 1,1,2,0
	};

	return sizes [(status & M_TYPE_MASK) >> 4];
}

/*
 *	midi_status.
 *
 */

byte
m_event::midi_status()
{
	if (event.generic.type >= M_SYSEX)
		return event.generic.type;
	return event.generic.type | event.generic.channel;
}

/*
 *	copy.
 *	
 */

void
m_event::copy (m_event * other)
{
	if (has_data())
		event.sysex.free_data();

	memcpy (&event, &other->event, sizeof (event));
	if (!other->has_data())
		return;

	event.sysex.data_buffer = new c_char_buffer (other->event.sysex.data_buffer);
}

double
m_pitch_event::get_pitch()
{
	return (double) ((msb << 7) | lsb) / (M_PITCH_MAX >> 1) - 1;
}

bool
m_pitch_event::set_pitch (double pitch)
{
	if (pitch < -1.0 || pitch > 1.0)
		return false;
		
	int p = (int) ((1.0 + pitch) * (M_PITCH_MAX >> 1));
	
	lsb = (p & 0x7F);
	msb = (p >> 7);

	return true;
}

/*
 *	m_meta/sysex_event.
 *
 */

void
m_sysex_event::init()
{
	int granularity = sizeof (int);
	int length = 0;
	
	if (type == M_META)
		length = m_meta_event::types[subtype].length;

	if (length <= 0)
		length = 0;
	else
		granularity = length;
		
	if (!data_buffer)
		data_buffer = new c_char_buffer (granularity);

	data_buffer->set_length (length);
	
	switch (subtype)
	{
		case M_META_TEMPO:
			((m_tempo_event *) this)->set_bpm (120);
			break;
		
		case M_META_TIME:
			((m_time_event *) this)->set_numerator_denominator (4, 4);
			((m_time_event *) this)->set_ticks_per_click (DEFAULT_TICKS_PER_CLICK);
			((m_time_event *) this)->set_32s_per_click (DEFAULT_32S_PER_CLICK);
			break;

		default:
			data_buffer->bzero();
			break;
	}
}

void
m_sysex_event::free_data()
{
	if (!data_buffer) return;

	delete data_buffer;
	data_buffer = 0;
}

bool
m_event::set_data (void * data, int length)
{
	if (!is_valid_data_length (length))
		return false;
		
	event.sysex.data_buffer->set (data, length);
	return true;
}

/*
 *	is this a valid data size?
 *
 */

bool
m_event::is_valid_data_length (int _length)
{
	if (_length < 0)
		return false;
	if (get_type() < M_SYSEX)
		return false;
		
	switch (get_type())
	{
		case M_SYSEX:
		case M_EOX:
			return true;

		case M_RT_SONG_POS:
			return _length == 2;

		case M_RT_QUARTER_FRAME:
		case M_RT_SONG_SELECT:
		case M_RT_BUS_SELECT:
			return _length == 1;
			
		case M_META:
			{
				int __length = m_meta_event::types[get_meta_type()].length;
				return __length < 0 || __length == _length;
			}
	}

	return _length == 0;
}

/*
 *	m_tempo_event.
 *
 */

int	
m_tempo_event::usec_per_quarter()
{
	return 
		 (int) data(2) | 
		((int) data(1) << 8) |
		((int) data(0) << 16);
}

void
m_tempo_event::set_usec_per_quarter (int usec)
{
	data(2) =  usec & 0xFF;
	data(1) = (usec >> 8) & 0xFF;
	data(0) = (usec >> 16) & 0xFF;
}

double	
m_tempo_event::bpm()				
{
	return (60 * TEN_TO_THE_SIXTH) / usec_per_quarter();
}

bool		
m_tempo_event::set_bpm (double bpm)	
{
	if (bpm < 4 || bpm > (60 * TEN_TO_THE_SIXTH)) 
		return 0; 
		
	long usec = (long) ((60 * TEN_TO_THE_SIXTH) / bpm);
	set_usec_per_quarter (usec);
	return true;
}

/*
 *	m_time_event.
 *
 */

byte
m_time_event::get_numerator()			
{
	return data(0);
}

byte
m_time_event::get_denominator()		
{
	return 1 << data(1);
}

byte
m_time_event::ticks_per_click()	
{
	return data(2);
}

byte
m_time_event::_32s_per_click()	
{
	return data(3);
}

m_tick
m_time_event::ticks_per_beat()		
{
	return (m_tick) (ticks_per_click() * 4) / (m_tick) get_denominator();
}

m_tick	
m_time_event::ticks_per_bar()		
{
	return (m_tick) get_numerator() * ticks_per_beat();
}

bool
m_time_event::set_numerator (int n)			
{
	if (n < 1 || n > 255) 
		return 0; 
	data(0) = n; 
	return true;
}

bool
m_time_event::set_denominator (int d)			
{
	if (d < 1 || d > 2048) 
		return false; 
		
	int denom = 0;
	
	while (d > 1)
	{
		denom++; 
		if (d > 1 && (d & 1))
			return false;
		d >>= 1;
	}
	
	data(1) = denom; 
	return true;
}

bool
m_time_event::set_numerator_denominator (int n, int d)
{
	return set_numerator (n) && set_denominator (d);
}
	
bool
m_time_event::set_ticks_per_click (int t)	
{
	data(2) = t; 
	return true;
}

bool		
m_time_event::set_32s_per_click (int t)	
{
	data(3) = t; 
	return true;
}

