/*	
 *	midi-event.h
 *	
 *	basic midi classes
 *	
 */

#ifndef _MIDI_CLASSES_
#define _MIDI_CLASSES_

#include <stdio.h>
#include <malloc.h>
#include <string.h>

#include "arch-types.h"
#include "arch-threads.h"
#include "common-list.h"
#include "common-refcount.h"
#include "midi-meta-types.h"

#define	M_TYPE_MASK						0xF0
#define M_CHANNEL_MASK				0x0F

#define	M_NOTE_OFF						0x80
#define M_NOTE_ON							0x90
#define	M_NOTE_TOUCH					0xA0
#define M_CONTROL_CHANGE			0xB0
#define	M_PROGRAM_CHANGE			0xC0
#define	M_CHANNEL_TOUCH				0xD0
#define	M_PITCH_WHEEL					0xE0

#define M_SYSEX								0xF0

#define	M_RT_QUARTER_FRAME		0xF1
#define	M_RT_SONG_POS					0xF2
#define	M_RT_SONG_SELECT			0xF3
#define	M_RT_BUS_SELECT				0xF5
#define	M_RT_TUNE_REQUEST			0xF6

#define M_EOX									0xF7

#define	M_RT_CLOCK						0xF8
#define	M_RT_START						0xFA
#define	M_RT_CONTINUE					0xFB
#define	M_RT_STOP							0xFC
#define	M_RT_ACTIVE_SENSE			0xFE
#define	M_RT_RESET						0xFF

#define	M_META								0xFF

#define	M_NOTE_OFF_NAME				"NoteOff"
#define M_NOTE_ON_NAME				"Note"
#define	M_NOTE_TOUCH_NAME			"NoteTouch"
#define M_CONTROL_CHANGE_NAME	"Controller"
#define	M_PROGRAM_CHANGE_NAME	"Program"
#define	M_CHANNEL_TOUCH_NAME	"ChannelTouch"
#define	M_PITCH_WHEEL_NAME		"Pitch"
#define M_SYSEX_NAME					"SysEx"
#define M_EOX_NAME						"SysExRaw"
#define	M_META_NAME						"Meta"

#define M_TICK_MAX						MAXINT
#define M_PITCH_MAX						0x4000

#define M_BROADCAST_PORT			255

#define DEFAULT_DIVISION				192
#define DEFAULT_NOTE						39
#define DEFAULT_LENGTH					(DEFAULT_DIVISION / 4)
#define DEFAULT_TICKS_PER_CLICK	DEFAULT_DIVISION
#define DEFAULT_32S_PER_CLICK		8

#define DEFAULT_VELOCITY				80

#define M_SONG_NAME				"Song"
#define M_SYNC_NAME				"Sync"
#define M_PART_NAME				"Part"
#define M_PORT_NAME				"Port"
#define M_LOOPBACK_NAME		"LoopbackPorts"
#define M_ASCII_PORT_NAME	"ASCIIPort"
#define M_ALSA_PORT_NAME	"ALSAPort"
#define M_EVENT_NAME			"Event"
#define M_IO_NAME					"IO"
#define M_CLOCK_NAME			"Clock"
#define M_RTC_CLOCK_NAME	"RTClock"

class m_event;
class m_part;

/*
 * accompanies an m_event being passed between connectors
 *
 */

class m_event_stamp
{
	public:
		int				port,
							channel;
		double		time;
		m_part *	part;
		
							m_event_stamp() {reset();}
		void			reset() {port = channel = -1; time = -1.0; part = NULL;}
		void			apply (m_event *);
};

/*
 *	m_generic_event
 *
 */	

class m_generic_event
{
	public:
		m_tick	tick;
		byte		port;

		byte		type;
		byte		channel;

		byte		p[2];

		byte		pad[4]; 
		
		double get_p0() {return (double) p[0] / 127.0;}
		double get_p1() {return (double) p[1] / 127.0;}
};


/*	
 *	m_controller_event
 *	
 */

typedef struct
{
	int controller;
	bool onoff;
	char * name;
} m_controller_type;

class m_controller_event
{
	public:
		m_tick	tick;
		byte		port;
		
		byte		type;
		byte		channel;

		byte		controller; 
		byte		value; 
		
		byte		pad[4]; 
						
		static m_controller_type types[];
};

/*	
 *	m_note_event
 *	
 *	a note on the notes :) we store a note and its length,
 *	not a note-on and a note-off like midi + midifiles do.
 *	
 */

class m_note_event
{
	public:
		m_tick	tick;
		byte		port;
		
		byte		type;
		byte		channel;

		byte		note;
		byte		velocity_on;
		byte		velocity_off;
		
		byte		pad[3];
			
		m_tick	length;

		/* string <-> note height translation */
		static char * note2string (int n, char * buffer);
		static int string2note (char * buffer);
		static bool flat;

		double get_velocity_on() {return (double) velocity_on / 127.0;}
		double get_velocity_off() {return (double) velocity_off / 127.0;}
};

/*	
 *	m_pitch_event
 *	
 */

class m_pitch_event
{
	public:
		m_tick	tick;
		byte		port;
		
		byte		type;
		byte		channel;

		byte		lsb; 
		byte		msb; 
		
		byte		pad[4]; 

	public:
		double	get_pitch();
		bool		set_pitch (double pitch);
};

/*	
 *	m_sysex_event. 
 *
 *	m_meta_event is derived from m_sysex_event.
 */
		
typedef struct
{
	int number;
	char * name;
	int length;
} m_meta_type;

class m_sysex_event
{
	public:
		m_tick	tick;
		byte		port;
		
		byte		type;
		byte		subtype;

		byte		pad[2]; 

		c_char_buffer *	data_buffer;
		
	public:
		void		init();
		void		free_data();
		
		byte &	data (int i) {return (*data_buffer)[i];}
};

class m_meta_event : public m_sysex_event
{
	public:
		static	m_meta_type types[];
};

/*
 *	m_tempo_event.
 *	
 */

class m_tempo_event : public m_meta_event
{ 
	public:
		int			usec_per_quarter();
		void		set_usec_per_quarter (int usec);
		double	bpm();
		bool		set_bpm (double bpm);
};

/*
 *	m_time_event.
 *
 */

class m_time_event : public m_meta_event
{ 
	public:
		byte		get_numerator();
		byte		get_denominator();
		byte		ticks_per_click();
		byte		_32s_per_click();
		
		m_tick	ticks_per_beat();
		m_tick	ticks_per_bar();
		
		bool		set_numerator (int n);
		bool		set_denominator (int d);
		bool		set_numerator_denominator (int n, int d);
		bool		set_ticks_per_click (int t);
		bool		set_32s_per_click (int t);
};

/*
 *	m_event
 *
 *	a union of the events along with some helper functions.
 *
 */

class m_event
:
public c_ref_obj
{
	private:
		union 
		{
			m_generic_event			generic;
		
			m_note_event 				note;
			m_controller_event	controller;
			m_pitch_event				pitch;

			m_sysex_event 			sysex;
			m_meta_event 				meta;
			m_tempo_event				tempo;
			m_time_event				time;
		}	
		event;

	public:
							m_event()	
								: c_ref_obj()
								{zero();}
						
							m_event (byte status, byte p0 = 255, byte p1 = 255) 
								: c_ref_obj()
								{zero(); init (status, p0, p1);}
							
							m_event (m_event * other)	
								: c_ref_obj()
								{zero(); copy (other);}
						
							m_event (m_event * other, m_tick tick)	
								: c_ref_obj()
								{zero(); copy (other); set_tick (tick);}
						
						 ~m_event();

	private:
		void			zero()
								{memset (&event, 0, sizeof (event));}

	public:
		void			init (byte status, byte p0 = 0, byte p1 = 0);

		void			copy (m_event *);
		
		byte			midi_status();

		/* get_* / set_* */
		byte			get_type()					{return event.generic.type;}
		void			set_type (byte type)			
																	{event.generic.type = type;}

		byte			get_channel()				{return event.generic.channel;}
		void			set_channel (byte channel)			
																	{event.generic.channel = channel;}

		m_tick		get_tick()					{return event.generic.tick;}
		void			set_tick (m_tick tick)			
																	{event.generic.tick = tick;}

		byte			get_port()					{return event.generic.port;}
		void			set_port (byte port)			
																	{event.generic.port = port;}

		byte			get_note()					{return event.note.note;}
		void			set_note (byte note)			
																	{event.note.note = note;}

		m_tick		get_note_length()		{return event.note.length;}
		void			set_note_length (m_tick length)			
																	{event.note.length = length;}

		byte			get_velocity_on()		{return event.note.velocity_on;}
		void			set_velocity_on (byte velocity_on)			
																	{event.note.velocity_on = velocity_on;}

		byte			get_velocity_off()	{return event.note.velocity_on;}
		void			set_velocity_off (byte velocity_off)			
																	{event.note.velocity_off = velocity_off;}

		byte			get_param_0()				{return event.generic.p[0];}
		void			set_param_0 (byte p0)
																	{event.generic.p[0] = p0;}

		byte			get_param_1()				{return event.generic.p[1];}
		void			set_param_1 (byte p1)
																	{event.generic.p[1] = p1;}

		byte			get_meta_type()			{return event.meta.subtype;}
		void			set_meta_type (byte meta)
																	{event.meta.subtype = meta;}

		double		get_pitch()					{return event.pitch.get_pitch();}
		void			set_pitch (double pitch)
																	{event.pitch.set_pitch (pitch);}

		bool			is_corresponding_note (m_event * other)
																	{
																		return event.generic.type == M_NOTE_ON &&
																			event.note.note == other->event.note.note &&
																			event.generic.channel == other->event.generic.channel;
																	}

		/* ********** */

		bool			is_valid_data_length (int length);
		c_char_buffer *	
							get_data_buffer()		{return has_data() ? event.sysex.data_buffer : NULL;}
		bool			set_data (void * data, int length);
			
		/* ********* */

		void			repr (c_char_buffer *);
		void			print (FILE * fp);

		char **		get_value_names();
		int				get_value (char * name);
		char *		set_value (char * name, int value);
	
		static int  precedence[8][8];

		int				compare (m_event * other);
		
		static int			get_bytecount (byte status);
		int							bytecount() 	{return get_bytecount (event.generic.type);}

		bool						has_data()		{return event.generic.type >= M_SYSEX && event.sysex.data_buffer;}

		bool						is_note_on() 	{return event.generic.type == M_NOTE_ON;}
		bool						is_note_off()	{return event.generic.type == M_NOTE_OFF;}
		bool						is_sysex()		{return event.generic.type >= M_SYSEX && event.generic.type < M_META;}
		bool						is_meta()			{return event.generic.type == M_META;}

		bool						is_meta (byte type)	{return is_meta() && event.meta.subtype == type ? &event.meta : NULL;}
		m_time_event *	is_time()						{return is_meta (M_META_TIME) ? &event.time : NULL;}
		m_tempo_event *	is_tempo()					{return is_meta (M_META_TEMPO) ? &event.tempo : NULL;}

		#ifdef DEBUG
		void						debug (char * label, bool cond = true);
		#endif
};

#endif
