/*
 * midi-file.cc
 *	
 */
 
#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <assert.h>
#include <errno.h>

#include "midi-file.h"

#define DBG_READ	0
#define DBG_WRITE	0

m_file::m_file()
: 
c_ref_obj("m_file"), 
parts()
{
	division = DEFAULT_DIVISION;
	name = NULL;
	
	file.fp = NULL;
	*file.error = 0;
}

m_file::m_file (char * _ref_name)
: 
c_ref_obj(_ref_name) 
{
	division = DEFAULT_DIVISION;
	name = NULL;
	
	file.fp = NULL;
	*file.error = 0;
}

/*
 *	 clean up a m_part_list
 *	
 */

m_file::~m_file()
{
	clear();
	
	if (name)
		free (name);
}

/*
 *	part list management
 *
 */

void
m_file::clear()
{
	for (int i = 0; i < parts.count; i++)
		C_DECREF (parts[i]);

	parts.clear();
}

bool
m_file::remove_part (m_part * part)
{
	if (!parts.remove (part))
		return false;
	
	C_DECREF (part);

	return true;
}

int
m_file::add_part (m_part * part)
{
	C_INCREF (part);
	
	parts.insert (part);

	return parts.count - 1;
}

void
m_file::reset_error()
{
	*file.error = 0;
}

/*
 *	 set_name
 *	
 */
 
void
m_file::set_name (char * _name)
{
	if (name)
		free (name);
	if (_name)
		name = strdup (_name);
	else
		name = NULL;
}

int
m_file::read (char * _name)
{
	clear();

	set_name (_name);

	reset_error();

	file.fp = fopen (name, "r");
	if (!file.fp)
	{
		strcpy (file.error, strerror (errno));
		return 0;
	}

	m_fileheader header;

	fread (&header, 1, 4, file.fp);
	if (strncmp (header.MThd, "MThd", 4))
	{
		strcpy (file.error, "MThd signature not found.");
		return 0;
	}

	header.length		= read_dword();
	header.format		= read_word();
	header.tracks		= read_word();
	
	division = read_word();

	/* skip additional data in the header chunk */
	fseek (file.fp, header.length - 6, SEEK_CUR);

	if (header.tracks == 0)
	{
		_dbgprintf (DBG_READ, " %s claims to have no tracks.\n", name);
		header.tracks = (word) -1;
	}

	for (int i = 0; !*file.error && i < header.tracks; i++)
	{
		_dbgprintf (DBG_READ, ">>> READING TRACK %d\n", i);
		file.tick = 0;

		fread (&header.MThd, 1, 4, file.fp);
		if (strncmp (header.MThd, "MTrk", 4))
			sprintf 
			(
				file.error, 
				"MTrk signature not found at %s offset 0x%08lx",
				name,
				ftell (file.fp)
			);

		file.remaining = read_dword();

		_dbgprintf (DBG_READ, ">>>>>>>>> file.remaining = %ld\n", file.remaining);

		m_part * part = new m_part();
		
		part->division = division;

		add_part (part);

		int i = 0;
		while (!*file.error && file.remaining && read_event (part))
			if (!(i++ % 100))
				_dbgprintf (0, "%-5d events read\n", i);
			
		part->initialize (division);
	}
	
	fclose (file.fp);

	return parts.count;
}

/*
 *	read an event into this part
 *
 */

bool
m_file::read_event (m_part * part)
{
	static byte	status = 0;

	dword delta		= read_varlen();
	byte rawevent	= read_byte();
	
	/* running status */
	if (rawevent < 0x80) 
	{
		ungetc (rawevent, file.fp);
		file.remaining++;
		rawevent = status;
	} 
	else 
	{
		status = rawevent;
	}

	file.tick += delta;

	m_event * e = new m_event;

	C_INCREF (e);

	void * data;
	int length;
	byte meta;
	
	switch (rawevent) 
	{
		case M_META:
			meta = read_byte();
			e->init (rawevent, meta);
		case M_EOX:
		case M_SYSEX:
			length = read_varlen();
			data = malloc (length);
			fread (data, 1, length, file.fp);
			e->set_data (data, length);
			free (data);
			file.remaining -= length;
			break;
		
		default:
			e->init (rawevent);
			length = e->bytecount();
			if (length > 0) /* always true. */
				e->set_param_0 (read_byte());
			if (length > 1)
				e->set_param_1 (read_byte());
			break;
	}

	/*
	 *	catch note_on which is note_off
	 *
	 */

	if (e->is_note_on())
	{
		if (!e->get_velocity_on())
	 		e->set_type (M_NOTE_OFF);
		e->set_velocity_off (0);
		e->set_note_length (2);
	}

	e->set_tick (file.tick);
	
	if (!e->is_note_off())
		part->add_event (e);
	else
		part->parse_note_off (e);
	
	C_DECREF (e);

	return true;
}

/*
 *	 read a varlen
 *	
 */

long
m_file::read_varlen()
{
   dword value;
   byte c;

   if ((value = read_byte()) & 0x80) 
	 {
      value &= 0x7f;
      do {
				value = (value << 7) | ((c = read_byte()) & 0x7f);
      } while (c & 0x80);
   }

   return value;
}

/*
 *	 read a 4-byte dword
 *	
 */

dword
m_file::read_dword()
{
	return (dword) 
	(
		(read_byte() << 24) | 
		(read_byte() << 16) | 
		(read_byte() << 8) | 
		(read_byte())
	);
}

/*
 *	 read a 2-byte word
 *	
 */

word
m_file::read_word()
{
	return (word) ((read_byte() << 8) | (read_byte()));
}

byte
m_file::read_byte()
{
	if (*file.error)
		return 0;
		
	if (feof (file.fp))
		strcpy (file.error, ("premature end-of-file.fp"));
	else if (ferror (file.fp))
		strcpy (file.error, strerror (ferror (file.fp)));
	else
	{
		file.remaining--;
		byte b = fgetc (file.fp);
		_dbgprintf (DBG_READ, ":%02x", b);
		return b;
	}

	return 0;
}

/*
 *	 write a midi file.fp
 *	
 */

int
m_file::write (char * _name)
{
	if (_name)
		set_name (_name);

	if (!name)
		return 0;
		
	file.fp = fopen (name, "w");
	
	if (!file.fp)
		return 0;

	reset_error();

	file.remaining = 0;
	
	write_data ((byte *) "MThd", 4);
	write_dword	(6);
	write_word (parts.count > 1 ? 1 : 0);
	write_word (parts.count);
	write_word (division);

	for (int i = 0; i < parts.count; i++) 
	{
		_dbgprintf (DBG_WRITE, ">> WRITING TRACK %d\n", i);

		write_data	((byte *) "MTrk", 4);

		long start_of_chunk = ftell (file.fp);
		write_dword	(0);
		file.remaining = 0;
		
		file.tick = 0;
		m_active_notes playing;
		
		m_event_cursor * c = part(i)->new_cursor (0);

		while (c->event)
		{
			m_event * e = c->event;
			
			write_note_offs 
				(&playing, e->get_tick());

			write_event	
				(&playing, e);

			c->step_fwd();
		}

		part(i)->del_cursor (c);

		write_note_offs (&playing, -1);

		m_event eot (M_META, M_META_EOT);
		eot.set_tick (file.tick);
		write_event (NULL, &eot);

		/*
		 * go back and fix the offset
		 *
		 */

		long end_of_chunk = ftell (file.fp);
		fseek (file.fp, start_of_chunk, SEEK_SET);
		write_dword	(file.remaining);
		fseek (file.fp, end_of_chunk, SEEK_SET);
		
		_dbgprintf (DBG_WRITE, ">>>>>>>>> file.remaining = %ld, end - start = %ld\n",
			file.remaining, end_of_chunk - start_of_chunk);
	}

	return true;
}

/*
 *	 write one event
 *	
 */

bool
m_file::write_event (m_active_notes * playing, m_event * e)
{
	#if defined (DEBUG) && DBG_WRITE > 0
	_dbgprintf (1, "\n>write> (%d) ", file.tick);
	e->print (__debug_fp);
	fputc ('\n', __debug_fp);
	#endif

	write_varlen (e->get_tick() - file.tick);

	file.tick = e->get_tick();

	write_byte (e->midi_status());
	
	int length;
	switch (e->get_type()) 
	{
		case M_NOTE_ON:
			playing->event_in (e);
		case M_NOTE_OFF:
		case M_NOTE_TOUCH:
		case M_CONTROL_CHANGE:
		case M_PITCH_WHEEL:
		case M_PROGRAM_CHANGE:
		case M_CHANNEL_TOUCH:
			length = e->bytecount();
			if (length > 0)
				write_byte (e->get_param_0());
			if (length > 1)
				write_byte (e->get_param_1());
			break;

		case M_META:
			/* meta */
			write_byte (e->get_meta_type());

		case M_EOX:
		case M_SYSEX:
			write_varlen (e->get_data_buffer()->count);
			write_data (e->get_data_buffer()->items, e->get_data_buffer()->count);
			break;
	}

	return !*file.error;
}

/*
 *	 write note off's for playing events ending before next
 *
 */

bool
m_file::write_note_offs (m_active_notes * playing, m_tick next)
{
	for (int j = 0; j < playing->events.count; ) 
	{
		m_event * e = playing->event(j);
			
		m_tick pos = e->get_tick();
			
		if (next < 0 || pos <= next) 
		{
			e->set_type (M_NOTE_OFF);

			write_event (NULL, e);

			playing->remove_event (e);
		} 
		else 
		{
			j++;
		}
	}

	return !*file.error;
}

bool
m_file::write_varlen (unsigned long value)
{
	unsigned long buffer;

	buffer = value & 0x7F;
	while ((value >>= 7) > 0) 
	{
		buffer <<= 8;
		buffer |= 0x80;
		buffer += (value & 0x7F);
	}

	while (1) 
	{
		write_byte (buffer);
		if (buffer & 0x80)
			buffer >>= 8;
		else
			break;
	}

	return true;
}

bool
m_file::write_dword (dword value)
{
	return 
		write_byte ((byte) (value >> 24)) ||
		write_byte ((byte) (value >> 16)) ||
		write_byte ((byte) (value >> 8)) ||
		write_byte ((byte) value);
}

bool
m_file::write_word (word value)
{
	return 
		write_byte ((byte) (value >> 8)) ||
		write_byte ((byte) value);
}

bool
m_file::write_byte (byte value)
{
	if (ferror (file.fp))
		strcpy (file.error, strerror (ferror (file.fp)));

	if (*file.error)
		return 0;
	
	file.remaining++;

	fputc (value, file.fp);

	return *file.error;
}

bool
m_file::write_data (void * pdata, long length)
{
	file.remaining	+= length;
	return (bool) fwrite (pdata, length, 1, file.fp);
}

/*
 *	repr.
 *
 */

void
m_file::repr (c_char_buffer * buffer, char limiter)
{
	*buffer << M_SONG_NAME" (" << limiter;

	for (int i = 0; i < parts.count; i++)
	{
		m_part * p = part(i)->lock_read();

		p->repr (buffer, limiter);

		*buffer << "," << limiter;
			
		p->unlock_read();
	}

	if (name)
		*buffer << "name='" << name << "'," << limiter;

	*buffer << "division=" << division << limiter << ")";
}
