/*
 *	midi-in-parser.cc
 *
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>

#include <stdexcept>

#include "midi-io.h"

#define DBG_INBYTES			0
#define DBG_EVENTS_IN		0


/*	
 *	m_in_parser
 *
 */

m_in_parser::m_in_parser()
{
	event = NULL;
	
	running = 0;
	expect = Status;
}

m_in_parser::~m_in_parser()
{
	C_XDECREF (event);
}

/*
 *	allocate the next event and set the parser state
 *
 */

void
m_in_parser::start_event (byte data, m_tick tick, byte port)
{
	event = new m_event (data, 0, 0);
	
	C_INCREF (event);
	
	event->set_port (port);
	event->set_tick (tick);

	if (data >= M_SYSEX)
		expect = Data;
	else
	{
		expect = ByteOne;
		running = data;
	}
}

/*
 *	dispatch this event and set the state
 *
 */
 
void
m_in_parser::dispatch (m_connector * out)
{
	if (!event)
		return;

	if (out)
		out->event_in (event);
		
	C_DECREF (event);

	event = NULL;

	expect = Status;
}

/*
 *	parse 1 byte.
 *
 *	the parser is based on code that i'd like to thank for:
 *
 *  	Paul Davis (libmidi++),
 *  	Tim Thompson (KeyKit).
 *
 */
 
void
m_in_parser::parse (byte data, m_tick tick, byte port, m_connector * out)
{
	if (!out)
		return;
		
	/* drop active sensing */
	if (data == M_RT_ACTIVE_SENSE)
		return;
	
	#if defined (DEBUG) && DBG_INBYTES
		#if (DBG_INBYTES > 1)
		static char * expect_names[] = {"Status", "ByteOne", "ByteTwo", "Data"};
		_dbgprintf (1, "[parse] port %d byte %2x expect %s\n", port, data, expect_names[expect]);
		#else
		_dbgprintf (1, "[%2x]", data);
		#endif
	#endif
	
	bool is_status = (data >= M_NOTE_OFF);

	if (is_status) 
	{
		/* real time message */
		if (data >= M_RT_CLOCK)
		{
			m_event * rt = new m_event (data, 0, 0);
			rt->set_tick (tick);
			
			C_INCREF (rt);
			out->event_in (rt);
			C_DECREF (rt);
			return;
		}

		/* sysex IS terminated by new status */
		if (expect == Data)
		{
			if (event->get_type() == M_SYSEX && data == M_EOX)
			{
				*(event->get_data_buffer()) << (char) data;
				dispatch (out);
				return;
			}

			dispatch (out);
		}

		/* start a new event */
		start_event (data, tick, port);
		return;
	}

	/* definitely a data byte. */
	switch (expect)
	{
		case Status:
			if (!running)
				return;

			start_event (running, tick, port);
			
			if (expect != ByteOne)
				BREAKPOINT;
				
			/* since we're in running status, we expect a data byte now */
			
		case ByteOne:
			event->set_param_0 (data);
			
			if (event->bytecount() > 1)
				expect = ByteTwo;
			else 
				dispatch (out);
			return;

		case ByteTwo:
			event->set_param_1 (data);
			
			/* catch note-on with 0 velocity */
			if (event->get_type() == M_NOTE_ON && !data)
				event->set_type (M_NOTE_OFF);
				
			dispatch (out);
			return;

		case Data:
			{
				c_char_buffer & buffer = *event->get_data_buffer();
				buffer << (char) data;
		
				if (event->get_type() > M_SYSEX)
					if (event->is_valid_data_length (buffer.count))
						dispatch (out);
			}
			return;
		
		default:
			BREAKPOINT;
	}
}
