/*
 *	midi-io.cc
 *
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/poll.h>
#include <unistd.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>

#include <stdexcept>

#include "midi-io.h"
#include "midi-clock.h"

#define DBG_THREAD			0
#define DBG_PORTS				0
#define DBG_INBYTES			0
#define DBG_EVENTS_OUT	1
#define DBG_EVENTS_IN		0

/*	
 *	m_io
 *
 */

m_io::m_io (m_connector * _out)
:
a_thread(),
m_connector ("m_io"),
sync(),
ports(),
mutex()
{
	sync.disable_refcounting();

	tick = 0;

	notification.fd[0] =
	notification.fd[1] = -1;

	set_output (_out);

	clock = m_clock::get();
	if (!clock)	
		throw runtime_error ("Cannot obtain a clock.");
	
	clock->add_client (this);
	clock->start();

	int r = socketpair (AF_LOCAL, SOCK_STREAM, 0, notification.fd);
	if (r < 0) 
		throw runtime_error ("Cannot obtain thread communication sockets.");
	
	a_thread::thread_start();
}

m_io::~m_io()
{
	if (clock)
	{
		clock->remove_client (this);
		m_clock::release();
	}

	if (a_thread::thread_running())
	{
		_dbgprintf (DBG_THREAD, "~m_io\n");

		a_thread::_quit_requested = true;

		int code = Quit;
		write (notification.fd[0], &code, sizeof (int));
		read (notification.fd[0], &code, sizeof (int));
		_dbgprintf (DBG_THREAD, "response read.\n");
	}

	if (notification.fd[0] != -1)
	{
		close (notification.fd[0]);
		close (notification.fd[1]);
	}
}

/*
 *	set_tick
 *
 *	stamp all incoming events with this tick.
 */
 
void
m_io::set_tick (m_tick _tick)
{
	tick = _tick;

	mutex.lock();
	
	for (int i = 0; i < ports.count; i++)
		port(i)->set_tick (tick);

	mutex.unlock();
}

/*
 *	set_output
 *
 */

void
m_io::set_output (m_connector * _out)
{
	mutex.lock();
	
	sync.set_through (_out);

	mutex.unlock();
}

/*
 *	adding / removing a port requires notifying the 
 *	polling thread, which then rebuilds his pollfd array.
 */

int
m_io::add_port (m_connector * _port)
{
	/* operate this port asynchronously from now on */
	_port->_async++;

	if (!a_thread::thread_running())
		return ports.insert (_port);

	mutex.lock();
	
	int code = PortAdded;
	write (notification.fd[0], &code, sizeof (int));

	int i = ports.insert (_port);

	mutex.unlock();

	return i;
}

void
m_io::clear()
{
	/* synchronous operation for these ports */
	for (int i = 0; i < ports.count; i++)
		port(i)->_async--;
		
	if (!a_thread::thread_running())
	{
		ports.clear();
		return;
	}

	mutex.lock();

	int code = PortRemoved;
	write (notification.fd[0], &code, sizeof (int));

	ports.clear();

	mutex.unlock();
}

bool
m_io::remove_port (m_connector * _port)
{
	/* operate this port synchronously from now on */
	_port->_async--;
	
	if (!a_thread::thread_running())
		return ports.remove (_port);

	mutex.lock();

	int code = PortRemoved;
	write (notification.fd[0], &code, sizeof (int));

	bool done = ports.remove (_port);

	mutex.unlock();

	return done;
}

bool
m_io::replace_port (int i, m_connector * _port)
{
	if (i < 0 || i >= ports.count)
		return false;
	
	/* toggle asynchronous operation for these ports */
	port(i)->_async--;
	_port->_async++;
	
	if (!a_thread::thread_running())
		return ports.replace (i, _port);

	mutex.lock();

	int code = PortReplaced;
	write (notification.fd[0], &code, sizeof (int));

	bool done = ports.replace (i, _port);

	mutex.unlock();

	return done;
}

/*
 *	event_in
 *
 *	this event actually goes out through one of the ports connected.
 */
 
bool
m_io::event_in (m_event * e, m_event_stamp * stamp)
{
	if (!ports.count)
		return false;

	mutex.lock();
	#if defined (DEBUG) && DBG_EVENTS_OUT > 0
	e->debug ("m_io::event_in", e->get_type() != 0xf8);
	#endif

	int _port = (stamp && stamp->port >= 0) ? stamp->port : e->get_port();

	if (_port != M_BROADCAST_PORT)
		port (_port % ports.count)->event_in (e, stamp);
	else
	{
		for (int i = 0; i < ports.count; i++)
			port(i)->event_in (e);
	}
	mutex.unlock();

	return true;
}

/*
 *	advance
 *
 *	relay this delta time to all ports
 *
 */
 
void
m_io::advance (double dt, m_connector * out)
{
	if (!ports.count)
		return;

	mutex.lock();

	/* the sync advances its song[s];
	 * this is called first to ensure proper timing of clock messages
	 */
	 
	sync.advance (dt, this);
	
	for (int i = 0; i < ports.count; i++)
		port(i)->advance (dt, NULL);
	
	mutex.unlock();
}

/*
 *	thread
 *
 *	polls the ports.
 */
 
void *
m_io::thread()
{
	/* rebuilding the pollfd array might be easier to do in 
	   add/remove but they have to stay constant during poll()
	 */
	
	bool rebuild = true;
	
	c_list<struct pollfd> poll_list;
	c_ref_list port_list;
	
	while (1)
	{
		mutex.lock();
		if (rebuild)
		{
			int n = 0;
			
			port_list.clear();
			poll_list.clear();
			
			/* first fd is myself, but it's no good idea to
				 hold a reference to myself.
			 */
			port_list.insert (NULL);
			
			poll_list.insert();
			poll_list[n].fd = notification.fd[1];
			poll_list[n].events = POLLIN;
			
			n++;
			
			for (int i = 0; i < ports.count; i++)
			{
				int fd = port(i)->get_read_fd();
				if (fd > 0)
				{
					port_list.insert (port(i));
					
					poll_list.insert();
			
					poll_list[n].fd = fd;
					poll_list[n].events = POLLIN;
					n++;
				}
			}

			_dbgprintf (DBG_PORTS, "[in::poll] got %d port/s\n", poll_list.count);
			rebuild = false;
		}
		mutex.unlock();
		
		/* now wait */
		int r = poll (&poll_list[0], poll_list.count, -1);
		if (r == -1)
		{
			_dbgprintf (2, "poll: %s\n", strerror (errno));
			continue;
		}
		
		_dbgprintf (DBG_THREAD, "thread notified of ");
		
		mutex.lock();
		for (int n = 0; n < poll_list.count; n++)
		{
			if (poll_list[n].revents & POLLIN)
			{
				if (port_list[n] == NULL)
				{
					int code;
					read (notification.fd[1], &code, sizeof (int));

					_dbgprintf (DBG_THREAD, "code %d.\n", code);
					if (code == Quit)
					{
						mutex.unlock();
						
						write (notification.fd[1], &code, sizeof (int));
						return NULL;
					}

					rebuild = true;
				}
				else
				{
					int p = ports.find (port_list[n]);
				
					_dbgprintf (DBG_THREAD, "port %d.\n", p);
					if (p >= 0)
						port(p)->parse_input (tick, p, &sync);
				}
			}
		}
		mutex.unlock();
	}
}



