/*
 *	midi-part.cc
 *	
 *	m_part / m_part_list
 *	
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "midi-event.h"
#include "midi-part.h"

#define DBG_CURSORS	0

/*
 *	m_part::
 *	
 */

m_part::m_part()
: 
m_connector ("m_part"), 
events(),
cursors(),
lock()
{
	division = DEFAULT_DIVISION;
	channel = -1;
	port = -1;
	
	last_add.tick = 0;
	last_add.index = 0;
}

m_part::m_part (m_part * other, int slice_start, int slice_end)
: 
m_connector ("m_part"), 
events(),
cursors(),
lock()
{
	division = DEFAULT_DIVISION;
	channel = -1;
	port = -1;

	last_add.tick = 0;
	last_add.index = 0;
	
	copy_slice (other, slice_start, slice_end);
}

/*
 *	copy.
 *	
 */

void
m_part::copy (m_part * other)
{
	copy_slice (other, 0, other->events.count);
}

void
m_part::copy_slice (m_part * other, int start, int end)
{
	division = other->division;

	events.clear();
	
	other->lock_read();
	
	int i = start;
	while (i < end)
		events.insert (other->event (i++), -1);

	other->unlock_read();
}

m_part::~m_part()
{
	clear();
}

/*
 *	clear all events.
 *
 */

void
m_part::clear()
{
	events.clear();
	
	for (int i = 0; i < cursors.count; i++)
		cursor(i)->init (this, 0);
}

/*
 *	construct/delete a new m_event_cursor
 *
 */

m_event_cursor *
m_part::new_cursor (int index)
{
	m_event_cursor * c = new m_event_cursor (this, index);
	
	_dbgprintf (DBG_CURSORS, "[new_cursor] (%p) : (%p)\n", this, c);
	
	C_INCREF (this);
	
	cursors.insert (c);

	return c;
}

bool
m_part::del_cursor (m_event_cursor * c)
{
	_dbgprintf (DBG_CURSORS, "[del_cursor] attempt (%p) : (%p)\n", this, c);

	if (!cursors.remove (c)) return false;
	
	C_DECREF (this);

	_dbgprintf (DBG_CURSORS, "[del_cursor] (%p) : (%p)\n", this, c);

	return true;
}

/*
 *	get the event index @ >= tick. 
 *	
 */

int
m_part::index_of_tick (m_tick tick)
{
	int i = 0;
	while (i < events.count)
	{
		m_event * e = event(i);

		if (e->get_tick() >= tick)
			return i;

		i++;
	}

	return i;
}

/*
 *	add.
 *
 */

int 
m_part::add_event (m_event * e)
{
	int index = last_add.index;
	
	if (index >= events.count)
	{
		index = 0;
		last_add.tick = 0;
	}

	if (e->get_tick() >= last_add.tick)
	{
		while (index < events.count && e->compare (event (index)) >= 0)
			index++;
	}
	else
	{
		while (index >= 0 && e->compare (event (index)) < 0)
			index--;
		index++;
	}
		
	index = events.insert (e, index);

	if (index < 0) return -1;
	
	for (int i = 0; i < cursors.count; i++)
		if (cursor(i)->index >= index)
			cursor(i)->init (this, cursor(i)->index + 1);
	
	last_add.index = index;
	last_add.tick = e->get_tick();

	return index;
}

/*
 *	parse_note_off.
 *
 *	sizes the corresponding note_on.
 *
 */
 
m_event *				
m_part::parse_note_off (m_event * e, m_tick tick)
{
	assert (e->is_note_off());

	if (tick < 0)
		tick = e->get_tick();

	m_event * on = find_note_on (tick, e->get_channel(), e->get_note());

	if (!on)
		return NULL;

	on->set_note_length (tick - on->get_tick());
	on->set_velocity_off (e->get_velocity_on());

	return on;
} 

/*
 *	remove.
 *	
 */

bool
m_part::remove (int index)
{
	if (index < 0 || index >= events.count) 
		return false;

	events.remove (index);
	
	for (int i = 0; i < cursors.count; i++)
	{
		if (cursor(i)->index == index && index < events.count)
			cursor(i)->init (this, index);
		else if (cursor(i)->index >= index)
			cursor(i)->init (this, cursor(i)->index - 1);
	}

	return true;
}

bool
m_part::remove_event (m_event * e)
{
	return remove (events.find (e));
}

/*
 *	remove a slice.
 *	
 */

void
m_part::remove_slice (int start, int end)
{
	events.remove_slice (start, end);
	
	if (end >= events.count)
		end = events.count - 1;
		
	for (int i = 0; i < cursors.count; i++)
	{
		if (cursor(i)->index >= start && cursor(i)->index < end)
			cursor(i)->init (this, end);
	}
}

/*
 *	move.
 *	
 */

bool
m_part::move_event (int from, m_tick to_tick)
{
	int to = index_of_tick (to_tick);
	if (to >= events.count) to--;
	if (to < 0) return false;

	if (to > from)
		to--;

	if (from == to)
		return false;

	_dbgprintf (0, "to = %d from = %d\n", to, from);

	event(from)->set_tick (to_tick);

	events.move (from, to);

	for (int i = 0; i < cursors.count; i++)
	{
		int index = cursor(i)->index;
		
		if (index == from)
			index = to;
		else 
		{
			if (to < from && index < from && index >= to)
				index++;
			else if (to > from && index > from && index <= to)
				index--;
		}

		if (cursor(i)->index != index)
			cursor(i)->init (this, index);
	}
	
	return true;
}

/*
 *	swap.
 *	
 *	helper for qsort().
 *
 */

void
m_part::swap_events (int i1, int i2)
{
	m_event t;
	memcpy (&t, event(i1), sizeof (m_event));
	memcpy (event(i1), event(i2), sizeof (m_event));
	memcpy (event(i2), &t, sizeof (m_event));

	for (int i = 0; i < cursors.count; i++)
	{
		int index = cursor(i)->index;
		if (index == i1)
			cursor(i)->init (this, i2);
		else if (index == i2)
			cursor(i)->init (this, i1);
	}
}

/*
 *	qsort.
 *	
 */

void
m_part::qsort (int lo, int hi)
{
	if (hi <= lo)
		return;
		
	int l = lo;
	int h = hi;
	m_event * p = event (hi);

	do {
		while ((l < h) && event(l)->compare (p) <= 0)
			l = l + 1;
		while ((h > l) && event(h)->compare (p) >= 0)
			h = h - 1;
		if (l < h) 
			swap_events (l, h);
	} while (l < h);

	swap_events (l, hi);

	qsort (lo, l - 1);
	qsort (l + 1, hi);
}

/*
 *	find the corresponding M_NOTE_ON.
 *	
 */

m_event *
m_part::find_note_on (m_tick t, byte channel, byte note)
{
	/* start at the end and count down, tis more reasonable */
	int i = events.count - 1;
	while (i >= 0 && event(i)->get_tick() > t)
		i--;
	
	while 
	(
		i >= 0 && 
		!(
			event(i)->is_note_on() &&
			event(i)->get_channel() == channel &&
			event(i)->get_note() == note
		)
	) i--;
		
	return i >= 0 ? event (i) : NULL; 
}

/*
 *	return the next meta event >= t
 *	
 */

m_event *
m_part::find_meta (byte type, m_tick t, int direction)
{
	int i = index_of_tick (t);
	if (i >= events.count) i--;
	if (i < 0) return false;

	if (direction >= 0)
		while (i < events.count) 
		{
			if (event(i)->is_meta (type))
				return event(i);
			i++;
		}
	else
		while (i >= 0)
		{
			if (event(i)->is_meta (type))
				return event(i);
			i--;
		}
		
	return NULL;
}

/*
 *	return the next non meta event >= t
 *	
 */

m_event *
m_part::find_non_meta (m_tick t, int direction)
{
	int i = index_of_tick (t);
	if (i >= events.count) i--;
	if (i < 0) return false;

	if (direction >= 0)
		while (i < events.count) 
		{
			if (!event(i)->is_meta())
				return event (i);
			i++;
		}
	else
		while (i >= 0)
		{
			if (!event(i)->is_meta())
				return event(i);
			i--;
		}
		
	return NULL;
}

/*
 *	initialize; e.g. after reading a midi file.
 *	
 */

void
m_part::initialize (int _division)
{
	division = _division;

	m_event * e;
	
	while ((e = find_meta (M_META_EOT)))
		remove_event (e);

	port = get_port();
	channel = get_channel();
}

/*
 *	get && set methods that use meta events for properties.
 *
 */

int
m_part::get_port()
{
	m_event * e = find_meta (M_META_PORT);
	return port = e ? (int) e->get_data_buffer()->items[0] : -1;
}

int
m_part::get_channel()
{
	m_event * e = find_meta (M_META_CHANNEL);
	
	return channel = e ? (int) e->get_data_buffer()->items[0] : -1;
}

void
m_part::set_port (int _port)
{
	port = _port;
	
	m_event * e = find_meta (M_META_PORT);
	if (port < 0)
	{
		remove_event (e);
		return;
	}

	if (!e) 
	{
		e = new m_event (M_META, M_META_PORT);
		add_event (e);
	}
		
	byte data = (byte) _port;
	e->set_data (&data, 1);
}

void
m_part::set_channel (int _channel)
{
	channel = _channel;

	m_event * e = find_meta (M_META_CHANNEL);
	if (channel < 0)
	{
		remove_event (e);
		return;
	}

	if (!e) 
	{
		e = new m_event (M_META, M_META_CHANNEL);
		add_event (e);
	}
		
	byte data = (byte) _channel;
	e->set_data (&data, 1);
}

/*
 *	repr
 *
 */

void
m_part::repr (c_char_buffer * buffer, char limiter)
{
	*buffer << M_PART_NAME" (" << limiter;
	
	for (int i = 0; i < events.count; i++)
	{
		event(i)->repr (buffer);
		*buffer << "," << limiter;
	}

	*buffer << "division=" << division << ")";
}

/*
 *	m_active_notes
 *	
 */

m_active_notes::m_active_notes()
:
m_part()
{
}

bool				
m_active_notes::event_in (m_event * e, m_event_stamp * stamp)
{
	if (!e->is_note_on())
		return false;

	int _port = (stamp && stamp->port >= 0) ? stamp->port : e->get_port();
	int _channel = (stamp && stamp->channel >= 0) ? stamp->channel : e->get_channel();

	m_event * off = 
		new m_event (
			M_NOTE_OFF | _channel, 
			e->get_note(), 
			e->get_velocity_off()
		);

	off->set_port (_port);
	off->set_tick (e->get_tick() + e->get_note_length());

	add_event (off);

	return true;
}

m_event *
m_active_notes::next_before (m_tick tick)
{
	for (int i = 0; i < events.count; i++)
		if (event(i)->get_tick() <= tick)
			return event(i);

	return NULL;
}



