/*
 *	midi-port.cc
 *
 *	m_port; generic midi port base class
 *	m_fd; midi port represented by a file descriptor
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>

#include <stdexcept>

#include "midi-io.h"

#define DBG_PORTS				0
#define DBG_EVENTS_OUT	1
#define DBG_BYTES_OUT		1
#define DBG_BYTES_IN		0
#define DBG_FLOW				0

/*	
 *	m_port
 *
 */

m_port::m_port (char * name) 
: 
m_connector (name), 
state(),
parser()
{
}

m_port::~m_port() 
{
}

bool
m_port::event_in (m_event * e, m_event_stamp * stamp)
{
	/* 0xFF means reset, and we don't dig it. */
	if (e->is_meta())
		return 0;

	state.event_in (e, stamp); 
	return write_event (e, stamp) >= 0;
}


void
m_port::parse_input (m_tick tick, byte port, m_connector * out)
{
	byte data;
	while (read (&data))
	{
		_dbgprintf (DBG_BYTES_IN, "%02x-", (unsigned int) data);
		parser.parse (data, tick, port, out);
	}
}


void
m_port::reset()
{
	m_event e;

	for (int channel = 0; channel < 16; channel++)
	{
		if (state.pitch[channel] != 0.0)
		{
			e.init (M_PITCH_WHEEL | channel, 0x00, 0x40);
			write_event (&e);
			
			state.pitch[channel] = 0.0;
		}
		for (int n = 0; n < 127; n++)
		{
			while (state.note[channel][n] > 0)
			{
				e.init (M_NOTE_OFF | channel, n, 0);
				write_event (&e);

				state.note[channel][n]--;
			}
		}
	}

	/* no need to reset the state. */
}
	










/*	
 *	m_fd 
 *
 *	a file descriptor describing a midi io port.
 *
 */
 
m_fd::m_fd (int _fd, int _mode) 
: 
m_port ("m_fd by fd"),
buffer()
{
	fd = _fd;
	mode = _mode;
	bytes_per_second = 31500.0 / 12.0;

	int r = fcntl (fd, F_SETFL, mode | O_NONBLOCK);
	if (r)
		_dbgprintf (2, "fcntl (O_NONBLOCK) returned %s\n", strerror (errno));
	
	setup();
}

/*
 *	open path and construct an m_fd on it
 *
 */
 
m_fd::m_fd (char * path, int _mode) 
: 
m_port ("m_fd"),
buffer()
{
	fd = open (path, _mode | O_NONBLOCK);
	mode = (_mode & O_ACCMODE);
	bytes_per_second = 31500.0 / 12.0;

	if (fd == -1)
		_dbgprintf (2, "cannot open %s in mode %d: %s\n", path, mode, strerror (errno));
	
	setup();
}

void
m_fd::setup()
{
	fcntl (fd, F_SETOWN, getpid());

	out.running = 0;
	out.pending = 0;
	out.written = 0.0;
	out.dt			= 1.0 / 1000.0;
}

m_fd::~m_fd()
{
	reset();
	close (fd);
}

int
m_fd::get_read_fd()
{
	if (mode == O_RDONLY || mode == O_RDWR)
		return fd;
	return -1;
}

int
m_fd::get_write_fd()
{
	if (mode == O_WRONLY || mode == O_RDWR)
		return fd;
	return -1;
}

/*	
 *	write
 *
 *	translate event into bytes and possibly write them.
 *
 */
 
int
m_fd::write_event (m_event * e, m_event_stamp * stamp)
{
	/* can we write this at all? */
	if (e->get_type() < M_NOTE_OFF || e->get_type() >= M_META)
		return 0;
	
	/* channel message */
	if (e->get_type() < M_SYSEX)
	{
		byte status;
		if (stamp && stamp->channel >= 0 && stamp->channel <= 15)
			status = (byte) (e->get_type() | stamp->channel);
		else
			status = e->midi_status();

		if (status != out.running)
		{
			out.running = status;
			buffer << (char) status;
		}

		int size = e->bytecount();
		
		if (size > 0)
			buffer << (char) e->get_param_0();
		if (size > 1)
			buffer << (char) e->get_param_1();
	}
	else if (e->get_type() == M_SYSEX)
	{
		buffer << (char) e->midi_status();

		if (e->has_data())
			buffer << e->get_data_buffer();
	}
	else 
	{
		/* write M_RT_* without delay; this does not terminate sysex data. */
		int b = e->midi_status();
		::write (fd, &b, 1);
		out.writable -= 1.0;

		if (e->has_data())
		{
			c_char_buffer * data = e->get_data_buffer();
			::write (fd, data->items, data->count);
			out.writable -= (double) data->count;
		}
	}
	
	/* if we do not get periodic advance() calls, we'll experience
	   data congestion. so we'll check for synchronous operation and
		 write the events now. 
	 */
	if (!_async)
		buffer.write (fd, -1); /* write it all */
		
	out.pending = buffer.count;
	
	#if 0
	if (out.pending)
		fprintf (stderr, "%d bytes pending, _async = %d\n", out.pending, _async);
	#endif

	return true;
}

/*
 *	operating the midi port at the specified speed because
 *	we don't want to block when there is a large sysex message
 *	to send.
 *
 */

void
m_fd::advance (double dt, m_connector * nil)
{
	if (!out.pending)
	{
		out.written = 0.0;
		out.writable = bytes_per_second * dt;
		return;
	}

	out.dt = dt;

	out.writable += bytes_per_second * out.dt;
	
	int n_bytes = (int) out.writable;
	
	#if DBG_FLOW
	fprintf (stderr, "this = %p dt = %1.4f n_bytes = %d pending = %d writable = %.4f B/s = %.3f\n",
		this, dt, n_bytes, out.pending, out.writable, bytes_per_second);
	#endif

	if (n_bytes > 0)
	{
		#if DBG_BYTES_OUT
		int n = n_bytes + buffer.next_write;
		if (n > buffer.count)
			n = buffer.count;
		_dbgprintf (1, " >>> [");
		for (int i = buffer.next_write; i < n; i++)
			_dbgprintf (1, "%02x ", (int) buffer[i]);
		_dbgprintf (1, "]\n");
		#endif

		out.pending = buffer.write (fd, n_bytes);
		out.writable -= n_bytes;
	}
}

/*
 *	input parsing
 *
 */
 
bool
m_fd::read (byte * data)
{
	return ::read (fd, data, 1) == 1;
}




