/*
 *	midi-song-cursor.cc
 *
 */
 
#include <assert.h>
#include <string.h>

#include "midi-part.h"
#include "midi-song.h"

#define DBG_BEAT			1
#define DBG_SEEK			0
#define DBG_TEMPO			0
#define DBG_CLICK			0
#define DBG_SONGTIME	0
#define DBG_ENQUEUE		0
#define DBG_SONG			0
#define DBG_PARSE			0

#define DBG_PLAY			0
#define DBG_LOOP			0

#define DBG_EVENTS		0

/*
 *	m_song.
 *
 *	collection of m_event_cursors. tick to time mapper. bar counter.
 *
 */
 
m_song::m_song()
:
m_file ("m_song"),
mutex(),
cursors(),
notes(),
beat(),
clock (M_RT_CLOCK, 0, 0),
default_tempo (M_META, M_META_TEMPO),
default_time (M_META, M_META_TIME)
{
	notes.disable_refcounting();
	
	clock.disable_refcounting();
	clock.set_port (M_BROADCAST_PORT);

	playing = 0;
	
	loop.start = 
	loop.end = 0;
	loop.count = 0;

	default_tempo.is_tempo()->set_bpm (120.0);
	
	default_time.is_time()->set_numerator (4);
	default_time.is_time()->set_denominator (4);
	
	tick.current = 0;
	time.current = 0.0;

	parse_tempo (default_tempo.is_tempo());
	beat.parse_time (default_time.is_time());

	reset();
}

m_song::~m_song()
{
	clear();
}

/*
 *	reset everything to song start
 *
 */

void
m_song::reset()
{
	tick.current = 0;
	time.current = 0.0;

	seek (0.0, 0, (m_connector *) 0);
	
	_dbgprintf (1, "m_song::reset() division = %d beat.division = %d\n", division, beat.division);
}

/*
 *	read
 *
 */
 
int
m_song::read (char * _name)
{
	if (!m_file::read (_name))
		return 0;
	
	reset();

	return parts.count;	
}

/*
 *	play and stop
 *
 */

bool
m_song::play()
{
	if (playing)
		return true;

	/* cache port && channel prefixes */
	for (int i = 0; i < cursors.count; i++)
	{
		m_part * p = cursors[i]->part;

		p->lock_read();
		p->get_port();
		p->get_channel();
		p->unlock_read();
	}

	return playing = true;
}

bool
m_song::stop()
{
	playing = false;

	return true;
}

/*
 *	time <-> tick translation
 *
 *	only valid within the quantize between the last and the next tempo change.
 */

double
m_song::tick2time (m_tick _tick)
{
	return time.of_last_tempo + (double) 
		(_tick - tick.of_last_tempo) * time.per_tick;
}

m_tick
m_song::time2tick (double _time)
{
	return tick.of_last_tempo + (m_tick) 
		((_time - time.of_last_tempo) / time.per_tick);
}

/*
 *	emit event
 *
 */

void
m_song::emit_event (m_event_cursor * c, m_connector * out)
{
	m_event_stamp stamp;
	m_event * e = c->event;
	m_part * p = c->part;

	stamp.port = p->port;
	stamp.channel = p->channel;
	stamp.time = tick2time (e->get_tick());
	stamp.part = p;

	out->event_in (e, &stamp);
				
	if (e->get_type() == M_NOTE_ON)
		notes.event_in (e, &stamp);
}

/*
 *	stop_active_notes
 *
 *	emits pending note-offs
 *
 */

void
m_song::stop_active_notes (m_tick until, m_connector * out)
{
	m_event_stamp stamp;
	stamp.time = time.current;
	
	for (int i = 0; i < notes.events.count; )
	{
		m_event * e = notes.event(i);
		
		if (e->get_tick() < until)
		{
			if (out) 
				out->event_in (e, &stamp);
			notes.remove (i);
			continue;
		}
		i++;
	}
}

/*
 *	advance
 *
 */

void
m_song::advance (double dt, m_connector * out)
{
	if (!playing)
		return;

	_dbgprintf (DBG_PLAY, 
		">>> t=%1.4f + %1.4f (tick=%d) out=%p>>>\n",
		time.current, dt, tick.current, out);

	advance_to_time (time.current + dt, out);
}

/*
 *	advance (time)
 *
 */

void
m_song::advance_to_time (double _time, m_connector * out)
{
	mutex.lock();

	while (1)
	{
		m_tick _tick = time2tick (_time); /* estimated tick of arrival */

		if (_tick <= tick.current) /* already arrived */
			break;
			
		cursors[0]->part->lock_read();

		m_event * e = cursors[0]->event;
		if (e && e->get_tick() < _tick)
			_tick = e->get_tick();

		cursors[0]->part->unlock_read();

		if (loop.start == loop.end || tick.current >= loop.end || _tick < loop.end)
			advance_to_tick (_tick, out);
		else /* hit a valid loop end */
		{
			advance_to_tick (loop.end, out);
			stop_active_notes (M_TICK_MAX, out);
			double dt = _time - time.current;
			seek ((double) TEN_TO_THE_NINTH, loop.start, (m_connector *) 0);
			_time = time.current + dt;
		}
	}

	time.current = _time;

	mutex.unlock();
}

/*
 *	advance (tick)
 *
 */

void
m_song::advance_to_tick (m_tick _tick, m_connector * out)
{
	stop_active_notes (_tick, out);

	/* emit loop */
	for (int i = 0; i < cursors.count; i++)
	{
		m_event_cursor * c = cursors[i];

		c->part->lock_read();

		while (c->event && c->tick <= _tick)
		{
			/* emit it */
			parse (c->event, out);
			if (out) emit_event (c, out);
			c->step_fwd();
		}
		
		c->part->unlock_read();
	}

	time.current = tick2time (tick.current = _tick);
		
	beat.advance (tick.current, out);
}

/*
 *	seek (bar, beat, fraction)
 *
 */

bool
m_song::seek (int _bar, int _beat, int _fraction, m_connector * out)
{
	if (_bar < 0 || _beat < 0 || _fraction < 0)
		return false;

	if (_bar <= beat.bar)
		seek (0.0);

	if (!cursors.count)
	{
		beat.seek (_bar, _beat, _fraction);
		tick.current = beat.tick.current;
		time.current = tick2time (tick.current);
		return true;
	}

	m_event_cursor * c = cursors[0];
	
	bool done = false;
	while (!done)
	{
		m_beat test (beat);
		test.seek (_bar, _beat, _fraction);
		
		m_tick _tick = test.tick.current;
		
		c->part->lock_read();
		
		if (!c->event)
			done = true;
		else if (c->tick >= test.tick.current)
			done = true;
		else
		{
			_tick = c->tick;
			c->step_fwd();
		}

		c->part->unlock_read();

		advance_to_tick (_tick);
	}

	return true;
}

/*
 *	parse an event if it's a tempo or time signature
 *
 */

bool
m_song::parse (m_event * e, m_connector * out)
{
	#if defined (DEBUG) && DBG_PARSE > 0
	e->debug ("m_song::parse");
	#endif

	if (e->is_time())
		beat.parse_time (e->is_time(), out);
	else if (e->is_tempo())
		parse_tempo (e->is_tempo());
	else
		return false;

	return true;
}

/*
 *	update tempo information.
 *
 */

void
m_song::parse_tempo (m_tempo_event * e)
{
	time.usec_per_quarter = e->usec_per_quarter();
	time.per_quarter = (double) time.usec_per_quarter / TEN_TO_THE_SIXTH;
	time.per_tick = time.per_quarter / (double) division;

	time.of_last_tempo = time.current;
	tick.of_last_tempo = tick.current;
	
	_dbgprintf (DBG_TEMPO,
		"[parse_tempo] parsed %1.2f bpm; usec/b = %d; t/t = %1.5f\n",
		e->bpm(),
		e->usec_per_quarter(),
		time.per_tick);
	
	_dbgprintf (DBG_TEMPO, "[after parse_tempo] %.4f %d\n", time.current, tick.current);
}


/*
 *	seek.
 *
 *	songpointer = min (_time, _tick): 
 *	which is first cannot be known without evaluating all events in 
 *	the part holding the tempo events.
 *
 *	if a seek to the current pointer is requested, a complete reparse
 *	is done.
 *
 */

bool
m_song::seek (double _time, m_tick _tick, m_connector * out)
{
	if (_time < 0 || _tick < 0)
		return false;

	if (!cursors.count)
	{
		m_tick time_tick = time2tick (_time);
		if (time_tick >= 0 && time_tick < _tick)
		{
			tick.current = time_tick;
			time.current = _time;
		}
		else
		{
			tick.current = _tick;
			time.current = tick2time (_tick);
		}
		beat.reset (division);
		beat.advance (tick.current);
		return true;
	}

	bool rewind = (_time <= time.current || _tick <= tick.current);
		
	if (rewind)
	{
		tick.current = 0;
		time.current = 0.0;

		beat.reset (division);
	}

	/* evaluate the time/tempo part and get the current tick */
	
	m_event_cursor * c = cursors[0];
	
	if (rewind) c->init();
	
	while (c->event)
	{
		m_tick tic = c->tick;
		double tim = tick2time (tic);
		
		if (tim >= _time)
			break;

		if (tic > _tick)
			break;

		parse (c->event, out);
		if (out) emit_event (c, out);

		if (!rewind)
			beat.advance (tic, out);

		_dbgprintf (DBG_SEEK, "tim = %1.3f _time = %1.4f\n", tim, _time);
		_dbgprintf (DBG_SEEK, "tic = %d     _tick = %d\n", tic, _tick);
		_dbgprintf (DBG_SEEK, "c->tick = %d\n", c->tick);

		c->step_fwd();
	}


	/* now syncing _time and _tick. */
	
	double time_from_tick = tick2time (_tick);

	/* tick earlier */
	if (time_from_tick < _time) 
		tick.current = _tick,
		time.current = time_from_tick;
	
	/* tick later [than time.] */
	else
		tick.current = time2tick (time.current = _time);

	beat.advance (tick.current, rewind ? NULL : out);


	/* all other parts */
	
	for (int i = 1; i < cursors.count; i++)
	{
		c = cursors[i];
		
		if (rewind) c->init();

		while (c->event && c->tick < tick.current)
		{
			if (out) emit_event (c, out);
			c->step_fwd();
		}
	}
	
	_dbgprintf (0, "[song] seek %d = %1.4f\n", tick.current, time.current);

	return true;
}


/*
 *	
 *
 */

void
m_song::set_division (m_tick _division)
{
	beat.division = 
	division = _division;
	tick.per_clock = division / 24;
}

/*
 *	
 *
 */

bool
m_song::set_file (m_file * _file)
{
	clear();
	
	if (_file)
	{
		for (int i = 0; i < _file->parts.count; i++)
			add_part (_file->part(i));
		_dbgprintf (DBG_SONG, "[set_file] division: %d ==> %d\n", division, _file->division);
		set_division (_file->division);
		set_name (_file->name);
		beat.reset (division);
	}
	else
	{
		set_name ("");
		set_division (DEFAULT_DIVISION);
		return true;
	}

	return true;
}

/*
 *	add a part	
 *
 *	return the part index in the list.
 */
 
int
m_song::add_part (m_part * part)
{
	if (m_file::add_part (part) < 0)
		return -1;

	if (!parts.count)
		beat.division = division = part->division;

	m_event_cursor * c = part->new_cursor (0);
	c->move_to (tick.current);
	cursors.insert_item (c);
	
	return cursors.count - 1;
}

/*
 *	
 *
 */

bool
m_song::remove_part (m_part * part)
{
	if (!m_file::remove_part (part))
		return false;

	for (int i = 0; i < cursors.count; i++)
	{
		m_event_cursor * c = cursors[i];
		if (c->part == part)
		{
			part->del_cursor (c);
			cursors.remove (i);
			return true;
		}
	}
	return false;
}

/*
 *	clear parts list
 *
 */
 
void
m_song::clear()
{
	while (cursors.count)
	{
		if (!remove_part (part(0)))
			fprintf (stderr, "m_song::clear() cannot remove this->part(0)\n");
	}
}





/*
 *	m_beat
 *
 */

m_beat::m_beat()
: 
click (M_NOTE_ON | 9, 37, 0),
clock (M_RT_CLOCK)
{
	click.disable_refcounting();
	clock.disable_refcounting();

	tick.quantize = 0;

	reset (DEFAULT_DIVISION);
}

m_beat::m_beat (m_beat &other)
: 
click (M_NOTE_ON | 9, 37, 0),
clock (M_RT_CLOCK)
{
	click.disable_refcounting();
	clock.disable_refcounting();

	tick.current = other.tick.current;
	tick.of_last_event = other.tick.of_last_event;
	tick.per_beat = other.tick.per_beat;
	
	division = other.division;

	numerator = other.numerator;
	denominator = other.denominator;

	bar = other.bar;
	beat = other.beat;
	fraction = other.fraction;
}

void
m_beat::reset (int _division)
{
	_dbgprintf (DBG_BEAT, "[beat::reset (%d)]\n", _division);

	tick.current = 
	tick.of_last_event = 0;

	tick.per_beat =
	division = _division;
	
	numerator = 4;
	denominator = 4;
	
	bar =
	beat = 
	fraction = 0;
}

/*
 *	parse a time event.
 *
 *	the song needs to have advanced the beat until this tick before.
 *
 */

void
m_beat::parse_time (m_time_event * e, m_connector * out)
{
	if (e->tick > tick.current)
		advance (e->tick, out);

	if (beat || fraction)
		bar++;

	beat = 0;
	fraction = 0;
		
	numerator = e->get_numerator();
	denominator = e->get_denominator();

	tick.per_beat = (4 * division) / denominator;
	
	tick.of_last_event = e->tick;
	
	_dbgprintf (DBG_CLICK, "m_beat::parse (%d/%d [%d / beat] @ %d)", numerator, denominator, tick.per_beat, e->tick);
}

/*
 *	advance until (bar, beat, fraction)
 *
 */

void
m_beat::seek (int _bar, int _beat, int _fraction)
{
	m_tick _tick = tick.current;
	
	_tick += (_bar - bar) * numerator * tick.per_beat;
	_tick += (_beat - beat) * tick.per_beat;
	_tick += _fraction - fraction;

	assert (_tick >= tick.current);
	
	advance_to_tick (_tick);
}

/*
 *	when is the next beat?
 *
 */
 
m_tick
m_beat::next_beat (m_tick _tick, m_tick _quantize)
{
	m_tick click = (_tick + _quantize);
	click -= (click - tick.of_last_event) % _quantize;

	return click;
}

/*
 *	update
 *
 */

void
m_beat::advance_to_tick (m_tick _tick)
{
	fraction += _tick - tick.current;
	beat += fraction / tick.per_beat;
	
	bar += beat / numerator;

	beat %= numerator;
	fraction %= tick.per_beat;

	tick.current = _tick;
}

/*
 *	advance, count, emit click and clock
 *
 */
 
void
m_beat::advance (m_tick _tick, m_connector * out)
{
	static int clicks = 0;

	if (_tick <= tick.current)
		return;

	if (!out)
	{
		advance_to_tick (_tick);
		return;
	}

	m_tick _quantize = tick.quantize ? tick.quantize : tick.per_beat / 24;
		
	_dbgprintf (DBG_CLICK, "%d > [%d] > %d (div=%d)\n", tick.current, _quantize, _tick, division);

	if (!fraction && !beat)
	{
		_dbgprintf (DBG_CLICK, "  %3d.%d.%-3d: %d clicks>\n", bar, beat, fraction, clicks);
		clicks = 0;
	}

	if (!(fraction % _quantize)) /* we are at a click */
	{
		_dbgprintf (DBG_CLICK, " <%3d.%d.%-3d> %d\n", bar, beat, fraction, tick.current);
		
		clicks++;
		clock.set_tick (tick.current);
		out->event_in (&clock);
		out->click_in (tick.current, bar, beat, fraction);
	}
		
	m_tick next = next_beat (tick.current, _quantize);

	while (next > tick.current && next < _tick)
	{
		_dbgprintf (DBG_CLICK, "          > %d\n", next);
		advance_to_tick (next);

		_dbgprintf (DBG_CLICK, "  %3d.%d.%-3d  %d\n", bar, beat, fraction, tick.current);
		if (!(fraction % _quantize))
		{
			_dbgprintf (DBG_CLICK, " [%3d.%d.%-3d] %d\n", bar, beat, fraction, tick.current);
			clicks++;
			clock.set_tick (tick.current);
			out->event_in (&clock);
			out->click_in (tick.current, bar, beat, fraction);
		}
		
		next = next_beat (tick.current, _quantize);
	}

	advance_to_tick (_tick);
}



