/*
 *	midi-sync.cc
 *
 */

#include "midi-song.h"

#define DBG_CLOCK	0

/*
 *	m_sync is responsible for synchronizing a m_song to MIDI time codes
 *	(or possibly other sorts of timecode as well)
 *
 *	not yet functional at all. sorry folks.
 */
 
m_sync::m_sync()
:
m_connector ("m_sync"),
songs()
{
	through = NULL;

	use = false;
	locked = false;

	time.current = 0.0;
	time.per_frame = 0.0;

	tick.current = 0;

	division = DEFAULT_DIVISION;

	mtc.mask = 0;
	mtc.quarter = 0;
}

/*
 *	list of songs maintenance
 *
 */

int
m_sync::add_song (m_song * _song)
{
	songs.mutex.lock();
	int i = songs.insert (_song);
	songs.mutex.unlock();

	if (!i) division = _song->division;

	return i;
}

bool
m_sync::remove_song (m_song * _song)
{
	songs.mutex.lock();
	bool done = songs.remove (_song);
	songs.mutex.unlock();

	return done;
}

/*
 *	clean up.
 *
 */
 
void
m_sync::set_through (m_connector * _through)
{
	C_XINCREF (_through);
	C_XDECREF (through);
	through = _through;
}

/*
 *	clean up.
 *
 */
 
m_sync::~m_sync()
{
	C_XDECREF (through);
	
	songs.clear();
}

/*
 *	receives a clock pulse (delta t value)
 *
 */
 
void
m_sync::advance (double dt, m_connector * out)
{
	if (!songs.count)
		return;

	songs.mutex.lock();
	
	for (int i = 0; i < songs.count; i++)
		song(i)->advance (dt, out);

	if (songs.count)
		tick.current = song(0)->get_tick();

	songs.mutex.unlock();
}

/*
 *	[to be] called by internals
 *
 */
 
void
m_sync::advance_to (double t)
{
	time.current = t;
	/* do something useful */
}

/*
 *	parse an event which has come in through a midi port
 *
 */
 
bool
m_sync::event_in (m_event * e, m_event_stamp * stamp)
{
	bool pass_on = true;

	e->set_tick (tick.current);

	switch (e->get_type())
	{
		/* possibly mtc sync */
		case M_SYSEX:
			pass_on = !do_sysex (e);
			break;

		case M_RT_QUARTER_FRAME:
			do_quarter_frame (e);
			return true;

		/* clock sync */
		case M_RT_SONG_POS:
			do_song_pos (e);
			return true;

		case M_RT_CLOCK:
			#if defined (DEBUG) && (DBG_CLOCK > 0)
				e->debug ("m_sync::M_RT_CLOCK");
			#endif
			do_clock();
			return true;
			
		case M_RT_START:
			do_start();
			return true;
			
		case M_RT_CONTINUE:
			do_continue();
			return true;
			
		case M_RT_STOP:
			do_stop();
			return true;
	}

	if (!(through && pass_on))
		return false;

	return through->event_in (e, stamp);
}

/*
 *	parse a SMPTE quarter frame message
 *
 */

void
m_sync::do_quarter_frame (m_event * e)
{
	byte b = e->get_data_buffer()->items[0];

	if (b & 0x80)
	{
		_dbgprintf (2, "received byte %02x\n", b);
		return;
	}
			
	if (locked)
		time.current += time.per_frame;
	
	int in_sequence = (b >> 4);
	
	mtc.quarter = in_sequence & 0x3;

	mtc.mask |= (1 << in_sequence);
	
	switch (in_sequence)
	{
		case 0:
			mtc.frame |= (b & 0xF);
			break;
		case 1:
			mtc.frame |= (b & 0xF) << 4;
			break;
		case 2:
			mtc.second |= (b & 0xF);
			break;
		case 3:
			mtc.second |= (b & 0xF) << 4;
			break;
		case 4:
			mtc.minute |= (b & 0xF);
			break;
		case 5:
			mtc.minute |= (b & 0xF) << 4;
			break;
		case 6:
			mtc.hour |= (b & 0xF);
			break;
		case 7:
			mtc.hour |= (b & 0x1) << 4;
			switch ((b & 0x5) >> 1)
			{
				case 0:
					time.per_frame = 1. / 24.;
					break;
				case 1:
					time.per_frame = 1. / 25.;
					break;
				case 2:
					time.per_frame = 1. / 29.97;
					break;
				case 3:
					time.per_frame = 1. / 30.;
					break;
			}
			break;
	}
	
	if (mtc.mask == 0xFF)
	{
		time.current =
			(double) mtc.quarter * time.per_frame / 4. + 
			(double) mtc.frame * time.per_frame + 
			(double) mtc.second +
			(double) mtc.minute * 60. +
			(double) mtc.hour * 60. * 60.;
		
		locked = true;

		mtc.mask = 0;
	}

	if (locked)
		advance_to (time.current);
}

/*
 *	parse a sysex message
 *
 *	return true if it was sync related.
 */

bool
m_sync::do_sysex (m_event * e)
{
	return false;
}

void
m_sync::do_song_pos (m_event * e)
{
}

void
m_sync::do_clock()
{
}

void
m_sync::do_start()
{
}

void
m_sync::do_stop()
{
}

void
m_sync::do_continue()
{
}










