/*
 *	python-gtk-event-view.cc
 *
 *	Python event view object
 *
 */

#ifdef WITH_PYGTK

#include "python-types.h"
#include "python-gtk-types.h"

extern "C" {
	#include <pygtk/pygtk.h>
}

/* * * * * * * * * * * * * * * * * * * * 
 *
 *	make it
 *
 */

gtk_view_object *
gtk_view_new_from_gtk_event_view (gtk_event_view * view)
{
	gtk_view_object * self = PyObject_NEW (gtk_view_object, &gtk_view_py_type);
	if (!self)
		return NULL;

	if (!(self->view = view))
	{
		self->view = new gtk_event_view();
		if (!self->view)
		{
			PyMem_DEL (self);
			PyErr_NoMemory();
			return NULL;
		}
	}
		
	_dbgprintf (DBG_REF, "<gtk_view_object %x>\n", (int) self);

	C_INCREF (self->view);

	return self;
}

/*
 *	
 *
 */

static void
gtk_view_del (gtk_view_object * self)
{
	_dbgprintf (DBG_REF, "<~m_song_object %x>\n", (int) self);

	C_DECREF (self->view);

	PyMem_DEL (self);
}

/* * * * * * * * * * * * *
 * 
 *	methods
 *
 */

static PyObject *
gtk_view_redraw (gtk_view_object * self, PyObject * args)
{
	self->view->full_redraw();
	
	Py_INCREF (Py_None);
	return Py_None;
}

static PyObject *
gtk_view_tick2x (gtk_view_object * self, PyObject * args)
{
	m_tick tick;
	
	if (args && !PyArg_ParseTuple (args, "i;tick to convert", &tick))
		return NULL;
	
	return PyInt_FromLong (self->view->x_mapper.tick2x (tick));
}

static PyObject *
gtk_view_x2tick (gtk_view_object * self, PyObject * args)
{
	int x;
	
	if (args && !PyArg_ParseTuple (args, "i;x position to convert", &x))
		return NULL;
	
	return PyInt_FromLong (self->view->x_mapper.x2tick (x));
}

static PyObject *
gtk_view_get_color (gtk_view_object * self, PyObject * args)
{
	int port, channel;
	
	if (args && !PyArg_ParseTuple (args, "ii;port and channel", &port, &channel))
		return NULL;
	
	return rgb_vector_new (self->view->get_color (port, channel));
}

static PyObject *
gtk_view_move (gtk_view_object * self, PyObject * args)
{
	double delta;
	
	if (args && !PyArg_ParseTuple (args, "d;delta time", &delta))
		return NULL;
	
	double dx = self->view->pixels.size.width * -delta;

	self->view->delayed_scroll ((int) dx);

	Py_INCREF (Py_None);
	return Py_None;
}

static PyMethodDef
gtk_view_methods[] = 
{
	{"redraw", (PyCFunction) gtk_view_redraw, METH_VARARGS, 
		"redraw()\n"
		"redraws the entire "G_VIEW_NAME"."
	},
	{"tick2x", (PyCFunction) gtk_view_tick2x, METH_VARARGS, 
		"tick2x (tick)\n"
		"returns the pixel position that corresponds to 'tick'."
	},
	{"x2tick", (PyCFunction) gtk_view_x2tick, METH_VARARGS, 
		"x2tick (x)\n"
		"returns the tick that is represented by pixel position 'x'."
	},
	{"get_color", (PyCFunction) gtk_view_get_color, METH_VARARGS, 
		"get_color (port, channel)\n"
		"returns the color that represents events on the 'port' / 'channel' combination."
	},
	{"move", (PyCFunction) gtk_view_move, METH_VARARGS,
		"move (delta_t)\n"
		"moves the viewport time coordinate; delta_t is a float between -1 and +1."
	},
	{NULL, NULL}
};

/*
 *	getattr
 *
 */

static PyObject *
gtk_view_getattr (gtk_view_object * self, char * name)
{
	if (!strcmp (name, "__members__"))
	{
		PyObject * list = PyList_New (7);
		
		if (!list)
			return NULL;
	
		PyList_SetItem (list, 0, PyString_FromString ("widget"));
		PyList_SetItem (list, 1, PyString_FromString ("song"));
		PyList_SetItem (list, 2, PyString_FromString ("staff_alpha"));
		PyList_SetItem (list, 3, PyString_FromString ("back"));
		PyList_SetItem (list, 4, PyString_FromString ("fore"));
		PyList_SetItem (list, 5, PyString_FromString ("follow"));
		PyList_SetItem (list, 6, PyString_FromString ("zoom"));
	
		if (PyErr_Occurred()) 
		{
			Py_DECREF (list);
			list = NULL;
		}

		return list;
	}
	
	if (!strcmp (name, "widget"))
		return PyGtk_New ((GtkObject *) self->view->widget);

	if (!strcmp (name, "song"))
		return (PyObject *) m_song_new_from_song (&self->view->song);

	if (!strcmp (name, "follow"))
		return (PyObject *) m_song_new_from_song (self->view->follow.song);

	if (!strcmp (name, "alpha"))
		return PyInt_FromLong (self->view->staff_alpha);

	if (!strcmp (name, "back"))
		return rgb_vector_new (&self->view->back);

	if (!strcmp (name, "fore"))
		return rgb_vector_new (&self->view->fore);
	
	if (!strcmp (name, "zoom"))
		return PyInt_FromLong (self->view->magnification);
	
	return Py_FindMethod (gtk_view_methods, (PyObject *) self, name);
}

/*
 *	setattr
 *
 */

static int
gtk_view_setattr (gtk_view_object * self, char * name, PyObject * value)
{
	if (!value) 
	{
		PyErr_SetString (PyExc_TypeError, "attribute deletion is not supported");
		return -1;
	}

	if (!strcmp (name, "alpha"))
	{
		int alpha = PyInt_AS_LONG (value);
		if (alpha < 0 || alpha > 255)
		{
			PyErr_SetString (PyExc_ValueError, "0 < alpha < 256.");
			return -1;
		}
		self->view->staff_alpha = alpha;
		return 0;
	}

	if (!strcmp (name, "back"))
	{
		c_color rgb;
		if (!PyArg_ParseTuple (value, "iii", &rgb.r, &rgb.g, &rgb.b))
			return -1;
		self->view->back = rgb;
		return 0;
	}

	if (!strcmp (name, "fore"))
	{
		c_color rgb;
		if (!PyArg_ParseTuple (value, "iii", &rgb.r, &rgb.g, &rgb.b))
			return -1;
		self->view->fore = rgb;
		return 0;
	}

	if (!strcmp (name, "follow"))
	{
		m_song * song;
		
		if (value == NULL || value == Py_None)
			song = NULL;
		else if (!m_song_Check (value))
			return -1;
		else
			song = ((m_song_object *) value)->song;

		self->view->set_follow (song);
		return 0;
	}

	if (!strcmp (name, "zoom"))
	{
		int magnification = PyInt_AS_LONG (value);
		if (magnification < -10 || magnification > 10)
		{
			PyErr_SetString (PyExc_ValueError, "-10 < zoom < 10.");
			return -1;
		}
		self->view->set_magnification (magnification);
		return 0;
	}

	if (!strcmp (name, "snap"))
	{
		int snap = PyInt_AS_LONG (value);
		if (snap < 1 || snap >= M_TICK_MAX)
		{
			PyErr_SetString (PyExc_ValueError, "1 <= snap <= max (int32)");
			return -1;
		}
		self->view->snap = snap;
		return 0;
	}

	return -1;
}

PyTypeObject 
gtk_view_py_type =
{
	PyObject_HEAD_INIT (&PyType_Type)
	0, // ob_size 
	G_VIEW_NAME,
	sizeof (gtk_view_object),
	0, // tp_itemsize 

	// methods 
	
	(destructor) gtk_view_del,
	(printfunc) 0,
	(getattrfunc) gtk_view_getattr,
	(setattrfunc) gtk_view_setattr,
	(cmpfunc) 0,
	(reprfunc) 0,
	0, // tp_as_number 
	0, // tp_as_sequence 
	0, // tp_as_mapping 
	(hashfunc) 0,
	(ternaryfunc) 0,
	(reprfunc) 0,

	// reserved ... 
	
	0,0,0,0,

	// __doc__ 
	
	"A "G_VIEW_NAME" is gaaf."
};

gtk_view_object *
module_gtk_view_new (PyObject * self, PyObject * args)
{
	#if 0
	if (!PyArg_ParseTuple (args, ""))
		return NULL;
		
	return gtk_view_new_from_gtk_event_view (NULL);
	#endif
	PyGtk_Object * object = NULL;
	
	if (!PyArg_ParseTuple (args, "|O!", &PyGtk_Type, &object))
		return NULL;
	
	if (object && !GTK_IS_DRAWING_AREA (object->obj))
	{
		PyErr_SetString (PyExc_TypeError, "Need a GtkDrawingArea.");
		return NULL;
	}

	gtk_event_view * view = 
		new gtk_event_view (object ? (GtkWidget *) object->obj : NULL);
	
	gtk_view_object * result = gtk_view_new_from_gtk_event_view (view);

	if (!result)
		delete view;

	return result;
}

bool
pygtk_init()
{
	PyObject * pygtk = PyImport_ImportModule("_gtk"); 
	if (!pygtk)
		return false;
		
	PyObject * module_dict = PyModule_GetDict (pygtk); 
	PyObject * cobject = PyDict_GetItemString (module_dict, "_PyGtk_API");
	
	if (!PyCObject_Check(cobject))
	{
		PyErr_SetString (PyExc_ImportError, "could not find _PyGtk_API object");
		return false;
	}

	_PyGtk_API = 
		(struct _PyGtk_FunctionStruct *) PyCObject_AsVoidPtr (cobject); 
		
	#if 0
	/* calling gtk_init multiple times does not hurt. */

	PyObject * args = PySys_GetObject ("argv");
	int argc = PyList_Size (args);
	char ** argv = (char **) malloc (argc * sizeof (char *));
	for (int i = 0; i < argc; i++)
		argv[i] = strdup (PyString_AsString (PyList_GetItem (args, i)));
	
	gtk_init (&argc, &argv);

	for (int i = 0; i < argc; i++)
		if (argv[i] != NULL)
			free (argv[i]);
	free (argv);
	#endif
	return true;
}

#endif /* WITH_PYGTK */
