/*
 *	python-midi-clock.cc
 *
 */

#include "python-types.h"

m_clock_object *
m_clock_new_from_clock (m_clock * clock)
{
	m_clock_object * self = PyObject_NEW (m_clock_object, &m_clock_py_type);
	if (!self) 
	{
		PyErr_NoMemory();
		return NULL;
	}
	
	_dbgprintf (DBG_REF, "<m_clock_object %x>\n", (int) self);

	self->clock = clock;

	C_INCREF (self->clock);
	
	return self;
}

static void
m_clock_del (m_clock_object * self)
{
	_dbgprintf (DBG_REF, "<~m_clock_object %x>\n", (int) self);

	C_DECREF (self->clock);
	
	PyMem_DEL (self);
}

static PyObject *
m_clock_start (m_clock_object * self)
{
	PyObject * result = Py_None;
	
	if (!self->clock->start())
	{
		PyErr_SetString (PyExc_SystemError, "dunno.");
		result = NULL;
	}

	Py_XINCREF (result);
	return result;
}

static PyObject *
m_clock_stop (m_clock_object * self)
{
	PyObject * result = Py_None;
	
	if (!self->clock->stop())
	{
		PyErr_SetString (PyExc_SystemError, "dunno.");
		result = NULL;
	}

	Py_XINCREF (result);
	return result;
}

static PyObject *
m_clock_set_resolution (m_clock_object * self, PyObject * args)
{
	unsigned hz;
	
	if (!PyArg_ParseTuple (args, "i;Hz", &hz))
		return NULL;
	
	PyObject * result = Py_None;

	if (!self->clock->set_resolution (hz))
	{
		PyErr_SetString (PyExc_SystemError, "dunno.");
		result = NULL;
	}

	Py_XINCREF (result);
	return result;
}

/*
 *	m_clock_methods
 *
 */

static PyMethodDef
m_clock_methods[] = 
{
	{"start", (PyCFunction) m_clock_start, METH_VARARGS, 
	 "start()\n"
	 "starts the clock. The clock is normally running from the "
	 "moment on it is created."},
	{"stop", (PyCFunction) m_clock_stop, METH_VARARGS, 
	 "stop()\n"
	 "stops the clock."},
	{"set_resolution", (PyCFunction) m_clock_set_resolution, METH_VARARGS, 
	 "set_resolution (hz)\n"
	 "sets the resolution of the clock in Hz."},
	{NULL, NULL}
};

/*
 *	m_clock.getattr
 *
 */

static PyObject *
m_clock_getattr (m_clock_object * self, char * name)
{
	if (!strcmp (name, "running"))
		return PyInt_FromLong (self->clock->running);
		
	if (!strcmp (name, "__members__"))
	{
		PyObject * result = PyList_New (1);
		if (result)
		{
			PyList_SetItem (result, 0, PyString_FromString ("running"));
		
			if (PyErr_Occurred()) 
			{
				Py_DECREF (result);
				result = NULL;
			}
		}
		return result;
	}
	
	return Py_FindMethod (m_clock_methods, (PyObject *) self, name);
}	

/*
 *	m_clock.setattr
 *
 */

static int
m_clock_setattr (m_port_object * self, char * name, PyObject * arg)
{
	if (!strcmp (name, "running"))
		PyErr_SetString (PyExc_NotImplementedError, "read only.");
	else
		PyErr_SetString (PyExc_AttributeError, name);
	return -1;
}	


/*
 *
 *
 */

PyTypeObject 
m_clock_py_type =
{
	PyObject_HEAD_INIT (&PyType_Type)
	0, // ob_size 
	M_CLOCK_NAME,
	sizeof (m_clock_object),
	0, // tp_itemsize 

	// methods 
	(destructor) m_clock_del,
	(printfunc) 0,
	(getattrfunc) m_clock_getattr,
	(setattrfunc) m_clock_setattr,
	(cmpfunc) 0,
	(reprfunc) 0,
	0, // tp_as_number 
	0, // tp_as_sequence 
	0, // tp_as_mapping 
	(hashfunc) 0,
	(ternaryfunc) 0,
	(reprfunc) 0,

	// more ... 
	
	0,0,0,0,

	// __doc__ 
	
	"The "M_CLOCK_NAME" object drives a "M_SYNC_NAME"."
};

m_clock_object *
module_m_clock_new (PyObject * self, PyObject * args)
{
	m_clock * clock;
	
	clock = m_clock::get();

	if (!clock)
	{
		PyErr_SetString (PyExc_SystemError, "Cannot instantiate a clock.");
		return NULL;
	}

	m_clock_object * clock_object = m_clock_new_from_clock (clock);

	if (!clock_object)
		delete clock;
		
	return clock_object;
}
