/*
 *	python-midi-event-values.cc
 *
 */

#include "python-types.h"

/*
 *	get the value names this event supports
 *
 */

char **
m_event::get_value_names()
{
	/* standard events */
	static char * note_values[] = 
		{"status", "note", "velocity", "tick", "length", "velocity_off", 
		 "channel", "port", "type", NULL};
		 
	static char * note_off_values[] = 
		{"status", "note", "velocity", "tick", 
		 "channel", "port", "type", NULL};
		 
	static char * note_touch_values[] = 
		{"status", "note", "velocity", "tick", 
		 "channel", "port", "type", NULL};
		 
	static char * control_values[] = 
		{"status", "controller", "value", "tick", 
		 "channel", "port", "type", NULL};
		 
	static char * program_values[] = 
		{"status", "program", "tick", 
		 "channel", "port", "type", NULL};
		 
	static char * channel_touch_values[] =
		{"status", "velocity", "tick", 
		 "channel", "port", "type", NULL};
		 
	static char * pitch_values[] =
		{"status", "pitch", "tick", 
		 "channel", "port", "type", NULL};
		 
	static char * sysex_values[] =
		{"status", "data", "tick", 
		 "port", "type", NULL};
		 
	static char * no_values[] = 
		{NULL};

	/* meta events */
	static char * meta_seq_no_values[] = 
		{"status", "meta", "number", "tick", "data", NULL};
		
	static char * meta_channel_values[] = 
		{"status", "meta", "channel", "tick", "data", NULL};
		
	static char * meta_port_values[] = 
		{"status", "meta", "port", "tick", "data", NULL};
		
	static char * meta_tempo_values[] = 
		{"status", "meta", "bpm", "tick", "usec_per_quarter", "data", NULL};
		
	static char * meta_smpte_values[] = 
		{"status", "meta", "hour", "minute", "second", "frame", "frame_fract", 
		 "tick", "data", NULL};
		 
	static char * meta_time_sig_values[] = 
		{"status", "meta", 
		 "signature", "tick",
		 "ticks_per_click", "thirtyseconds_per_click", 
		 "numerator", "denominator", "data", NULL};
		 
	static char * meta_key_sig_values[] = 
		{"status", "meta", 
		 "key", "flavour", "tick", "data", NULL};

	/* catch-all event. 
		 let op! unlike all others, data comes before tick here! 
	 */
	static char * meta_values[] = 
		{"status", "meta", 
		 "data", "tick", NULL};

	if (get_type() < M_NOTE_OFF)
		return no_values;
		
	if (get_type() < M_META)
	{
		switch (get_type())
		{
			case M_NOTE_ON:
				return note_values;

			case M_NOTE_OFF:
				return note_off_values;

			case M_NOTE_TOUCH:
				return note_touch_values;

			case M_CONTROL_CHANGE:
				return control_values;

			case M_PROGRAM_CHANGE:
				return program_values;

			case M_CHANNEL_TOUCH:
				return channel_touch_values;

			case M_PITCH_WHEEL:
				return pitch_values;

			case M_SYSEX:
			case M_EOX:
			default:
				return sysex_values;
		}
	}

	switch (get_meta_type())
	{
		case M_META_SEQ_NO:
			return meta_seq_no_values;

		case M_META_CHANNEL:
			return meta_channel_values;
			
		case M_META_PORT:
			return meta_port_values;
			
		case M_META_TEMPO:
			return meta_tempo_values;
			
		case M_META_SMPTE:
			return meta_smpte_values;
			
		case M_META_TIME:
			return meta_time_sig_values;
			
		case M_META_KEY:
			return meta_key_sig_values;
		
		case M_META_TEXT:
		case M_META_COPY:
		case M_META_SEQ_NAME:
		case M_META_INSTRUMENT:
		case M_META_LYRIC:
		case M_META_MARKER:
		case M_META_CUE:
		case M_META_EOT:
		default:
			return meta_values;
	}
	
	return no_values;
}

/*
 *	
 *
 */

PyObject *
m_event_get_value (m_event * e, char * name)
{
	#define INT(x) Py_BuildValue ("i", x)
	#define DBL(x) Py_BuildValue ("d", x)
	
	if (!strcmp (name, "tick"))
		return INT (e->get_tick());
		
	if (!strcmp (name, "status"))
		return INT (e->midi_status());

	if (!strcmp (name, "port"))
		return INT (e->get_port());

	if (!strcmp (name, "type"))
		return INT (e->get_type());

	int type = e->get_type();

	if (type < M_META)
	{
		if (type < M_SYSEX && !strcmp (name, "channel"))
			return INT (e->get_channel());

		switch (type)
		{
			case M_NOTE_OFF:
				if (!strcmp (name, "note"))
					return INT (e->get_note());
				if (!strcmp (name, "velocity"))
					return INT (e->get_velocity_on());
				return NULL;
				
			case M_NOTE_ON:
				if (!strcmp (name, "note"))
					return INT (e->get_note());
				if (!strcmp (name, "velocity"))
					return INT (e->get_velocity_on());
				if (!strcmp (name, "length"))
					return INT (e->get_note_length());
				if (!strcmp (name, "velocity_off"))
					return INT (e->get_velocity_off());
				return NULL;

			case M_NOTE_TOUCH:
				if (!strcmp (name, "note"))
					return INT (e->get_note());
				if (!strcmp (name, "velocity"))
					return INT (e->get_velocity_on());
				return NULL;

			case M_CONTROL_CHANGE:
				if (!strcmp (name, "controller"))
					return INT (e->get_param_0());
				if (!strcmp (name, "value"))
					return INT (e->get_param_1());
				return NULL;

			case M_PROGRAM_CHANGE:
				if (!strcmp (name, "program"))
					return INT (e->get_param_0());
				return NULL;

			case M_CHANNEL_TOUCH:
				if (!strcmp (name, "velocity"))
					return INT (e->get_param_0());
				return NULL;

			case M_PITCH_WHEEL:
				if (!strcmp (name, "pitch"))
					return DBL (e->get_pitch());
				return NULL;

			case M_SYSEX:
			case M_EOX:
				if (!strcmp (name, "length"))
					return INT (e->get_data_buffer()->count);
				if (!strcmp (name, "data"))
					return Py_BuildValue (
						"s#", 
						e->get_data_buffer()->items, 
						e->get_data_buffer()->count
					);
				return NULL;
		}
		return NULL;
	}

	c_char_buffer & data = *e->get_data_buffer();

	/* it is a meta event. */
	if (!strcmp (name, "meta"))
		return INT (e->get_meta_type());
	if (!strcmp (name, "length"))
		return INT (e->get_data_buffer()->count);
	if (!strcmp (name, "data"))
		return Py_BuildValue (
			"s#", 
			e->get_data_buffer()->items, 
			e->get_data_buffer()->count
		);

	switch (e->get_meta_type())
	{
		case M_META_SEQ_NO:
			if (!strcmp (name, "number"))
				return INT ((data[0] << 8) | data[1]);
			return NULL;

		case M_META_TEXT:
		case M_META_COPY:
		case M_META_SEQ_NAME:
		case M_META_INSTRUMENT:
		case M_META_LYRIC:
		case M_META_MARKER:
		case M_META_CUE:
			return NULL;
			
		case M_META_CHANNEL:
			if (!strcmp (name, "channel"))
				return INT (data[0]);
			return NULL;

		case M_META_PORT:
			if (!strcmp (name, "port"))
				return INT (data[0]);
			return NULL;

		case M_META_EOT:
			return NULL;

		case M_META_TEMPO:
			if (!strcmp (name, "bpm"))
				return DBL (e->is_tempo()->bpm());
			if (!strcmp (name, "usec_per_quarter"))
				return INT (e->is_tempo()->usec_per_quarter());
			return NULL;

		case M_META_SMPTE:
			if (!strcmp (name, "hour"))
				return INT (data[0]);
			if (!strcmp (name, "minute"))
				return INT (data[1]);
			if (!strcmp (name, "second"))
				return INT (data[2]);
			if (!strcmp (name, "frame"))
				return INT (data[3]);
			if (!strcmp (name, "frame_fract"))
				return INT (data[4]);
			return NULL;

		case M_META_TIME:
			if (!strcmp (name, "signature"))
				return Py_BuildValue (
					"(ii)", 
					e->is_time()->get_numerator(), 
					e->is_time()->get_denominator()
				);
			if (!strcmp (name, "numerator"))
				return INT (e->is_time()->get_numerator());
			if (!strcmp (name, "denominator"))
				return INT (e->is_time()->get_denominator());
			if (!strcmp (name, "ticks_per_click"))
				return INT (e->is_time()->ticks_per_click());
			if (!strcmp (name, "thirtyseconds_per_click"))
				return INT (e->is_time()->_32s_per_click());
			return NULL;

		case M_META_KEY:
			if (!strcmp (name, "key"))
				return INT ((int) (signed char) data[0]);
			if (!strcmp (name, "flavour"))
				return INT (data[1]);
			return NULL;
	}
	return NULL;

	#undef INT
	#undef NOTE
	#undef DBL
}

/*
 *	convert a python int or string to a note.
 *
 */
 
static int
PyObject_AsNote (PyObject * o, int * note)
{
	if (PyInt_Check (o) || PyFloat_Check (o))
	{
		long i = PyInt_AsLong (o);
		if (i < 0) *note = 0;
		else if (i > 127) *note = 127;
		else *note = i;
			
		return !PyErr_Occurred();
	}

	if (!PyString_Check (o))
	{
		PyErr_SetString (PyExc_TypeError, "not a note.");
		return 0;
	}
	
	char * s = PyString_AS_STRING (o);
	
	*note = m_note_event::string2note (s);

	if (*note != -1)
		return 1;
	
	PyErr_Format (PyExc_TypeError, "'%s' is not a note.", s);
	return 0;
}

/*
 * set a value by name
 *
 */

bool
m_event_set_value (m_event * e, char * name, PyObject * py_value)
{
	/* 
	 * these macros parse the py_value into value or dbl_value.
	 *
	 */
	#if 1
	#define INT(mi,ma)	\
		value = PyInt_AsLong(py_value); \
		if (PyErr_Occurred()) return false; \
		if (value < mi || value > ma) \
		{ \
			PyErr_Format (PyExc_ValueError, "%d <= %s <= %d", mi, name, ma); \
			return false; \
		}	
	#else
	#define INT(mi,ma)	\
		value = PyInt_AsLong(py_value); \
		if (PyErr_Occurred()) return false; \
		if (value < (mi)) value = (mi); \
		else if (value > (ma)) value = (ma); 
	#endif
	
	#define NOTE() \
		if (!PyObject_AsNote (py_value, &value)) return false;

	#define DBL(mi,ma)	\
		dbl_value = PyFloat_AsDouble(py_value); \
		if (PyErr_Occurred()) return false; \
		if (dbl_value < mi || dbl_value > ma) \
		{ \
			PyErr_Format (PyExc_ValueError, name); \
			return false; \
		}	\
		value = (int) (dbl_value * 127.0);
	
	int value = 0;
	double dbl_value = 0.0;
	
	if (!strcmp (name, "status"))
	{
		if (e->is_meta())
		{
			PyErr_SetString (PyExc_NotImplementedError, "status is immutable.");
			return false;
		}

		INT (0x80, 0xEF);

		e->set_type (value & M_TYPE_MASK);
		e->set_channel (value & M_CHANNEL_MASK);
		return true;
	}

	if (!strcmp (name, "port"))
	{
		INT (0, 0xFF);
		e->set_port (value);
		return true;
	}

	if (!strcmp (name, "tick"))
	{
		INT (0, 0x7FFFFFFF);
		e->set_tick (value);
		return true;
	}

	if (e->get_type() < M_SYSEX)
	{
		if (!strcmp (name, "channel"))
		{
			INT (0, 15);
			e->set_channel (value & M_CHANNEL_MASK);
			return true;
		}

		if (!strcmp (name, "type"))
		{
			INT (0, 0xFF);
			e->set_type (value & M_TYPE_MASK);
			return true;
		}

		switch (e->get_type())
		{
			case M_NOTE_OFF:
				if (!strcmp (name, "note"))
					goto SET_NOTE;
				else if (!strcmp (name, "velocity"))
					goto SET_P1;
				else
					return false;
				return true;

			case M_NOTE_ON:
				if (!strcmp (name, "note"))
					goto SET_NOTE;
				else if (!strcmp (name, "velocity"))
					goto SET_P1;
				else if (!strcmp (name, "velocity_off"))
					{INT (0,127); e->set_velocity_off (value);}
				else if (!strcmp (name, "length"))
					{INT (0, M_TICK_MAX); e->set_note_length (value);}
				else
					return false;
				return true;

			case M_NOTE_TOUCH:
				if (!strcmp (name, "note"))
					goto SET_NOTE;
				else if (!strcmp (name, "velocity"))
					goto SET_P1;
				else
					return false;
				return true;

			case M_CONTROL_CHANGE:
				if (!strcmp (name, "controller"))
					goto SET_P0;
				else if (!strcmp (name, "value"))
					goto SET_P1;
				else
					return false;
				return true;

			case M_PROGRAM_CHANGE:
				if (!strcmp (name, "program"))
					goto SET_P0;
				else
					return false;
				return true;

			case M_CHANNEL_TOUCH:
				if (!strcmp (name, "velocity"))
					goto SET_P0;
				else
					return false;
				return true;

			case M_PITCH_WHEEL:
				if (!strcmp (name, "pitch"))
				{
					DBL (-1.0, 1.0);
					e->set_pitch (dbl_value);
				}
				else
					return false;
				return true;
		}
		return false;
	
		SET_NOTE:
			NOTE(); 
			e->set_note (value); 
			return true;
			
		SET_P0:
			INT (0,127); 
			e->set_param_0 (value);
			return true;
			
		SET_P1:
			INT (0,127); 
			e->set_param_1 (value);
			return true;
	}

	/* it is a meta event. */
	if (!strcmp (name, "status"))
	{
		PyErr_Format (PyExc_NotImplementedError, name);
		return false;
	}
	
	if (!strcmp (name, "length"))
	{
		PyErr_Format (PyExc_NotImplementedError, name);
		return false;
	}
	
	#if 0
	if (!strcmp (name, "meta"))
	{
		INT (0, 127);
		e->set_meta_type (value);
		return true;
	}
	#endif

	if (!strcmp (name, "data"))
	{
		byte * new_value;
		int length;

		if (!PyArg_Parse (py_value, "s#", &new_value, &length))
			return false;
		
		if (!e->is_valid_data_length (length))
		{
			PyErr_SetString (PyExc_ValueError, "the data length is invalid.");
			return false;
		}
		e->set_data (new_value, length);
		return true;
	}
	
	if (!e->is_meta())
		return false;

	c_char_buffer & data = *e->get_data_buffer();

	switch (e->get_meta_type())
	{
		case M_META_SEQ_NO:
			if (!strcmp (name, "number"))
			{
				INT (0, 0xFFFF);
				data[0] = value & 0xFF;
				data[1] = value >> 8;
			}
			else 
				return false;
			return true;

		case M_META_TEXT:
		case M_META_COPY:
		case M_META_SEQ_NAME:
		case M_META_INSTRUMENT:
		case M_META_LYRIC:
		case M_META_MARKER:
		case M_META_CUE:
			return false;
			
		case M_META_CHANNEL:
			if (!strcmp (name, "channel"))
				{INT (0,15); data[0] = value;}
			else
				return false;
			return true;

		case M_META_PORT:
			if (!strcmp (name, "port"))
				{INT (0,255); data[0] = value;}
			else
				return false;
			return true;

		case M_META_EOT:
			return false;

		case M_META_TEMPO:
			if (!strcmp (name, "bpm"))
				{DBL (1.0, 999.0); e->is_tempo()->set_bpm (dbl_value);}
			else if (!strcmp (name, "usec_per_quarter"))
				{INT (1, 0xFFFFFF); e->is_tempo()->set_usec_per_quarter (value);}
			else
				return false;
			return true;

		case M_META_SMPTE:
			if (!strcmp (name, "hour"))
				{INT (0,255); data[0] = value;}
			else if (!strcmp (name, "minute"))
				{INT (0,59); data[1] = value;}
			else if (!strcmp (name, "second"))
				{INT (0,59); data[2] = value;}
			else if (!strcmp (name, "frame"))
				{INT (0,29); data[3] = value;}
			else if (!strcmp (name, "frame_fract"))
				{INT (0,99); data[4] = value;}
			else
				return false;
			return true;

		case M_META_TIME:
			if (!strcmp (name, "signature"))
			{
				int n, d;
				if (
					!PyArg_ParseTuple 
					(
						py_value, "ii",
						&n, &d
					)
				) return false;
				if (n < 1 || n > 255)
				{
					PyErr_Format (PyExc_ValueError, "invalid numerator: %d", n);
					return false;
				}
				if (d < 1 || d > 2048)
				{
					PyErr_Format (PyExc_ValueError, "invalid denominator: %d", d);
					return false;
				}
				if (!e->is_time()->set_denominator (n))
					goto WRONG_DENOMINATOR;
				e->is_time()->set_numerator (n);
				return true;
			}
			else if (!strcmp (name, "numerator"))
			{
				INT (1,255); 
				e->is_time()->set_numerator (value);
				return true;
			}
			else if (!strcmp (name, "denominator"))
			{
				INT (1,2048); 
				if (!e->is_time()->set_denominator (value))
				{
				WRONG_DENOMINATOR:
					PyErr_Format (PyExc_ValueError, "invalid denominator: %d", value);
					return false;
				}
			}
			else if (!strcmp (name, "ticks_per_click"))
				{INT (1,65535); e->is_time()->set_ticks_per_click (value);}
			else if (!strcmp (name, "thirtyseconds_per_click"))
				{INT (1,65535); e->is_time()->set_32s_per_click (value);}
			else
				return false;
			return true;

		case M_META_KEY:
			if (!strcmp (name, "key"))
				{INT (-7, 7); data[0] = (byte) value;}
			else if (!strcmp (name, "flavour"))
				{INT (0,1); data[1] = value ? 1 : 0;}
			else
				return false;
			return true;
	}

	return false;

	#undef INT
	#undef DBL
}

