/*
 *	python-midi-event.cc
 *
 */

#include "python-types.h"

/*
 *	m_event.new and del m_event
 *
 */

m_event_object *
m_event_new()
{
	return m_event_new_from_event (NULL);
}

m_event_object *
m_event_new_from_event (m_event * e)
{
	m_event_object * self = PyObject_NEW (m_event_object, &m_event_py_type);
	if (!self)
		return NULL;

	_dbgprintf (DBG_REF, "<m_event_object %x>\n", (int) self);
	
	if (e == NULL)
		e = new m_event;
	
	C_INCREF (e);
	
	self->event = e;

	return self;
}

static void
m_event_del (m_event_object * self)
{
	_dbgprintf (DBG_REF, "<~m_event_object %x>\n", (int) self);
	
	C_DECREF (self->event);

	PyMem_DEL (self);
}

/*
 *	m_event.repr
 *
 */

static PyObject *
m_event_repr (m_event_object * self)
{
	c_char_buffer buffer;

	self->event->repr (&buffer);

	return PyString_FromStringAndSize (buffer.items, buffer.count);
}

/*
 *	m_event.print
 *
 */

static int
m_event_print (m_event_object * self, FILE * fp, int flags)
{
	c_char_buffer buffer;

	self->event->repr (&buffer);

	buffer >> fp;
	
	return 0;
}

static PyObject *
m_event_parameters (m_event_object * self, PyObject * args)
{
	if (args && !PyArg_ParseTuple (args, ""))
		return NULL;
		
	char ** members = self->event->get_value_names();
	
	int n = 0;
	
	while (members[n])
		n++;
		
	PyObject * result = PyList_New (n);

	if (!result)
	{
		PyErr_NoMemory();
		return NULL;
	}
	
	if (result)
	{
		for (int i = 0; i < n; i++)
			PyList_SetItem (result, i, PyString_FromString (members[i]));
		
		if (PyErr_Occurred()) 
		{
			Py_DECREF (result);
			return NULL;
		}
	}
	
	return result;
}

/*
 *	m_event.clone()
 *
 */

static m_event_object *
m_event_copy (m_event_object * self, PyObject * args)
{
	if (PyTuple_Size (args))
	{
		PyErr_SetString (PyExc_ValueError, "This function takes no arguments.");
		return NULL;
	}

	m_event_object * result = m_event_new_from_event (NULL);

	if (result)
		result->event->copy (self->event);

	return result;
}

static PyMethodDef
m_event_methods[] = 
{
	{"copy", (PyCFunction) m_event_copy, METH_VARARGS, 
		"copy()\n"
		"returns an exact copy of this "M_EVENT_NAME"."},
	{"parameters", (PyCFunction) m_event_parameters, METH_VARARGS, 
		"parameters()\n"
		"returns a list of the parameters of this "M_EVENT_NAME" in the order "
		"in which the constructor "M_EVENT_NAME"() evaluates them if you do not "
		"specify keywords with them."},
	{NULL, NULL}
};


/*
 *	m_event.getattr
 *
 */

static PyObject *
m_event_getattr (m_event_object * self, char * name)
{
	if (!strcmp (name, "__members__"))
		return m_event_parameters (self, NULL);

	PyObject * result = m_event_get_value (self->event, name);

	if (result)
		return result;
		
	return Py_FindMethod (m_event_methods, (PyObject *) self, name);
}

/*
 *	m_event.setattr
 *
 */

static int
m_event_setattr (m_event_object * self, char * name, PyObject * value)
{
	if (!value) 
	{
		PyErr_SetString (PyExc_NotImplementedError, "cannot delete an attribute");
		return -1;
	}

	if (m_event_set_value (self->event, name, value))
		return 0;
		
	/* ..._set_value sets only ValueErrors */
	if (!PyErr_Occurred())
		PyErr_SetString (PyExc_AttributeError, name);
		
	return -1;
}

static int
m_event_compare (m_event_object * self, m_event_object * other)
{
	if (other->event == self->event)
		return 0;
	return self->event->compare (other->event);
}


/*
 *	(PyTypeObject) m_event
 *
 */

PyTypeObject 
m_event_py_type =
{
	PyObject_HEAD_INIT (&PyType_Type)
	0, // ob_size 
	M_EVENT_NAME,
	sizeof (m_event_object),
	0, // tp_itemsize 

	// methods 
	(destructor) m_event_del,
	(printfunc) m_event_print,
	(getattrfunc) m_event_getattr,
	(setattrfunc) m_event_setattr,
	(cmpfunc) m_event_compare,
	(reprfunc) m_event_repr,
	0, // tp_as_number 
	0, // tp_as_sequence 
	0, // tp_as_mapping 
	(hashfunc) 0,
	(ternaryfunc) 0,
	(reprfunc) 0,

	// more ... 
	0,0,0,0,

	// __doc__ 
	M_EVENT_NAME" objects are the basic unit of midi messages.",
};

/*
 *	proudly presenting the allround event constructor.
 *
 *	make any event.
 *
 *	not optimal when it comes to performance, maintainable instead.
 *
 */

static m_event_object *
m_any_event_new (int status, int start_arg, PyObject * args, PyObject * keywds)
{
	/* this is the 'assume something useful' parameter for m_event::init() */
	int first_parameter = 255;
	int n_args = PyTuple_Size (args);
	if (status == M_META)
	{
		if (n_args < 1)
		{
			PyErr_SetString (PyExc_ValueError, "need an "M_META_NAME" type.");
			return NULL;
		}

		PyObject * arg = PyTuple_GetItem (args, 0);
	
		if (!PyInt_Check (arg))
		{
			PyErr_SetString (PyExc_ValueError, "need a "M_META_NAME" type.");
			return NULL;
		}
	
		first_parameter = PyInt_AsLong (arg);
		if (first_parameter < 0 || first_parameter > 0x7F)
		{
			PyErr_SetString (PyExc_ValueError, "invalid "M_META_NAME" type.");
			return NULL;
		}

		start_arg++;
	}
	
	/* can now make the event. */
	m_event_object * self = m_event_new();
	if (!self) return self;
	
	m_event * e = self->event;
	e->init (status, first_parameter);

	/* iterate the remaining arguments that have no keyword associated. */
	char ** names = e->get_value_names();
	int i;
	for (i = start_arg; i < n_args && names[i+1]; i++)
	{
		PyObject * value = PyTuple_GetItem (args, i);
		m_event_set_value (e, names[i+1], value);
		
		if (PyErr_Occurred())
		{
			Py_DECREF (self);
			return NULL;
		}
	}
		
	if (!names[i+1] && i < n_args)
	{
		PyErr_SetString (PyExc_ValueError, "too many arguments.\n");
		Py_DECREF (self);
		return NULL;
	}

	/* iterate the arguments that have a keyword associated. */
	if (keywds)
	{
		int n = 0;
		PyObject * name, * value;
		
		while (PyDict_Next (keywds, &n, &name, &value) && !PyErr_Occurred())
		{
			char * attr_name = PyString_AS_STRING (name);
			m_event_set_value (e, attr_name, value);
		}

		if (PyErr_Occurred())
		{
			Py_DECREF (self);
			self = NULL;
		}
	}

	return self;
}

/*
 *	the different constructors
 *
 */

m_event_object *
module_m_note_new (PyObject * self, PyObject * args, PyObject * keywds)
{
	return m_any_event_new (M_NOTE_ON, 0, args, keywds);
}

m_event_object *
module_m_note_off_new (PyObject * self, PyObject * args, PyObject * keywds)
{
	return m_any_event_new (M_NOTE_OFF, 0, args, keywds);
}

m_event_object *
module_m_note_touch_new (PyObject * self, PyObject * args, PyObject * keywds)
{
	return m_any_event_new (M_NOTE_TOUCH, 0, args, keywds);
}

m_event_object *
module_m_controller_new (PyObject * self, PyObject * args, PyObject * keywds)
{
	return m_any_event_new (M_CONTROL_CHANGE, 0, args, keywds);
}

m_event_object *
module_m_program_new (PyObject * self, PyObject * args, PyObject * keywds)
{
	return m_any_event_new (M_PROGRAM_CHANGE, 0, args, keywds);
}

m_event_object *
module_m_channel_touch_new (PyObject * self, PyObject * args, PyObject * keywds)
{
	return m_any_event_new (M_CHANNEL_TOUCH, 0, args, keywds);
}

m_event_object *
module_m_pitch_new (PyObject * self, PyObject * args, PyObject * keywds)
{
	return m_any_event_new (M_PITCH_WHEEL, 0, args, keywds);
}

m_event_object *
module_m_sysex_new (PyObject * self, PyObject * args, PyObject * keywds)
{
	return m_any_event_new (M_SYSEX, 0, args, keywds);
}

m_event_object *
module_m_meta_new (PyObject * self, PyObject * args, PyObject * keywds)
{
	return m_any_event_new (M_META, 0, args, keywds);
}

/*
 *	all events constructor.
 *
 */
 
m_event_object *
module_m_event_new (PyObject * self, PyObject * args, PyObject * keywds)
{
	if (!PyTuple_Size (args))
	{
		PyErr_SetString (PyExc_ValueError, "Need a "M_EVENT_NAME" type.");
		return NULL;
	}

	PyObject * arg = PyTuple_GetItem (args, 0);

	if (!PyInt_Check (arg))
	{
		PyErr_SetString (PyExc_ValueError, "need a "M_EVENT_NAME" type.");
		return NULL;
	}

	int status = PyInt_AsLong (arg);
	if (status < 0 || status > 0xFF)
	{
		PyErr_SetString (PyExc_ValueError, "invalid "M_EVENT_NAME" type.");
		return NULL;
	}

	return m_any_event_new (status, 1, args, keywds);
}

