/*
 *	python-midi-io.cc
 *
 */

#include "python-types.h"

static int
m_io_set_on_input (m_io_object * self, PyObject * value);

/*
 *	m_io.new and del m_io
 *
 */

static m_io_object *
m_io_new()
{
	m_io_object * self = PyObject_NEW (m_io_object, &m_io_py_type);
	if (!self) return NULL;
		
	_dbgprintf (DBG_REF, "<m_io_object %x>\n", (int) self);

	self->io = new m_io (new m_event_debugger ("in"));
	C_INCREF (self->io);

	self->on_input = new py_callback();
	C_INCREF (self->on_input);
	
	return self;
}

static void
m_io_del (m_io_object * self)
{
	_dbgprintf (DBG_REF, "<~m_io_object %x>\n", (int) self);

	C_DECREF (self->on_input);
	C_DECREF (self->io);

	PyMem_DEL (self);
}

/*
 *	m_io.add()
 *
 */

static
PyObject *
m_io_add (m_io_object * self, PyObject * args)
{
	m_port_object * port_object;

	if (!PyArg_ParseTuple 
		(args, "O!;"M_PORT_NAME" object",
			&m_port_py_type, &port_object
		)
	) return NULL;
		
	_dbgprintf (0, "port %s has %d references.\n", port_object->port->get_connector_name(), port_object->ob_refcnt);

	int i;
	Py_INCREF ((PyObject *) port_object);
	Py_BEGIN_ALLOW_THREADS
	i = self->io->add_port (port_object->port);
	Py_END_ALLOW_THREADS
	Py_DECREF ((PyObject *) port_object);

	if (i < 0)
	{
		PyErr_SetString (PyExc_ValueError, "cannot add this "M_PORT_NAME".");
		return NULL;
	}

	return Py_BuildValue ("i", i);
}

/*
 *	m_io.remove()
 *
 */

static
PyObject *
m_io_remove (m_io_object * self, PyObject * args)
{
	m_port_object * port_object;

	if (!PyArg_ParseTuple 
		(
			args, "O!;"M_PORT_NAME" object",
			&m_port_py_type, &port_object
		)
	) return NULL;
		
	Py_INCREF ((PyObject *) port_object);
	if (!self->io->remove_port (port_object->port))
	{
		Py_DECREF ((PyObject *) port_object);
		PyErr_SetString (PyExc_IndexError, "cannot remove this "M_PORT_NAME".");
		return NULL;
	}

	Py_DECREF ((PyObject *) port_object);
	
	Py_INCREF (Py_None);
	return Py_None;
}

/*
 *	m_io.index()
 *
 */

static
PyObject *
m_io_index (m_io_object * self, PyObject * args)
{
	m_port_object * port_object;
	
	if (!PyArg_ParseTuple 
		(
			args, "O!;"M_PORT_NAME" object", 
			&m_port_py_type, &port_object
		)
	)	return NULL;
	
	Py_INCREF ((PyObject *) port_object);
	int i = self->io->ports.find (port_object->port);
	Py_DECREF ((PyObject *) port_object);

	if (i < 0)
	{
		PyErr_SetObject (PyExc_IndexError, (PyObject *) port_object);
		return NULL;
	}

	return PyInt_FromLong (i);
}

/*
 *	m_io.clear()
 *
 */

static PyObject *
m_io_clear (m_io_object * self, PyObject * args)
{
	if (!PyArg_ParseTuple (args, ""))
		return NULL;
	
	Py_BEGIN_ALLOW_THREADS
	self->io->clear();
	Py_END_ALLOW_THREADS

	Py_INCREF (Py_None);
	return Py_None;
}

/*
 *	m_io as sequence
 *
 */

static int
m_io_len (m_io_object * self)
{
	return self->io->ports.count;
}

/*
 *	(m_port) = m_io[i]
 *
 */

static m_port_object *
m_io_item (m_io_object * self, int i)
{
	if (i < 0 || i >= m_io_len (self))
	{
		PyErr_SetString (PyExc_IndexError, "out of range.");
		return NULL;
	}

	m_connector * port = self->io->port (i);

	return m_port_new_from_port (port);
}

/*
 *	m_io[i] = (m_port)
 *
 */

static int
m_io_ass_item (m_io_object * self, int i, m_port_object * port_object)
{
	if (i < 0 || i >= m_io_len (self))
	{
		PyErr_SetString (PyExc_IndexError, "out of range.");
		return -1;
	}

	/* remove an event */
	if (port_object == NULL)
	{
		Py_BEGIN_ALLOW_THREADS
		self->io->port(i)->reset();
		self->io->remove_port (self->io->port(i));
		Py_END_ALLOW_THREADS
		return 0;
	}
	
	Py_INCREF (port_object);
	
	if (!m_port_Check (port_object))
	{
		Py_DECREF (port_object);
		PyErr_SetString (PyExc_ValueError, "need a "M_PORT_NAME".");
		return -1;
	}
	
	Py_BEGIN_ALLOW_THREADS
	self->io->port(i)->reset();
	self->io->replace_port (i, port_object->port);
	Py_END_ALLOW_THREADS

	Py_DECREF (port_object);

	return 0;
}

/*
 *	m_io
 *
 */

PySequenceMethods 
m_io_as_sequence = 
{
	(inquiry) m_io_len, //sq_length
	(binaryfunc) 0, //sq_concat
	(intargfunc) 0, //sq_repeat
	(intargfunc) m_io_item, //sq_item
	(intintargfunc) 0, //sq_slice
	(intobjargproc) m_io_ass_item, //sq_ass_item
	(intintobjargproc) 0, //sq_ass_slice
	(objobjproc) 0, //sq_contains
};

/*
 *	m_io.write
 *
 */

static 
PyObject *
m_io_write (m_io_object * self, PyObject * args)
{
	m_event_object * event_object;
	
	if (!PyArg_ParseTuple 
		(
			args, "O!", 
			&m_event_py_type, &event_object
		)
	) return NULL;

	if (!m_io_len (self))
	{
		PyErr_SetString (PyExc_IOError, "No ports.");
		return NULL;
	}

	Py_INCREF (event_object);
	
	/* since m_io can call back into python, 
	   we need to release the python lock here. 
	*/
	Py_BEGIN_ALLOW_THREADS
	self->io->event_in (event_object->event);
	Py_END_ALLOW_THREADS
	
	Py_DECREF (event_object);

	Py_INCREF (Py_None);
	return Py_None;
}


/*
 *	m_io.methods
 *
 */

static PyMethodDef
m_io_methods[] = 
{
	{"add", (PyCFunction) m_io_add, METH_VARARGS, 
		"add (port)\n"
		"adds a "M_PORT_NAME" to the list of "M_PORT_NAME"s and returns the index "
		"at which it was inserted."},
	{"remove", (PyCFunction) m_io_remove, METH_VARARGS, 
		"remove (port)\n"
		"removes a "M_PORT_NAME" from the list of "M_PORT_NAME"s."},
	{"index", (PyCFunction) m_io_index, METH_VARARGS, 
		"index (port)\n"
		"returns the index of "M_PORT_NAME" or raises an error."},
	{"clear", (PyCFunction) m_io_clear, METH_VARARGS, 
		"clear()\n"
		"removes all "M_PORT_NAME"s from the "M_IO_NAME" object."},
	{"write", (PyCFunction) m_io_write, METH_VARARGS, 
		"write (event)\n"
		"writes an "M_EVENT_NAME" to the port that is specified by the "M_EVENT_NAME"."},
	{NULL, NULL}
};

/*
 *	m_io.getattr
 *	
 */

static PyObject *
m_io_getattr (m_io_object * self, char * name)
{
 	if (!strcmp (name, "on_input"))
	{
		if (!self->on_input->callback)
		{
			Py_INCREF (Py_None);
			return Py_None;
		}
		
		Py_INCREF (self->on_input->callback);
		return self->on_input->callback;
	}

 	if (!strcmp (name, "sync"))
		return (PyObject *) m_sync_new_from_io (self->io);

 	if (!strcmp (name, "clock"))
		return (PyObject *) m_clock_new_from_clock (self->io->clock);

	if (!strcmp (name, "__members__"))
	{
		PyObject * list = PyList_New (3);
		if (!list)
			return NULL;

		PyList_SetItem (list, 0, PyString_FromString ("on_input"));
		PyList_SetItem (list, 1, PyString_FromString ("sync"));
		PyList_SetItem (list, 2, PyString_FromString ("clock"));
		
		if (PyErr_Occurred()) 
		{
			Py_DECREF(list);
			list = NULL;
		}
		return list;
	}

	return Py_FindMethod (m_io_methods, (PyObject *) self, name);
}

/*
 *	set the on_input Python callback.
 *
 */

static int
m_io_set_on_input (m_io_object * self, PyObject * value)
{
	if (value == Py_None || value == NULL)
	{
		if (self->on_input->callback)
		{
			self->io->set_output (NULL);
			self->on_input->set_callback (NULL);
		}
		return 0;
	}

	/* sets PyErr_ appropriately */
	if (!self->on_input->set_callback (value))
		return -1;
	
	self->io->set_output (self->on_input);
	return 0;
}

/*
 *	 m_io.setattr
 *	
 */

static int
m_io_setattr (m_io_object * self, char * name, PyObject * value)
{
	if (!value) 
	{
		PyErr_SetString (PyExc_TypeError, "attribute deletion is not supported");
		return -1;
	}
						
 	if (!strcmp (name, "on_input"))
		return m_io_set_on_input (self, value);

	PyErr_SetString (PyExc_AttributeError, name);
	return -1;
}


PyTypeObject 
m_io_py_type =
{
	PyObject_HEAD_INIT (&PyType_Type)
	0, // ob_size 
	M_IO_NAME,
	sizeof (m_io_object),
	0, // tp_itemsize 

	// methods 
	(destructor) m_io_del,
	(printfunc) 0,
	(getattrfunc) m_io_getattr,
	(setattrfunc) m_io_setattr,
	(cmpfunc) 0,
	(reprfunc) 0,
	0, // tp_as_number 
	&m_io_as_sequence, // tp_as_sequence 
	0, // tp_as_mapping 
	(hashfunc) 0,
	(ternaryfunc) 0,
	(reprfunc) 0,

	// more ... 
	
	0,0,0,0,

	// __doc__ 
	
	"The "M_IO_NAME" object handles "M_EVENT_NAME" I/O and timing."
};

/*
 *	new m_io
 *
 */

m_io_object *
module_m_io_new (PyObject * self, PyObject * args)
{
	if (!PyArg_ParseTuple (args, ""))
		return NULL;
		
	return m_io_new();
}


