/*
 *	python-midi-event.cc
 *
 */

#include "python-types.h"

static int m_part_len (m_part_object *);

/*
 *	m_part.new / .del
 *
 */

static m_part_object *
m_part_new (m_part * part)
{
	m_part_object * self = PyObject_NEW (m_part_object, &m_part_py_type);
	if (!self)
		return NULL;
	
	_dbgprintf (DBG_REF, "<m_part_object %x>\n", (int) self);

	C_INCREF (part);
	self->part = part;
	
	return self;
}

/*
 * = m_part[x:y]
 *
 */
 
m_part_object *
m_part_slice_new (m_part_object * self, int start, int end)
{
	/* XXX a lot of this code is unnecessary i think. */
	if (start < 0)
		start = 0;
	else if (start > self->part->events.count)
		start = self->part->events.count;
	if (end < start)
		end = start;
	if (end >= self->part->events.count)
		end = self->part->events.count - 1;

	m_part * slice = new m_part (self->part, start, end);

	self = m_part_new (slice);
	if (!self) 
	{
		delete slice;
		return (m_part_object *) PyErr_NoMemory();
	}
	
	_dbgprintf (DBG_REF, "<m_part_slice [%d:%d] %p>\n", start, end, self);

	return self;
}

m_part_object *
m_part_new_from_song (m_song_object * song_object, int i)
{
	return m_part_new (song_object->song->part(i));
}

static void
m_part_del (m_part_object * self)
{
	_dbgprintf (DBG_REF, "<~m_part_object %x>\n", (int) self);

	C_DECREF (self->part);

	PyMem_DEL (self);
}

/*
 *	_m_part_add
 *
 *	given a python object in arg, add this if arg == event
 *	or recursively add members if arg == part or arg == [part,...]
 */
 
static bool
_m_part_add (m_part_object * self, PyObject * arg, m_tick tick)
{
	if (m_event_Check (arg))
	{
		m_event * e = ((m_event_object *) arg)->event;
		if (tick >= 0)
		{
			e = new m_event (e);
			e->set_tick (tick);
		}

		self->part->lock_write();
		self->part->add_event (e);
		self->part->unlock_write();
	}
	else if (m_part_Check (arg))
	{
		m_part_object * other = (m_part_object *) arg;

		other->part->lock_read();
		
		int n = other->part->events.count;
		
		if (n)
		{
			self->part->lock_write();
			if (tick >= 0)
			{
				/* try and be smart about the tick offset */
				tick -= other->part->event(0)->get_tick();
				
				for (int i = 0; i < n; i++)
				{
					m_event * e = new m_event (other->part->event (i));
					e->set_tick (e->get_tick() + tick);
					self->part->add_event (e);
				}
			}
			else
			{
				for (int i = 0; i < n; i++)
					self->part->add_event (other->part->event(i));
			}
			self->part->unlock_write();
		}
		other->part->unlock_read();
	}
	else if (PyList_Check (arg))
	{
		/* recursively parse PyList */
		for (int i = 0; i < PyList_Size (arg); i++)
		{
			if (!_m_part_add (self, PyList_GetItem (arg, i), tick))
				return false;
		}
	}
	else if (PyTuple_Check (arg))
	{
		/* recursively parse PyTuple */
		for (int i = 0; i < PyTuple_Size (arg); i++)
		{
			if (!_m_part_add (self, PyTuple_GetItem (arg, i), tick))
				return false;
		}
	}
	else
	{
		return false;
	}
	return true;
}

static PyObject *
m_part_add (m_part_object * self, PyObject * args, PyObject * keywds)
{
	static char * keywords[] = {"object", "tick", NULL};
	PyObject * object;
	int tick = -1;
	
	if (!PyArg_ParseTupleAndKeywords 
		(
			args, keywds,
			"O|i;"M_EVENT_NAME"[,tick=0]", keywords,
			&object, &tick
		)
	) return NULL;

	Py_INCREF (object);
	if (!_m_part_add (self, object, tick))
	{
		Py_DECREF (object);
		PyErr_SetString (PyExc_TypeError, "Cannot add this object.");
		
		return NULL;
	}
	Py_DECREF (object);

	Py_INCREF (Py_None);
	return Py_None;
}

/*
 *	m_part.remove
 *
 */

static PyObject *
m_part_remove (m_part_object * self, PyObject * args)
{
	m_event_object * event_object;
	
	if (!PyArg_ParseTuple 
		(
			args, 
			"O!;"M_EVENT_NAME, 
			&m_event_py_type, &event_object
		)
	) return NULL;

	Py_INCREF (event_object);

	if (self->part->events.find (event_object->event) < 0)
	{
		PyErr_SetString (PyExc_ValueError, 
			"this "M_EVENT_NAME" is not in this "M_PART_NAME".");
		Py_DECREF (event_object);

		return NULL;
	}
	
	self->part->lock_write();
	int done = self->part->remove_event (event_object->event);
	self->part->unlock_write();

	PyObject * result;
	
	/* better to check again? */
	if (!done)
	{
		PyErr_SetString (PyExc_AssertionError, 
			"this "M_EVENT_NAME" is not in this "M_PART_NAME".");
		result = NULL;
	}
	else
	{
		Py_INCREF (Py_None);
		result = Py_None;
	}

	Py_DECREF (event_object);

	return result;
}

/*
 *	m_part.index
 *
 */

static PyObject *
m_part_index (m_part_object * self, PyObject * args)
{
	m_event_object * event_object;
	int tick = -1;
	
	if (!PyArg_ParseTuple 
		(
			args, 
			"O!;"M_EVENT_NAME, 
			&m_event_py_type, &event_object, &tick
		)
	) return NULL;

	Py_INCREF (event_object);

	self->part->lock_read();
	int index = self->part->events.find (event_object->event);
	self->part->unlock_read();

	Py_DECREF (event_object);

	if (index < 0)
	{
		PyErr_SetString (PyExc_ValueError, 
			"this "M_EVENT_NAME" is not in this "M_PART_NAME".");
		return NULL;
	}

	return Py_BuildValue ("i", index);
}

/*
 *	m_part.clear
 *
 */

static PyObject *
m_part_clear (m_part_object * self, PyObject * args)
{
	self->part->lock_write();
	self->part->clear();
	self->part->unlock_write();
	
	Py_INCREF (Py_None);
	return Py_None;
}

/*
 *	m_part.sort
 *
 */

static PyObject *
m_part_sort (m_part_object * self, PyObject * args)
{
	self->part->lock_write();
	self->part->sort();
	self->part->unlock_write();
	
	Py_INCREF (Py_None);
	return Py_None;
}

/*
 *	m_part.iterate
 *
 */

static PyObject *
m_part_iterate (m_part_object * self, PyObject * args)
{
	/* return right away if this part is empty. */
	if (!m_part_len (self))
	{
		Py_INCREF (Py_None);
		return Py_None;
	}
	
	PyObject * callback = PyTuple_GetItem (args, 0);
	if (!callback)
	{
		PyErr_SetString (PyExc_TypeError, 
			"give me a callback.");
		return NULL;
	}
	if (!PyCallable_Check (callback))
	{
		PyErr_SetString (PyExc_TypeError, 
			"first parameter must be callable.");
		return NULL;
	}
	
	/* make a new tuple for the callback arguments. */

	int len = PyTuple_Size (args);
	PyObject * passing_in = PyTuple_New (len);
	
	if (!passing_in)
	{
		PyErr_SetString (PyExc_MemoryError, "oops.");
		return NULL;
	}
	
	Py_INCREF (passing_in);
	
	m_event_object * 
	event_object = m_event_new_from_event (self->part->event(0));
		
	Py_INCREF ((PyObject *) event_object);

	PyTuple_SET_ITEM (passing_in, 0, (PyObject *) event_object);

	for (int i = 1; i < len; i++)
	{
		PyObject * o = PyTuple_GET_ITEM (args, i);

		Py_INCREF (o);
		PyTuple_SET_ITEM (passing_in, i, o);
	}
	
	Py_INCREF (callback);
	

	/* the iteration loop */

	self->part->lock_read();
	
	bool tick_changed = 0;
	
	int i = 0;
	while (i < self->part->events.count)
	{
		/* exchange the event in the iterator event_object */
		C_DECREF (event_object->event);
		event_object->event = self->part->event (i);
		C_INCREF (event_object->event);
			
		self->part->unlock_read();
		
		m_tick tick = event_object->event->get_tick();

		int count = self->part->events.count;

		if (!PyEval_CallObject (callback, passing_in) || PyErr_Occurred())
		{
			Py_DECREF (passing_in);
			Py_DECREF (callback);
			return NULL;
		}

		self->part->lock_read();

		tick_changed |= (tick != event_object->event->get_tick());
		
		if (self->part->events.count == count)
			i++;
	}
	
	self->part->unlock_read();
	
	Py_DECREF (passing_in);
	Py_DECREF (callback);

	if (tick_changed)
		m_part_sort (self, NULL);

	Py_INCREF (Py_None);
	return Py_None;
}

/*
 *	m_part as sequence
 *
 */

static int
m_part_len (m_part_object * self)
{
	return self->part->events.count;
}

/*
 *	(m_event) = m_part[i]
 *
 */

static m_event_object *
m_part_item (m_part_object * self, int i)
{
	if (i < 0 || i >= m_part_len (self))
	{
		PyErr_SetObject (PyExc_IndexError, PyInt_FromLong (i));
		return NULL;
	}

	return m_event_new_from_event (self->part->event (i));
}

/*
 *	m_part[i] = (m_event)
 *
 */

static int
m_part_ass_item (m_part_object * self, int i, m_event_object * event_object)
{
	if (i < 0 || i >= m_part_len (self))
	{
		PyErr_SetObject (PyExc_IndexError, PyInt_FromLong (i));
		return -1;
	}

	/* remove an event */
	if (event_object == NULL)
	{
		self->part->remove_event (self->part->event (i));
		return 0;
	}
	
	/* add an event */
	if (!m_event_Check (event_object))
	{
		PyErr_SetString (PyExc_TypeError, "not an "M_EVENT_NAME);
		return -1;
	}

	self->part->events.replace (i, event_object->event);

	return 0;
}

/*
 *	m_part == x ?
 *
 *	return 0 if equal, -1 if less events, +1 if more
 */

static int
m_part_compare (m_part_object * self, m_part_object * other)
{
	if (other->part == self->part)
		return 0;
	return self->part->events.count > other->part->events.count ? 1 : -1;
}

/*
 *	m_part_methods
 *
 */

static PyMethodDef
m_part_methods[] = 
{
	{"add", (PyCFunction) m_part_add, METH_VARARGS|METH_KEYWORDS, 
	 "add (event[, tick])\n"
	 "adds an "M_EVENT_NAME" to the "M_PART_NAME". The tick argument may be "
	 "omitted in which case the "M_EVENT_NAME"s tick will be used. The "
	 M_PART_NAME" will insert a copy of the "M_EVENT_NAME" and return "
	 "a reference to the new "M_EVENT_NAME"."},
	{"remove", (PyCFunction) m_part_remove, METH_VARARGS, 
	 "remove (event)\n"
	 "removes an "M_EVENT_NAME" from the "M_PART_NAME". An error is raised if "
	 "the "M_EVENT_NAME" is not in the "M_PART_NAME"."},
	{"index", (PyCFunction) m_part_index, METH_VARARGS, 
	 "index (event)\n"
	 "returns the index at which an "M_EVENT_NAME" resides in a "M_PART_NAME". "
	 "An error is raised if the "M_EVENT_NAME" is not in the "M_PART_NAME"."},
	{"clear", (PyCFunction) m_part_clear, METH_VARARGS, 
	 "clear()\n"
	 "removes all "M_EVENT_NAME"s from a "M_PART_NAME"."},
	{"iterate", (PyCFunction) m_part_iterate, METH_VARARGS, 
	 "iterate (func[, ...])\n"
	 "calls 'func' with every "M_EVENT_NAME" in the "M_PART_NAME". Any values "
	 "passed in after 'func' are in turn passed to 'func' by the iterator after "
	 "the "M_EVENT_NAME", which is always the first argument to 'func'."},
	{"sort", (PyCFunction) m_part_sort, METH_VARARGS, 
	 "sort()\n"
	 "sorts the "M_PART_NAME". Use this if you modified "M_EVENT_NAME" ticks."},
	{NULL, NULL}
};


/*
 *	m_part.getattr
 *
 */

static
PyObject *
m_part_getattr (m_part_object * self, char * name)
{
	if (!strcmp (name, "division"))
		return Py_BuildValue ("i", self->part->division);
	if (!strcmp (name, "channel"))
		return Py_BuildValue ("i", self->part->channel);
	if (!strcmp (name, "port"))
		return Py_BuildValue ("i", self->part->port);

	if (!strcmp (name, "first_non_meta"))
	{
		self->part->lock_read();
		m_event * e = self->part->find_non_meta (0);
		self->part->unlock_read();
		
		if (!e)
			return Py_BuildValue ("i", 0);
		return Py_BuildValue ("i", e->get_tick());
	}

	if (!strcmp (name, "_cursors_"))
		return Py_BuildValue ("i", self->part->cursors.count);

	if (!strcmp (name, "name"))
	{
		m_event * e = self->part->find_meta (M_META_SEQ_NAME);
		if (!e)
		{
			Py_INCREF (Py_None); 
			return Py_None;
		}
		c_char_buffer & data = *e->get_data_buffer();
		return Py_BuildValue ("s#", data.items, data.count);
	}

	if (!strcmp (name, "__members__"))
	{
		PyObject * result = PyList_New (5);
		if (result)
		{
			PyList_SetItem (result, 0, PyString_FromString ("name"));
			PyList_SetItem (result, 1, PyString_FromString ("division"));
			PyList_SetItem (result, 2, PyString_FromString ("channel"));
			PyList_SetItem (result, 3, PyString_FromString ("port"));
			PyList_SetItem (result, 4, PyString_FromString ("first_non_meta"));
		
			if (PyErr_Occurred()) 
			{
				Py_DECREF (result);
				result = NULL;
			}
		}
		return result;
	}
	
	return Py_FindMethod (m_part_methods, (PyObject *) self, name);
}	

/*
 *	m_part.setattr
 *
 */

static int
m_part_setattr (m_part_object * self, char * name, PyObject * arg)
{
	if (!strcmp (name, "name"))
	{
		char * s;
		int length;
		if (!PyArg_Parse (arg, "s#", &s, &length))
			return -1;

		m_event * e = self->part->find_meta (M_META_SEQ_NAME);

		if (!e)
		{
			e = new m_event (M_META, M_META_SEQ_NAME);
			self->part->add_event (e);
		}

		e->set_data (s, length);

		return 0;
	}
		
	int new_value;
	if (!PyArg_Parse (arg, "i", &new_value))
		return 1;
		

	if (!strcmp (name, "start"))
		;
	else if (!strcmp (name, "division"))
		self->part->division = new_value;
	else if (!strcmp (name, "port"))
		self->part->set_port (new_value);
	else if (!strcmp (name, "channel"))
		self->part->set_channel (new_value);
	else
	{
		PyErr_SetString (PyExc_AttributeError, name);
		return -1;
	}
	
	return 0;
}	

/*
 *	m_part.repr
 *
 */

static PyObject *
m_part_repr (m_part_object * self)
{
	self->part->lock_read();

	c_char_buffer buffer;
	
	self->part->repr (&buffer, '\n');

	self->part->unlock_read();

	return PyString_FromStringAndSize (buffer.items, buffer.count);
}

/*
 *	m_part.print
 *
 */

static int
m_part_print (m_part_object * self, FILE * fp, int flags)
{
	self->part->lock_read();

	c_char_buffer buffer;
	
	self->part->repr (&buffer, '\n');

	self->part->unlock_read();

	buffer >> fp;

	return 0;
}

/*
 *	m_part
 *
 */

PySequenceMethods 
m_part_as_sequence = 
{
	(inquiry) m_part_len, //sq_length
	(binaryfunc) 0, //sq_concat
	(intargfunc) 0, //sq_repeat
	(intargfunc) m_part_item, //sq_item
	(intintargfunc) m_part_slice_new, //sq_slice
	(intobjargproc) m_part_ass_item, //sq_ass_item
	(intintobjargproc) 0, //sq_ass_slice
	(objobjproc) 0, //sq_contains
};

PyTypeObject 
m_part_py_type =
{
	PyObject_HEAD_INIT (&PyType_Type)
	0, // ob_size 
	M_PART_NAME,
	sizeof (m_part_object),
	0, // tp_itemsize 

	// methods 
	
	(destructor) m_part_del,
	(printfunc) m_part_print,
	(getattrfunc) m_part_getattr,
	(setattrfunc) m_part_setattr,
	(cmpfunc) m_part_compare,
	(reprfunc) m_part_repr,
	0, // tp_as_number 
	&m_part_as_sequence, // tp_as_sequence 
	0, // tp_as_mapping 
	(hashfunc) 0,
	(ternaryfunc) 0,
	(reprfunc) 0,

	// more ... 
	
	0,0,0,0,

	// __doc__ 
	
	"A "M_PART_NAME" object is a list of "M_EVENT_NAME"s sorted by MIDI tick."
};

/*
 *	make a part from a tuple of events, parts, part slices & keywds
 *
 */

m_part_object *
module_m_part_new (PyObject * self, PyObject * args, PyObject * keywds)
{
	m_part * part = new m_part();
	m_part_object * part_object = m_part_new (part);
	if (!part_object) 
	{
		delete part; 
		return (m_part_object *) PyErr_NoMemory();
	}

	if (PyTuple_Size (args))
	{
		for (int i = 0; i < PyTuple_Size (args); i++)
		{
			PyObject * arg = PyTuple_GetItem (args, i);
			if (!_m_part_add (part_object, arg, 0))
			{
				delete part; 
				PyErr_SetObject (PyExc_TypeError, arg);
				return NULL;
			}
		}
	}
	if (keywds)
	{
		int n = 0;
		PyObject * name, * value;
		
		while (PyDict_Next (keywds, &n, &name, &value) && !PyErr_Occurred())
		{
			char * attr_name = PyString_AS_STRING (name);
			m_part_setattr (part_object, attr_name, value);
		}

		if (PyErr_Occurred())
		{
			Py_DECREF (part_object);
			return NULL;
		}
	}
	return part_object;	
}

