/*
 *	python-midi-event.cc
 *
 */

#include <sys/fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
			 
#include "alsa-midi-io.h"

#include "python-types.h"

/*
 *	m_port.new / .del
 *
 */

m_port_object *
m_port_new_from_port (m_connector * port)
{
	m_port_object * self = PyObject_NEW (m_port_object, &m_port_py_type);
	if (!self) 
	{
		PyErr_NoMemory();
		return NULL;
	}
	
	_dbgprintf (DBG_REF, "<m_port_object %x>\n", (int) self);

	self->port = port;

	C_INCREF (self->port);
	
	return self;
}

static void
m_port_del (m_port_object * self)
{
	_dbgprintf (DBG_REF, "<~m_port_object %x>\n", (int) self);

	C_DECREF (self->port);
	
	PyMem_DEL (self);
}

/*
 *	m_port.reset
 *
 */

static PyObject *
m_port_reset (m_port_object * self, PyObject * args)
{
	if (!PyArg_ParseTuple
		(args, "")
	) return NULL;
	
	self->port->reset();

	Py_INCREF (Py_None);
	return Py_None;
}

/*
 *	m_port.write
 *
 */

static PyObject *
m_port_write (m_port_object * self, PyObject * args)
{
	m_event_object * event_object;
	
	if (!PyArg_ParseTuple
		(args, "O!;"M_EVENT_NAME" object to send",
		 &m_event_py_type, &event_object)
	) return NULL;
	
	self->port->event_in (event_object->event);

	Py_INCREF (Py_None);
	return Py_None;
}

/*
 *	m_port.state
 *
 */

static PyObject *
m_port_state (m_port_object * self, PyObject * args)
{
	if (!PyArg_ParseTuple
		(args, "")
	) return NULL;
	
	m_state * state = self->port->get_state();

	if (!state)
	{
		Py_INCREF (Py_None);
		return Py_None;
	}

	PyObject * result = PyList_New (16);
	
	for (int channel = 0; channel < 16; channel++)
	{
		PyObject * note_list = PyDict_New();
		PyObject * ctrl_list = PyDict_New();
		PyObject * pres_list = PyDict_New();
		for (int i = 0; i < 128; i++)
		{
			if (state->note[channel][i])
				PyDict_SetItem (
					note_list, 
					PyInt_FromLong (i), 
					PyInt_FromLong (state->note[channel][i]
				)
			);
			if (state->ctrl[channel][i] < 0x80)
				PyDict_SetItem (
					ctrl_list, 
					PyInt_FromLong (i), 
					PyInt_FromLong (state->ctrl[channel][i]
				)
			);
			if (state->ctrl[channel][i] < 0x80)
				PyDict_SetItem (
					ctrl_list, 
					PyInt_FromLong (i), 
					PyInt_FromLong (state->ctrl[channel][i]
				)
			);
		}
		
		PyObject * chan_dict = PyDict_New();
		PyDict_SetItem (chan_dict, PyString_FromString ("pitch"), PyFloat_FromDouble (state->pitch[channel]));
		PyDict_SetItem (chan_dict, PyString_FromString ("notes"), note_list);
		PyDict_SetItem (chan_dict, PyString_FromString ("controllers"), ctrl_list);
		PyDict_SetItem (chan_dict, PyString_FromString ("pressure"), pres_list);

		PyList_SetItem (result, channel, chan_dict);
	}

	return result;
}

/*
 *	m_port_methods
 *
 */

static PyMethodDef
m_port_methods[] = 
{
	{"write", (PyCFunction) m_port_write, METH_VARARGS, 
	 "write (event)\n"
	 "write an "M_EVENT_NAME" to this port."},
	{"reset", (PyCFunction) m_port_reset, METH_VARARGS, 
	 "reset()\n"
	 "resets the port state: silences all notes on this port and "
	 "pulls pitch back to 0.0."},
	{"state", (PyCFunction) m_port_state, METH_VARARGS, 
	 "state()\n"
	 "returns a list of 16 dictionaries describing for each channel: "
	 "the current pitchbend value, "
	 "which (and how many) notes are sounding, the last controller values "
	 "and the last NoteTouch values."},
	{NULL, NULL}
};

/*
 *	m_port.getattr
 *
 */

static
PyObject *
m_port_getattr (m_port_object * self, char * name)
{
	if (!strcmp (name, "mode"))
	{
		char buffer[] = {0,0,0,0};
		char * b = buffer;
		if (self->port->can_read())
			*b++ = 'r';
		if (self->port->can_write())
			*b++ = 'w';
		return PyString_FromString (buffer);
	}
	
	if (!strcmp (name, "__members__"))
	{
		PyObject * result = PyList_New (1);
		if (result)
		{
			PyList_SetItem (result, 0, PyString_FromString ("mode"));
		
			if (PyErr_Occurred()) 
			{
				Py_DECREF (result);
				result = NULL;
			}
		}
		return result;
	}
	
	return Py_FindMethod (m_port_methods, (PyObject *) self, name);
}	

/*
 *	m_port.setattr
 *
 */

static int
m_port_setattr (m_port_object * self, char * name, PyObject * arg)
{
	PyErr_SetString (PyExc_AttributeError, name);
	return -1;
}	


static PyObject *
m_port_repr (m_port_object * self)
{
	return PyString_FromString (self->port->get_connector_name());
}

/*
 * 	the python port type
 *
 */

PyTypeObject 
m_port_py_type =
{
	PyObject_HEAD_INIT (&PyType_Type)
	0, // ob_size 
	M_PORT_NAME,
	sizeof (m_port_object),
	0, // tp_itemsize 

	// methods 
	
	(destructor) m_port_del,
	(printfunc) 0,
	(getattrfunc) m_port_getattr,
	(setattrfunc) m_port_setattr,
	(cmpfunc) 0,
	(reprfunc) m_port_repr,
	0, // tp_as_number 
	0, // tp_as_sequence 
	0, // tp_as_mapping 
	(hashfunc) 0,
	(ternaryfunc) 0,
	(reprfunc) 0,

	// more ... 
	
	0,0,0,0,

	// __doc__ 
	
	M_PORT_NAME" objects are representing MIDI output ports."
};

m_port_object *
module_m_port_new (PyObject * self, PyObject * args)
{
	int o_mode;
	m_fd * port = NULL;
	
	/* case 1 is we are given an integer fd */
	PyObject * first_arg = PyTuple_GetItem (args, 0);
	if (PyTuple_Size (args) == 1 && PyInt_Check (first_arg))
	{	
		int fd = PyInt_AsLong (first_arg);
		o_mode = fcntl (fd, F_GETFL);
		if (o_mode == -1)
		{
			PyErr_SetFromErrno (PyExc_IOError);
			return NULL;
		}
		port = new m_fd (fd, o_mode);
		port->set_connector_name ("Port (%d)", fd);
	}
	/* case 2 is we are given a path and an open mode */
	else
	{
		char * path = NULL;
		char * mode = "rw";

		if (!PyArg_ParseTuple 
			(
				args, 
				"s|s;path to midi port to add, read/write mode", 
				&path, &mode
			)
		)	return NULL;
			
		int o_mode = 0;
		char * m = mode;
		while (*m)
		{
			switch (*m++)
			{
			case 'r':
				o_mode |= O_RDONLY;
				break;
			case 'w':
				o_mode |= O_WRONLY;
				break;
			case ' ':
				continue;
				
			default:
				PyErr_Format (PyExc_ValueError, "invalid open mode for "M_PORT_NAME"(): %s", mode);
				return NULL;
			}
		}
		
		if (o_mode == (O_RDONLY | O_WRONLY))
			o_mode = O_RDWR;

		port = new m_fd (path, o_mode);
		port->set_connector_name ("Port ('%s', '%s')", path, mode);
	}

	if (port->fd < 0)
	{
		PyErr_SetFromErrno (PyExc_IOError);
		delete port;
		return NULL;
	}
	
	m_port_object * port_object = m_port_new_from_port (port);
	if (!port_object)
		delete port;
	return port_object;
}



/*
 *	create two ports representing a socketpair(2)
 *	aka a loopback port.
 *
 */
 
PyObject *
module_m_loopback_new (PyObject * self, PyObject * args)
{
	if (!PyArg_ParseTuple (args, ""))
		return NULL;
	
	PyObject * result = PyTuple_New (2);
	if (!result)
		return NULL;

	int fd[2];
	int r = socketpair (AF_LOCAL, SOCK_STREAM, 0, fd);
	if (r < 0)
	{
		PyErr_SetFromErrno (PyExc_IOError);
		return NULL;
	}
	
	m_fd * port_0 = new m_fd (fd[0], O_RDWR);
	port_0->set_connector_name ("<LoopbackPort fd=%d>", fd[0]);
	
	m_fd * port_1 = new m_fd (fd[1], O_RDWR);
	port_1->set_connector_name ("<LoopbackPort fd=%d>", fd[1]);

	m_port_object * port_object_1 = NULL;
	m_port_object * port_object_0 = NULL;
	
	port_object_0 = m_port_new_from_port (port_0);
	if (!port_object_0)
		goto FAIL;

	port_object_1 = m_port_new_from_port (port_1);
	if (!port_object_1)
		goto FAIL;

	PyTuple_SetItem (result, 0, (PyObject *) port_object_0);
	PyTuple_SetItem (result, 1, (PyObject *) port_object_1);
	return result;
	
	/* TODO: mail python-list. does this add one reference too many? */
	return Py_BuildValue ("OO", port_object_0, port_object_1);

	FAIL:
	if (port_object_1)
		m_port_del (port_object_1);
	else
		delete port_1;
		
	if (port_object_0)
		m_port_del (port_object_0);
	else
		delete port_0;
	PyObject_DEL (result);
	return NULL;
}


/*
 *	create an ascii port (see midi-ascii-port.cc and midi-io.h)
 *
 */
 
PyObject *
module_m_ascii_port_new (PyObject * self, PyObject * args)
{
	if (PyTuple_Size (args) != 1)
	{
		PyErr_SetString (PyExc_SyntaxError, "needs a file descriptor or "
			"the path to a writeable file/fifo.");
		return NULL;
	}
	
	PyObject * arg = PyTuple_GetItem (args, 0);
	
	if (PyInt_Check (arg)) /* is a file descriptor */
	{
		return (PyObject *) m_port_new_from_port (
			new m_ascii_port (PyInt_AS_LONG (arg), false) /* we don't own the fd */
		);
	}
	
	if (!PyString_Check (arg)) /* not a string either */
	{
		PyErr_SetString (PyExc_SyntaxError, "needs a file descriptor or "
			"the path to a writeable file/fifo.");
		return NULL;
	}

	int fd = open (PyString_AS_STRING (arg), O_WRONLY | O_APPEND);
	if (fd == -1)
	{
		PyErr_SetFromErrnoWithFilename (PyExc_IOError, PyString_AS_STRING (arg));
		return NULL;
	}

	return (PyObject *) m_port_new_from_port (
		new m_ascii_port (fd, true) /* we do own the fd */
	);
}







#ifdef WITH_ALSA

class py_alsa_iterator
:
public alsa_port_iterator
{
	public:
		PyObject * list;
		
									py_alsa_iterator() {list = PyList_New (0);}
		virtual			 ~py_alsa_iterator() {Py_XDECREF (list);}
								 
		virtual void	next (snd_seq_client_info_t * c, snd_seq_port_info_t * p);
};

void
py_alsa_iterator::next (snd_seq_client_info_t * c, snd_seq_port_info_t * p)
{
	char mode[3]="  ";
	
	int capability = snd_seq_port_info_get_capability (p);
	
	if (capability & SND_SEQ_PORT_CAP_SUBS_READ)
		mode[0]='r';
	if (capability & SND_SEQ_PORT_CAP_SUBS_WRITE)
		mode[1]='w';
	
	PyList_Append (
		list,
		Py_BuildValue (
			"(ssii)", 
			snd_seq_port_info_get_name (p),
			mode,
			snd_seq_port_info_get_client (p),
			snd_seq_port_info_get_port (p)
		)
	);
}

m_port_object *
module_m_alsa_port_new (PyObject * self, PyObject * args)
{
	int major, minor;
	
	alsa_port * port = new alsa_port();
	if (errno)
	{
		delete port;
		PyErr_SetFromErrno (PyExc_SystemError);
		return NULL;
	}
	
	if (!PyTuple_Size (args))
	{
		py_alsa_iterator iterator;

		port->iterate (&iterator);
		
		delete port;

		if (PyErr_Occurred())
			return NULL;

		Py_INCREF (iterator.list);
		return (m_port_object *) iterator.list;
	}
	
	if (PyTuple_Size (args) == 1)
	{
		char * name;
		if (!PyArg_ParseTuple 
			(
				args,
				"s;ALSA client name",
				&name
			)
		) return NULL;

		port->connect (name);
		port->set_connector_name (name);
	}
	else 
	{
		if (!PyArg_ParseTuple 
			(
				args, 
				"ii;ALSA client and port numbers", 
				&major, &minor
			)
		)	return NULL;

		port->connect (major, minor);
		port->set_connector_name ("ALSAPort (%d, %d)", major, minor);
	}
		
	if (!port->can_read() && !port->can_write())
	{
		delete port;
		PyErr_SetFromErrno (PyExc_SystemError);
		return NULL;
	}

	m_port_object * port_object = m_port_new_from_port (port);
	if (!port_object)
		delete port;
	return port_object;
}

#endif  /* WITH_ALSA */

