/*
 *	python-midi-sync.cc
 *
 */

#include "python-types.h"

m_sync_object *
m_sync_new_from_io (m_io * io)
{
	m_sync_object * self = PyObject_NEW (m_sync_object, &m_sync_py_type);
	if (!self) 
	{
		PyErr_NoMemory();
		return NULL;
	}
	
	_dbgprintf (DBG_REF, "<m_sync_object %x>\n", (int) self);

	self->io = io;
	self->sync = &io->sync;

	C_INCREF (self->io);
	
	return self;
}

static void
m_sync_del (m_sync_object * self)
{
	_dbgprintf (DBG_REF, "<~m_sync_object %x>\n", (int) self);

	C_DECREF (self->io);
	
	PyMem_DEL (self);
}

/*
 *	m_sync.add (m_part *)
 *
 */

static PyObject *
m_sync_add (m_sync_object * self, PyObject * args)
{
	m_song_object * object;
	if (!PyArg_ParseTuple 
		(
			args,
			"O!;"M_SONG_NAME, 
			&m_song_py_type, &object
		)
	) return NULL;

	self->sync->add_song (object->song);

	Py_INCREF (Py_None); 
	return Py_None;
}

/*
 *	m_sync.remove (m_part *)
 *
 */

static PyObject *
m_sync_remove (m_sync_object * self, PyObject * args)
{
	m_song_object * object;
	if (!PyArg_ParseTuple 
		(
			args,
			"O!;"M_SONG_NAME, 
			&m_song_py_type, &object
		)
	) return NULL;

	if (!self->sync->remove_song (object->song))
	{
		PyErr_SetString (PyExc_AssertionError, "this "M_SONG_NAME" is not in the "M_SYNC_NAME".");
		return NULL;
	}

	Py_INCREF (Py_None); 
	return Py_None;
}

/*
 *	m_sync_methods
 *
 */

static PyMethodDef
m_sync_methods[] = 
{
	{"add", (PyCFunction) m_sync_add, METH_VARARGS, 
	 "add ("M_SONG_NAME")\n"
	 "adds a "M_SONG_NAME"."},
	{"remove", (PyCFunction) m_sync_remove, METH_VARARGS, 
	 "remove ("M_SONG_NAME")\n"
	 "removes a "M_SONG_NAME"."},
	{NULL, NULL}
};

/*
 *	m_sync.getattr
 *
 */

static PyObject *
m_sync_getattr (m_sync_object * self, char * name)
{
	if (!strcmp (name, "locked"))
		return PyInt_FromLong (self->sync->locked);
	if (!strcmp (name, "use"))
		return PyInt_FromLong (self->sync->use);
		
	if (!strcmp (name, "__members__"))
	{
		PyObject * result = PyList_New (2);
		if (result)
		{
			PyList_SetItem (result, 0, PyString_FromString ("locked"));
			PyList_SetItem (result, 1, PyString_FromString ("use"));
		
			if (PyErr_Occurred()) 
			{
				Py_DECREF (result);
				result = NULL;
			}
		}
		return result;
	}
	
	return Py_FindMethod (m_sync_methods, (PyObject *) self, name);
}	

/*
 *	m_sync.setattr
 *
 */

static int
m_sync_setattr (m_sync_object * self, char * name, PyObject * arg)
{
	if (!strcmp (name, "use"))
	{
		int use = PyInt_AsLong (arg);
		if (PyErr_Occurred())
			return -1;
		self->sync->use = use;
		return 0;
	}
	if (!strcmp (name, "locked"))
		PyErr_SetString (PyExc_AttributeError, "read-only attribute.");
	else
		PyErr_SetString (PyExc_AttributeError, name);
	return -1;
}	


/*
 *	m_sync as sequence
 *
 */

static int
m_sync_len (m_sync_object * self)
{
	return self->sync->songs.count;
}

/*
 *	= m_sync[i]
 *
 */

static m_song_object *
m_sync_item (m_sync_object * self, int i)
{
	if (i < 0 || i >= self->sync->songs.count)
	{
		PyErr_SetString (PyExc_IndexError, "index is out of range.");
		return NULL;
	}
	
	self->sync->songs.mutex.lock();
	m_song_object * song_object = m_song_new_from_song (self->sync->song (i));
	self->sync->songs.mutex.unlock();

	if (!song_object)
	{
		PyErr_SetString (PyExc_MemoryError, "failed to create a "M_SONG_NAME".");
		return NULL;
	}

	return song_object;
}
	
/*
 *	m_sync[i] =
 *
 */

static int
m_sync_ass_item (m_sync_object * self, int i, PyObject * sync_object)
{
	if (i < 0 || i >= self->sync->songs.count)
	{
		PyErr_SetString (PyExc_IndexError, "index out of range.");
		return -1;
	}

	if (sync_object == NULL)
	{
		self->sync->remove_song (self->sync->song(i));
		return 0;
	}

	PyErr_SetString (PyExc_NotImplementedError, "sorry I can't do that.");
	return -1;
}

/*
 *	m_sync python type
 *
 */

static 
PySequenceMethods 
m_sync_as_sequence = 
{
	(inquiry) m_sync_len, //sq_length
	(binaryfunc) 0, //sq_concat
	(intargfunc) 0, //sq_repeat
	(intargfunc) m_sync_item, //sq_item
	(intintargfunc) 0, //sq_slice
	(intobjargproc) m_sync_ass_item, //sq_ass_item
	(intintobjargproc) 0, //sq_ass_slice
	(objobjproc) 0, //sq_contains
};


PyTypeObject 
m_sync_py_type =
{
	PyObject_HEAD_INIT (&PyType_Type)
	0, // ob_size 
	M_SYNC_NAME,
	sizeof (m_sync_object),
	0, // tp_itemsize 

	// methods 
	(destructor) m_sync_del,
	(printfunc) 0,
	(getattrfunc) m_sync_getattr,
	(setattrfunc) m_sync_setattr,
	(cmpfunc) 0,
	(reprfunc) 0,
	0, // tp_as_number 
	&m_sync_as_sequence, // tp_as_sequence 
	0, // tp_as_mapping 
	(hashfunc) 0,
	(ternaryfunc) 0,
	(reprfunc) 0,

	// more ... 
	
	0,0,0,0,

	// __doc__ 
	
	"The "M_SYNC_NAME" object drives a "M_SONG_NAME"."
};

#if 0
m_sync_object *
module_m_sync_new (PyObject * self, PyObject * args)
{
	m_sync * sync;
	
	sync = m_sync::get();

	if (!sync)
	{
		PyErr_SetString (PyExc_SystemError, "Cannot instantiate a sync.");
		return NULL;
	}

	m_sync_object * sync_object = m_sync_new_from_sync (sync);

	if (!sync_object)
		delete sync;
		
	return sync_object;
}
#endif
